
/*+ --------------------------------------------------- + 
  +		IDA*_CR Search Algorithms		+
  + --------------------------------------------------- +*/

#define	PRINT_ITER_PROFILE	printf \
("iter=%d, approx=%g, accu=%g, exp_opt(glub=%g)=%g, gllb=%g, threshold=%g, time=%d\n", \
iter, child_sp->approx.approx, approx2accu (child_sp->approx.approx), \
(float) child_sp->glub, (float) expected_opt(child_sp), \
(float) child_sp->gllb, \
(float) child_sp->constr.bound, get_time(child_sp));

void set_ida_cr_device ();

void ida_cr_algorithm (sp)
  search_ *sp;
{
  search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal; /** not important, only as a placeholder **/
  int iter = 0;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  set_search_message (&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search (sp);

  while (++iter) {
    set_ida_cr_device (child_sp, iter);
    put_child_constr (child_sp, sp, 1.0, 1.0, 1.0);

    idpd_init();
    msg.feasible = NO;
    switch (sp->strategy) {
    case DFS:  term_signal = dfs_primitive (child_sp, sp, &msg); break;
    case GDFS: term_signal = gdfs_primitive (child_sp, sp, &msg); break;
    case BAND: term_signal = band_primitive (child_sp, sp, &msg); break;
    case FBS:  term_signal = first_band_primitive (child_sp, sp, &msg); break;
    default:   error ("ida_algorithm: no such feasible primitive"); break;
    }

    eval_final_approx (child_sp);
#ifdef DEBUG
if (cmd.debug >= 1) { PRINT_ITER_PROFILE }
#endif
    merge_stat_to_parent (child_sp, sp);
    merge_sol_to_parent (child_sp, sp);
    if (get_approx (child_sp) <= sp->approx.achieved)
      merge_approx_to_parent (child_sp, sp);
    merge_threshold_to_parent (child_sp, sp);

    if (msg.feasible == YES || get_time (sp) >= sp->constr.time) break;
    if (end_of_opt_sol_search (sp) == YES) break;

    (sp->child = child_sp = create_brother_search (child_sp))->parent = sp;
  }
}


void set_ida_cr_device (sp, iter)
  search_ *sp;
  int iter;
{
  int i;
  long time, count;

  if (iter == 1)
  {
       sp->constr.bound = ROOT->lowb+1;
  } else {
      time = sp->brother->stat.expanded * (sp->brother->alg.ida_cr->r - 1);
      for (i = 0, count = 0; i < IDA_CR_NUM_BIN; i++)
      {
	count += sp->brother->alg.ida_cr->bin[i];
	if (count >= time) break;
      }
      if (count < time)
      {
	for (i = IDA_CR_NUM_BIN - 1; i >= 0; i--)
	{
	  if (sp->brother->alg.ida_cr->bin[i]) break;
	}
#ifdef DEBUG
	if (cmd.debug >= 1)
	  printf ("set_ida_cr_device: not sufficient nodes\n");
#endif
      }
#ifdef DEBUG
	if (cmd.debug >= 1)
	  printf ("set_ida_cr_device: time=%d, count=%d, bin number=%d\n",
			time, count, i);
#endif
      sp->constr.bound = sp->brother->alg.ida_cr->x * (1 + ((float) (++i)) / IDA_CR_NUM_BIN);
  }

  sp->alg.ida_cr->x = sp->constr.bound;

  Next_IDA_Threshold = problem.huge;
}


void ida_cr_bin_update (sp, node)
search_ *sp;
node_ *node;
{
    domain x = sp->alg.ida_cr->x;
    domain diff = node->lowb - x;
    int which;

    if (diff > 0 && diff <= x)
    {
	which = ((float) diff) * IDA_CR_NUM_BIN / x;
	(sp->alg.ida_cr->bin[which])++;
#ifdef DEBUG
        if (cmd.debug >= 4)
	{
	  printf ("node->lowb=%f, bin=%d\n", ((float) node->lowb), which);
	}
#endif
    }
}

