/*+ --------------------------------------------------- + 
  +		IRD* Search Algorithms			+
  + algorithms:						+
  +   . unary IRD* (uIRD*): u[a] & u[th]		+
  +	using only one pruning device.			+
  +   . binary IRD* (bIRD*):				+
  +	using both pruning devices.			+
  + --------------------------------------------------- +*/


void set_uird_device (), set_uird_approx (), set_uird_threshold ();
void set_bird_device (), set_bird_approx (), set_bird_threshold ();
void ird_set_typeI_approx (), ird_set_typeII_approx ();
void ird_set_regress_approx (), ird_set_last_approx ();
void ird_set_ird_threshold (), ird_set_naive_threshold ();
void ird_set_static_threshold (), ird_set_typeI_static_threshold ();
void ird_set_typeII_static_threshold (), ird_set_predictive_threshold ();
void ird_set_typeI_predictive_threshold ();
void ird_set_typeII_predictive_threshold ();
void reset_th_environ (), set_th_entry ();

float get_g_factor (), get_r_factor ();
domain ird_get_predicted_glub ();
yesno_ check_span (), get_bin_vs_time (), get_th_vs_time ();
yesno_ get_glub_vs_time (), get_approx_vs_time (), get_glub_vs_approx ();
yesno_ get_accu_vs_time ();
iter_attr_ get_iteration_attribute ();
set_th_ *get_set_th ();


/* 		uIRD*			*/

void uird_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal;
  yesno_ break_signal = NO;
  int iter;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;
  set_search_message(&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search(sp);
  break_signal = NO;
  for (iter = 1; 1; iter++) {
    RTS_IRD_iter = iter;
#ifndef NO_INHERIT
    inherit_sol_from_parent(child_sp, sp);
#endif
    put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
    set_uird_device(child_sp, iter);
    idpd_init();
    term_signal = (sp->strategy == DFS) ? dfs_primitive(child_sp, sp, &msg) :
					  gdfs_primitive(child_sp, sp, &msg);
    eval_final_approx(child_sp);
#ifdef DEBUG
if (cmd.debug >= 1)
  printf("iter=%d, approx(glub=%g)=%g, threshold=%g, time=%d\n",
	iter, (float) child_sp->glub, (float) expected_opt(child_sp),
        (float) child_sp->alg.uird->th, get_time(child_sp));
#endif
    merge_stat_to_parent(child_sp, sp);
    if (get_approx(child_sp) <= sp->approx.achieved) {
      merge_approx_to_parent(child_sp, sp);
      merge_sol_to_parent(child_sp, sp);
    }
    merge_threshold_to_parent(child_sp, sp);

    switch (sp->alg.uird->alg) {
      case UNARY_APPROX: 
	if (child_sp->approx.approx <= 0.0) break_signal = YES;
	break;
      case UNARY_TH:
	if (msg.feasible == YES) break_signal = YES;
	break;
      default:	error("uird_algorithm: no such pruning device");  break;
    }
    if (break_signal == YES) break;

    if (end_of_first_sol_search (sp) == YES) break;
    if (end_of_opt_sol_search (sp) == YES) break;

    msg.feasible = NO;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
  }
}


void set_uird_device (sp, iter)
  search_ *sp;
  int iter;
{
  switch (sp->alg.uird->alg) {
    case UNARY_APPROX:	set_uird_approx(sp, iter);	break;
    case UNARY_TH:	set_uird_threshold(sp, iter);	break;
    default:	error("set_uird_device: no such pruning device");  break;
  }
}


void set_uird_approx (sp, iter)
  search_ *sp;
  int iter;
{
  switch (sp->alg.uird->a_type) {
    case TYPE_I:  ird_set_typeI_approx(sp, iter, &(sp->alg.uird->basic)); break;
    case TYPE_II: ird_set_typeII_approx(sp, iter, &(sp->alg.uird->basic)); break;
    default: error("set_uird_approx: no such type"); break;
  }
}


void set_uird_threshold (sp, iter)
  search_ *sp;
  int iter;
{ uird_ *alg = sp->alg.uird;
  domain ub = sp->parent->root_ub;
  domain lb = sp->parent->root_lb;
  float g = get_g_factor (sp);
 
  if (iter == 1) {
    /* set init uird threshold */
    sp->constr.bound = alg->th = lb + g * (ub - lb);
  }
  else { /* set intermediate-iteration uird threshold */
    ird_set_ird_threshold(sp, iter, &(alg->basic));
  }

  reset_th_environ(sp);
}


void uird_entry (sp, node)
  search_ *sp;
  node_ *node;
{
  switch (sp->alg.uird->alg) {
    case UNARY_APPROX:				break;
    case UNARY_TH:	set_th_entry(sp, node);	break;
    default: error("uird_entry: no such pruning device");  break;
  }
}


/* 		bIRD*			*/
void bird_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal;
  int iter;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;
  set_search_message(&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search(sp);
  for (iter = 1; 1; iter++) {
    RTS_IRD_iter = iter;
#ifndef NO_INHERIT
    inherit_sol_from_parent(child_sp, sp);
#endif
    put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
    set_bird_device(child_sp, iter);
    idpd_init();
    term_signal = (sp->strategy == DFS) ? dfs_primitive(child_sp, sp, &msg) :
					  gdfs_primitive(child_sp, sp, &msg);
    eval_final_approx(child_sp);
#ifdef DEBUG
if (cmd.debug >= 1)
  printf("iter=%d, approx(glub=%g)=%g, threshold=%g, time=%d\n",
	iter, (float) child_sp->glub, (float) expected_opt(child_sp),
        (float) child_sp->alg.bird->th, get_time(child_sp));
#endif
    merge_stat_to_parent(child_sp, sp);
    if (get_approx(child_sp) <= sp->approx.achieved) {
      merge_approx_to_parent(child_sp, sp);
      merge_sol_to_parent(child_sp, sp);
    }
    merge_threshold_to_parent(child_sp, sp);
    if (child_sp->constr.bound == problem.huge) {
      if (child_sp->approx.approx <= 0.0) break;
    }
    else if (msg.feasible == YES) break;

    if (end_of_first_sol_search (sp) == YES) break;
    if (end_of_opt_sol_search (sp) == YES) break;

    msg.feasible = NO;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
  }
}


void set_bird_device (sp, iter)
  search_ *sp;
  int iter;
{
    /* Ideally, approx and threshold should be set iteratively,
     * until both converge.
     * However, it may not converge and here it is emulated by
     * setting approx first and then setting threshold.
     */
    set_bird_approx(sp, iter);
    set_bird_threshold(sp, iter);
}


void set_bird_approx (sp, iter)
  search_ *sp;
  int iter;
{ 
  ird_set_typeI_approx(sp, iter, &(sp->alg.bird->basic));
}


void set_bird_threshold (sp, iter)
  search_ *sp;
  int iter;
{ bird_ *alg = sp->alg.bird;
  domain ub = sp->parent->root_ub;
  domain lb = sp->parent->root_lb;
  float g = get_g_factor (sp);
 
  if (iter == 1) {
    /* set init bird threshold */
    sp->constr.bound = alg->th = lb + g * (ub - lb);
  }
  else { /* set intermediate-iteration uird threshold */
    ird_set_ird_threshold(sp, iter, &(alg->basic));
  }

  reset_th_environ(sp);
}


void bird_entry (sp, node)
  search_ *sp;
  node_ *node;
{ set_th_entry(sp, node); }


/*+ --------------------------------------------------- + 
  + 	  IRD* Approximation Setting Routines		+
  + algorithms:						+
  +   . naive approximation (Type I):			+
  +     linear gradient, like sTCGD.			+
  +   . regression approximation (Type II):		+
  +	regressed gradient, like pTCGD.			+
  +   * last approximation: regression gradient.	+
  + --------------------------------------------------- +*/


/** linear-gradient approximation **/
void ird_set_typeI_approx (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ float sched;
  float x_root = sp->parent->approx.x_root_approx;
  float g = get_g_factor (sp);
  int count;
  search_ *brother;
  float glub[2];

  if (cmd.xon == YES) {
    x_root = approx2accu (x_root);
    sched = x_root + (1.0 - x_root) * iter * g;
  } else {
    sched = x_root * (1.0 - iter * g);
  }

  if (basicp->shortcut && check_span (sp, FULL_SPAN) == YES) {

    count = 0;
    for (brother = sp->brother; brother; brother = brother->brother)
      if (is_feasible_time (get_time (brother))) {
        glub[count++] = (float) (brother->glub);
        if (count >= 2) break;
      } else break;

    if (glub[1] - glub[0] <= glub[0] * basicp->margin)
      /** if the upper bounds changes only a little bit,
       ** then it is likely that upb is very close to optimum. **/
      sched = (cmd.xon == YES) ? 1.0 : 0.0;
  }

  if (cmd.xon == YES) sched = accu2approx (sched);
  sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}


void ird_set_regress_approx (sp, iter, span, basicp)
  search_ *sp;
  int iter;
  span_ span;
  ird_basic_ *basicp;
{ float sched, a;
  double beta[2];
  float x_root = sp->parent->approx.x_root_approx;
  float g = get_g_factor (sp);

  if (cmd.xon == YES) {
    x_root = approx2accu (x_root);
    sched = x_root + (1.0 - x_root) * iter * g;
  } else {
    sched = x_root * (1.0 - iter * g);
  }

  if (cmd.xon == YES) {
    if (get_accu_vs_time(sp, beta, span) == YES) {
      sched = beta[0] + beta[1] * log10(((double) sp->constr.time));
      a = get_approx(sp->parent);
      a = approx2accu (a);
      if (sched <= a) sched = a + (1.0 - x_root) * g;
    }
    sched = accu2approx (sched);
  } else {
    if (get_approx_vs_time(sp, beta, span) == YES) {
      sched = beta[0] + beta[1] * log10(((double) sp->constr.time));
      if (sched >= get_approx(sp->parent))
	sched = get_approx(sp->parent) - sp->parent->approx.root_approx * g;
    }
  }

  sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}


void ird_set_typeII_approx (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ ird_set_regress_approx (sp, iter, FULL_SPAN, basicp); }



/*+ --------------------------------------------------- + 
  + 	    IRD* Threshold Setting Routines		+
  + algorithms:						+
  +   . naive algorithm.				+
  +   . static algorithm: type I & type II		+
  +   . predictive algorithm: type I & type II		+
  + --------------------------------------------------- +*/


void ird_set_ird_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{
  switch ((get_set_th(sp))->alg) {
    case NAIVE_TH:      ird_set_naive_threshold(sp, iter, basicp);      break;
    case STATIC_TH:     ird_set_static_threshold(sp, iter, basicp);     break;
    case PREDICTIVE_TH: ird_set_predictive_threshold(sp, iter, basicp); break;
    default: error("ird_set_ird_threshold: no such threshold setting"); break;
  }
}


void ird_set_naive_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ domain lb = sp->parent->root_lb;
  domain ub = sp->parent->root_ub;
  float g = get_g_factor (sp);
  domain th;
  double beta[2];
  int count;
  search_ *brother;
  float glub[2];

  sp->constr.bound = lb + iter * g * (ub - lb);

  if (basicp->shortcut && check_span (sp, FULL_SPAN) == YES) {

    count = 0;
    for (brother = sp->brother; brother; brother = brother->brother)
      if (is_feasible_time (get_time (brother))) {
        glub[count++] = (float) (brother->glub);
        if (count >= 2) break;
      } else break;
 
    if (glub[1] - glub[0] <= glub[0] * basicp->margin)
      /** if the upper bounds changes only a little bit,
       ** then it is likely that upb is very close to optimum. **/
      sp->constr.bound = glub[0];
  }

  put_threshold (sp, sp->constr.bound);
}


void ird_set_static_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{
  switch ((get_set_th(sp))->type) {
    case TYPE_I:   ird_set_typeI_static_threshold(sp, iter, basicp);  break;
    case TYPE_II:  ird_set_typeII_static_threshold(sp, iter, basicp); break;
    default: 	   error("ird_set_static_threshold: no such type");   break;
  }
}


void ird_set_typeI_static_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ domain u = ird_get_predicted_glub (sp, iter, basicp);

  sp->constr.bound = (u + sp->brother->glub) * 0.5 / (1.0 + sp->approx.approx);
  put_threshold(sp, sp->constr.bound);
}


void ird_set_typeII_static_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ search_ *brother = sp->brother;
  domain prev_th = get_threshold (sp->brother);
  domain prev_prev_glub, prev_glub, dth;
  float g = get_g_factor (sp);
  float ub, lb, delta, base;

  if (iter == 2) {
    prev_prev_glub = sp->parent->root_lb;
    prev_glub = brother->glub;
  }
  else {
    prev_prev_glub = brother->brother->glub;
    prev_glub = brother->glub;
  }
  base = 1.0 + brother->approx.approx;
  lb = prev_glub / base;
  ub = prev_prev_glub / base;
  if (prev_th < lb) {
    /* case I */
    delta = (ub - lb) * iter * g;
    dth = lb - prev_th + delta;
  }
  else {
    /* case II */
    dth = delta = (ub - lb) * g;
  }
  sp->constr.bound = prev_th + dth;
  ub = sp->glub / (1.0 + sp->approx.approx);
  if (sp->constr.bound > ub) sp->constr.bound = ub;
  put_threshold(sp, sp->constr.bound);
}


void ird_set_predictive_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{
  switch ((get_set_th(sp))->type) {
    case TYPE_I:   ird_set_typeI_predictive_threshold(sp, iter, basicp);  break;
    case TYPE_II:  ird_set_typeII_predictive_threshold(sp, iter, basicp); break;
    default: 	   error("ird_set_predictive_threshold: no such type");   break;
  }
}


void ird_set_typeI_predictive_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ search_ *brother = sp->brother;
  domain lb = sp->parent->root_lb;
  domain ub = sp->parent->root_ub;
  float g = get_g_factor(sp);
  long this_time;
  double beta[2];

  sp->constr.bound = lb + iter * g * (ub - lb);

  if (get_th_vs_time (sp, beta, FULL_SPAN) == YES) {
    this_time = get_r_factor (sp) * get_time (brother);
    if (this_time > sp->constr.time) this_time = sp->constr.time;
    sp->constr.bound = beta[0] + beta[1] * log10 ((double) this_time);
    if (sp->constr.bound < brother->constr.bound)
      sp->constr.bound = brother->constr.bound + g * (ub - lb);
  }

  put_threshold (sp, sp->constr.bound);
}


void ird_set_typeII_predictive_threshold (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ search_ *brother = sp->brother;
  domain lb = sp->parent->root_lb;
  domain ub = sp->parent->root_ub;
  float g = get_g_factor(sp);
  long this_time;
  double beta[2];

  sp->constr.bound = lb + iter * g * (ub - lb);

  if (get_bin_vs_time (sp, beta, FULL_SPAN) == YES) {
    this_time = get_r_factor (sp) * get_time (brother);
    if (this_time > sp->constr.time) this_time = sp->constr.time;
    sp->constr.bound = beta[0] + beta[1] * log10 ((double) this_time);
    if (sp->constr.bound < brother->constr.bound)
      sp->constr.bound = brother->constr.bound + g * (ub - lb);
  }

  put_threshold (sp, sp->constr.bound);
}



/*+ --------------------------------------------------- + 
  +		IRD*	Library Routines		+
  +	     Many Library Routines are in RTS		+
  + --------------------------------------------------- +*/ 


domain ird_get_predicted_glub (sp, iter, basicp)
  search_ *sp;
  int iter;
  ird_basic_ *basicp;
{ span_ span = FULL_SPAN;
  float step;
  double beta[2];
 
  if (get_glub_vs_approx (sp, beta, span) == YES)
    return ((domain) (beta[0] + beta[1] * sp->approx.approx));

  /* if situation is too tough to handle, simply use naive glub */
  step = iter * get_g_factor (sp) * (sp->parent->root_ub - sp->parent->root_lb);
  return ((domain) (sp->parent->root_ub - step));
}

