
/* June 29, 1990
** Arthur Ieumwananonthachai
**
** Add code for using IDA* search strategy
*/

/*
** April 10, 1992
** L.-C. Chu
** linked this to ISE.
*/

#define MIDA_INIT_TH_MUL	1.05

static double	sumx, sumx2, sumx3, sumx4, sumy, sumxy, sumx2y;
static int	sumn;

void mida_algorithm (sp)
search_ *sp;
{
  search_ *child_sp;
  search_message_ msg;
  float factor;
  search_ending_ termination = SEARCH_IS_COMPLETED;

  mida_init (sp);
  set_search_message (&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search (sp);
  child_sp->algorithm = sp->algorithm;


/*** LC, I don't know what is going on here.  4/10/92
  if (condition not to do BFS search)
    child_sp->constr.bound = (domain) (((float) sp->root_lb) * 0.99);
  else {
*/


  child_sp->constr.bound = sp->root_lb;
  sp->child = child_sp = create_brother_search (child_sp);
  mida_bfs_search (sp, child_sp, &msg);

  while ((expected_opt (sp) > sp->alg.mida->threshold) ||
	 (sp->alg.mida->iter == 0)) {
    mida_set_threshold (sp);

    sp->child = child_sp = create_brother_search (child_sp);
    inherit_sol_from_parent (child_sp, sp);
    child_sp->constr.bound = sp->alg.mida->threshold;
    /*    dom_init(); */

    put_child_constr (child_sp, sp, 1.0, 1.0, 1.0);
    termination = (sp->strategy == DFS) ? dfs_primitive (child_sp, sp, &msg) :
                                          gdfs_primitive (child_sp, sp, &msg);
    merge_stat_to_parent (child_sp, sp);

    if (termination == SEARCH_IS_ABORTED) mida_accounting (sp, child_sp);
    
    if ((get_approx (child_sp) < sp->approx.achieved) ||
	(termination != SEARCH_IS_ABORTED)) {
      merge_approx_to_parent (child_sp, sp);
      merge_sol_to_parent (child_sp, sp);
    }

    if (termination == SEARCH_IS_ABORTED) break;

    if (end_of_first_sol_search (sp) == YES) break;
    if (end_of_opt_sol_search (sp) == YES) break;

    if (sp->alg.mida->iter > 1)
      factor = ((float) (child_sp->stat.expanded)) /
	       child_sp->brother->stat.expanded;
    
#ifdef DEBUG
    if(cmd.debug >= 1) {
      printf("approx(global_upb=%g)=%g, mida_threshold=%g, expand=%d",
	(float) sp->glub, (float) expected_opt(sp),
	(float) child_sp->constr.bound, child_sp->stat.expanded);
      if (sp->alg.mida->iter > 1) printf(" factor=%g\n", factor);
      else printf("\n");
    }
#endif
  }

#ifndef AVOID_MIDA_FREE
  mida_free (sp);
#endif

  /* Report IDA statistics */
  if (termination != SEARCH_IS_ABORTED)
    mida_report (sp, sp->alg.mida->iter, factor);
}


/* Subroutine for initialize variable for mida* search */
mida_init (sp)
search_ *sp;
{
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;
  sp->approx.achieved = sp->approx.root_approx;

  /* set init threshold */
  sp->alg.mida->threshold = (domain) (((float) sp->root_lb) * MIDA_INIT_TH_MUL);

  if (sp->alg.mida->bin) free (sp->alg.mida->bin);
  sp->alg.mida->bin = (int *) malloc(((unsigned) sp->alg.mida->mida_n) *
				    sizeof (int));

#ifdef UPAST
  if (sp->alg.mida->oldbin) free (sp->alg.mida->oldbin);
  sp->alg.mida->oldbin = (int *) malloc(((unsigned) sp->alg.mida->mida_n) *
				       sizeof (int));
#endif

  sumx2y = sumxy = sumy = sumx4 = sumx3 = sumx2 = sumx = (double) 0.0;
  sumn = 0;
}


/* Free up allocation */
mida_free (sp)
search_ *sp;
{
/* Arthur and L.-C. could not figure out what is wrong with this
** statement.
** It caused Bus Error.
*/
  if (sp->alg.mida->bin) free (sp->alg.mida->bin);
#ifdef UPAST
  if (sp->alg.mida->oldbin) free (sp->alg.mida->oldbin);
#endif
}


/* Execute for reset variable for each iteration of IDA* */
mida_reset(sp)
search_ *sp;
{
  int i, n = sp->alg.mida->mida_n;
  int *mida_bin = sp->alg.mida->bin;
#ifdef UPAST
  int *old_bin = sp->alg.mida->oldbin;
#endif

  for(i = 0; i < n; i++) *(mida_bin+i) = 0;
#ifdef UPAST
  for(i = 0; i < n; i++) *(old_bin+i) = 0;
#endif
  sp->alg.mida->data.opt_value = sp->alg.mida->threshold;
}


/* Set threshold according to lower bound distribution in last iteration */
mida_set_threshold (sp)
search_ *sp;
{
  mida_ *mida_data = sp->alg.mida;
  double slope, dslope, term2, sqrt();
  double val, b1, b0, det, g = sp->alg.mida->g_factor;
  float x, d2, factor;
  int k, y, mida_n = mida_data->mida_n, *mida_bin = mida_data->bin;
#ifdef UPAST
  int *old_bin = mida_data->oldbin;
#endif
  domain delta, prev_threshold;
  domain up = sp->glub;

  mida_data->iter++;
  if (mida_data->iter > 1) {
    prev_threshold = sp->child->brother->constr.bound;
    delta = mida_data->threshold - prev_threshold;

#ifdef UPAST
    x = (float) sp->root_lb;
    y = 0;
    if (mida_data->iter > 2) {
      sumx4 = x * (sumx3 = x * (sumx2 = x * (sumx = x)));
      sumx2y = sumxy = sumy = 0;
      d2 = ((float) (prev_threshold - sp->root_lb)) / mida_n;
      for (k = 0; k < mida_n; k++) {
	sumx += val = x += d2;
	sumx2 += val *= x;
	sumx3 += val *= x;
	sumx4 += val * x;
#ifdef EXPFIT
	y += *(old_bin + k);
	sumy += log((double) y);
	sumxy += val = x * log((double) y);
#else EXPFIT
	sumy += y += *(old_bin + k);
	sumxy += val = x * y;
#endif EXPFIT
	sumx2y += val * x;
      }
      sumn = mida_n + 1;
#ifdef DEBUG
      if (cmd.debug >= 1)
	printf("# nodes below previous threshold = %d\n",y);
#endif DEBUG
    } else
      x = (float) prev_threshold;
#else UPAST
    x = (float) prev_threshold;
    y = sp->child->brother->stat.expanded;
#ifdef NOPAST
#ifdef EXPFIT
    if (y > 0) {
      sumx4 = x * (sumx3 = x * (sumx2 = x * (sumx = x)));
      sumx2y = x * (sumxy = x * (sumy = log((double) y)));
      sumn = 1;
    } else
      sumx4 = sumx3 = sumx2 = sumx = sumx2y = sumxy = sumy = sumn = 0;
#else EXPFIT
    sumx4 = x * (sumx3 = x * (sumx2 = x * (sumx = x)));
    sumx2y = x * (sumxy = x * (sumy = y));
    sumn = 1;
#endif EXPFIT
#endif NOPAST
#endif UPAST
    d2 = ((float) delta) / mida_n;
    for (k = 0; k < mida_n; k++) {
#ifdef EXPFIT
      y += *(mida_bin + k);
      if (y > 0) {
#endif EXPFIT
      sumx += val = x += d2;
      sumx2 += val *= x;
      sumx3 += val *= x;
      sumx4 += val * x;
#ifdef EXPFIT
      sumy += log((double) y);
      sumxy += val = x * log((double) y);
#else EXPFIT
      sumy += y += *(mida_bin + k);
      sumxy += val = x * y;
#endif EXPFIT
      sumx2y += val * x;
#ifdef EXPFIT
      } else sumn--;
#endif EXPFIT
    }
    sumn += mida_n;
#ifdef EXPFIT
    dslope = 0;
    b0 = (sumx2 * sumy) - (sumx * sumxy);
    if (y > 5) {
      det = (sumn * sumx2) - (sumx * sumx);
      b1 = (sumn * sumxy) - (sumx * sumy);
    } else {
      det = 2 * (x - prev_threshold);
      b1 = log((double) g);
    }
#else EXPFIT
    det = (sumn * sumx2 * sumx4) + (2 * sumx * sumx2 * sumx3) -
	  (sumx2 * sumx2 * sumx2) - (sumn * sumx3 * sumx3) -
	  (sumx * sumx * sumx4);
    b0 = sumy * ((sumx2 * sumx4) - (sumx3 * sumx3)) +
         sumxy * ((sumx2 * sumx3) - (sumx * sumx4)) +
         sumx2y * ((sumx * sumx3) - (sumx2 * sumx2));
    b1 = sumy * ((sumx2 * sumx3) - (sumx * sumx4)) +
	  sumxy * ((sumn * sumx4) - (sumx2 * sumx2)) +
	  sumx2y * ((sumx * sumx2) - (sumn * sumx3));
    dslope = sumy * ((sumx * sumx3) - (sumx2 * sumx2)) +
	     sumxy * ((sumx * sumx2) - (sumn * sumx3)) +
	     sumx2y * ((sumn * sumx2) - (sumx * sumx));
#endif EXPFIT
    slope = b1 + 2 * dslope * x;
    if ((slope/det) <= 0) {
#ifdef DEBUG
      if (cmd.debug >= 1)
        printf("dslope = %g, slope = %g < 0: change to 1st order approx at midpoint\n",
               dslope*2/det,slope/det);
#endif
      slope = b1 + 2 * dslope * (x - (delta / 2.0));
      dslope = 0;
      if ((slope/det) <= 0) {
#ifdef DEBUG
        if (cmd.debug >= 1)
          printf("midpoint slope = %g < 0: so set slope = 0\n",slope/det);
#endif
        slope = 0;
      }
    }

#ifndef TAYLOR1
    if (dslope != 0) {
      term2 = 4 * (dslope / det) * (sp->child->stat.expanded * (g - 1));
      term2 += (slope * slope) / (det * det);
      if (term2 >= 0)
        mida_data->threshold += (float) ((sqrt(term2) - (slope/det)) * det / dslope);
#ifdef DEBUG
      else if (cmd.debug >= 1)
	printf("square root term = %g < 0: so change to 1st order approx\n",
               term2);
#endif
    }
    if ((term2 < 0) || (dslope == 0))
#endif TAYLOR1
      if (slope < 1.0e-5) {
        if (fabs((float) mida_data->threshold) <= 1.0)
          mida_data->threshold = 100;
        else if (mida_data->threshold > 1)
          mida_data->threshold *= 100;
        else
          mida_data->threshold *= -99;
      } else
#ifdef EXPFIT
        mida_data->threshold += log((double) g) * det / slope;
#else EXPFIT
        mida_data->threshold += (sp->child->stat.expanded * (g - 1)) *
	                       det / slope;
#endif EXPFIT

    if (mida_data->threshold > up) {
#ifdef DEBUG
      if (cmd.debug >= 1)
	printf("threshold = %g > global upb = %g: set threshold to upb\n",
               (float) mida_data->threshold,(float) up);
#endif
       mida_data->threshold = up;
    }

     /* Code for update statistics */
    if (mida_data->iter > 2) {
      factor = ((float) (sp->child->stat.expanded)) /
	       sp->child->brother->stat.expanded;
#ifdef DEBUG
      if (cmd.debug >= 1)
	printf(" factor=%g\n", factor);
#endif
      mida_data->data.sqr_exp_factor += (factor * factor);
      mida_data->data.tot_exp_factor += factor;
      if (factor > mida_data->data.max_exp_factor)
	mida_data->data.max_exp_factor = factor;
      if (factor < mida_data->data.min_exp_factor)
	mida_data->data.min_exp_factor = factor;
    }
  }

#ifdef DEBUG
  if (cmd.debug >= 1)
    if (mida_data->iter > 1) {
      x = (float) prev_threshold;
      y = sp->child->brother->stat.expanded;
#ifdef NOPAST
      printf("{{%g,%d},",x,y);
#endif
      for (k = 1; k <= mida_n; k++) {
	x += d2;
	y += *(mida_bin + k - 1);
        printf("{%g,%d},",x,y);
      }
      printf("}\n slope = %g, dslope = %g\n",slope/det,dslope*2/det);
      printf("b0 = %g, b1 = %g, b2 = %g\n",b0/det,b1/det,dslope/det);
    }
  if (cmd.debug >= 3) {
    printf("IDA iteration %d\n",mida_data->iter);
    printf("Previous iteration expansion = %d\n",sp->child->stat.expanded);
    printf("New threshold = %g\n",(float) mida_data->threshold);
  }
#endif

  mida_reset(sp);              /* Reset other mida variable */
}

/* This routine to update histogram of lower bound distribution */
mida_bin_update(sp, node)
search_ *sp;
node_ *node;
{
  int index;
  domain prev_threshold;

  if (sp->strategy == BFS) {
    index = sp->stat.expanded;
    *(sp->parent->alg.mida->bfsbin + index) =
       sp->parent->alg.mida->threshold = node->lowb;
    return;
  }

  prev_threshold = sp->brother->constr.bound;
  if (node->lowb > prev_threshold) {
    index = ceil((node->lowb - prev_threshold) * sp->parent->alg.mida->mida_n /
		 ((float) (sp->constr.bound - prev_threshold))) - 1;
    (*(sp->parent->alg.mida->bin + index))++;
  }
#ifdef UPAST
  else {
    if (node->lowb == sp->root_lb) index = 0;
    else
      index = ceil((node->lowb - sp->root_lb) * sp->parent->alg.mida->mida_n /
		   ((float) (prev_threshold - sp->root_lb))) - 1;
    if ((index < 0) || (index >= sp->parent->alg.mida->mida_n))
      printf("MIDA ERROR: index = %d, node lowb = %g less than root_lb = %g\n",
	     index, (float) node->lowb, (float) sp->root_lb);
    else
      (*(sp->parent->alg.mida->oldbin + index))++;
  }
#endif
}

/* This routine to update statistics when get a better solution */
mida_entry(sp, node)
search_ *sp;
node_ *node;
{
  mida_stat_ *mida_data = &(sp->parent->alg.mida->data);
  int index1,index2, i, mida_n = sp->parent->alg.mida->mida_n;
  int *mida_bin = sp->parent->alg.mida->bin;
  domain prev_threshold = sp->brother->constr.bound, upb = node->upb;

  if ((sp->strategy != BFS) && (upb < mida_data->opt_value)) {
    index1 = ceil((mida_data->opt_value - prev_threshold) * mida_n /
                  ((float) (sp->constr.bound - prev_threshold))) - 1;
    index2 = ceil((upb - prev_threshold) * mida_n /
                  ((float) (sp->constr.bound - prev_threshold))) - 1;
    mida_data->opt_value = upb;
    mida_data->b_cost = 0;
    for (i = index1; i > index2; i--) mida_data->b_cost += *(mida_bin+i);
    mida_data->mida_cost += mida_data->b_cost;
    for (i = mida_n-1; i > index1; i--)
      mida_data->b_cost += *(mida_bin + i);
  }
}


mida_report(sp,iter,factor)
search_ *sp;
int iter;
float factor;
{
  mida_stat_ *mida_data = &(sp->alg.mida->data);
  search_ *temp;
  int i, index, bfs_node, mida_node, count = 0;
  int mida_n = sp->alg.mida->mida_n, *mida_bin = sp->alg.mida->bin;
  double q_ratio, ratio, u_ratio, mean;
  domain prev_threshold;

  if (cmd.debug >= 1) {
    printf("IDA Summary Report\n");
    printf("Number of iteration = %d, Total expand = %d, Total gen = %d\n",
           iter,sp->stat.expanded,sp->stat.generated);
  }

  prev_threshold = sp->child->brother->constr.bound;
  index = ceil((mida_data->opt_value - prev_threshold) * mida_n /
	       ((float) (sp->child->constr.bound - prev_threshold))) - 1;
  for (i = mida_n-1; i > index; i--)
    count += *(mida_bin + i);
  if (mida_data->b_cost > 0)
    q_ratio = ((double) mida_data->b_cost) / count;
  else
    q_ratio = (double) 0;

  bfs_node = sp->child->stat.expanded - count;
  mida_node = sp->stat.expanded - count + mida_data->mida_cost;

  ratio = ((double) mida_node) / bfs_node;
  u_ratio = ((double) mida_node) / (bfs_node - *(mida_bin + index));

  printf("** q=%g BFS=%d-%d IDA=%d BFS ratio=%g+%g iter=%d\n",
	 q_ratio, bfs_node, *(mida_bin + index), mida_node,
	 ratio, u_ratio - ratio, iter);

  if (iter > 1) {
    mida_data->sqr_exp_factor += (factor * factor);
    mida_data->tot_exp_factor += factor;
    if (factor > mida_data->max_exp_factor) mida_data->max_exp_factor = factor;
    if (factor < mida_data->min_exp_factor) mida_data->min_exp_factor = factor;
    mean = mida_data->tot_exp_factor/(float) (iter-1);
    printf("** Factor of expanded node:  Max=%g Avg=%g Min=%g",
	   mida_data->max_exp_factor, mean, mida_data->min_exp_factor);
    printf(" StdDev=%g\n",
	   sqrt((double) ((mida_data->sqr_exp_factor/(float) (iter-1)) -
			  mean*mean)));
				      

    if (cmd.debug >= 1) {
      printf("*** ");
      temp = sp->child;
      while (temp->brother->brother != NIL) {
	printf(" %g",((float) temp->stat.expanded) /
	       temp->brother->stat.expanded);
	temp = temp->brother;
      }
      printf("\n");
    }
  }

#ifdef DEBUG
  if (cmd.debug >= 3) {
    temp = sp->child;
    do {
      printf("Iteration=%d  Threshold=%g  Node=%d\n",
	     iter,(float) temp->constr.bound,temp->stat.expanded);
      temp = temp->brother;
    } while (--iter > 0);
  }
#endif
}


mida_accounting(sp, child_sp)
search_ *sp, *child_sp;
{
#ifdef DEBUG
  int mida_n = sp->alg.mida->mida_n, *mida_bin = sp->alg.mida->bin;
  float x, d2;
  int k, y;

  if (cmd.debug >= 1) {
    printf("Previous Up-bound=%g, Previous Low-bound=%g, ",
           (float) sp->glub,(float) sp->gllb);
    printf("Up-bound=%g, Low-bound=%g\n",(float) child_sp->glub,
	   (float) child_sp->gllb);

    if (sp->alg.mida->iter > 1) {
      x = (float) child_sp->brother->constr.bound;
      y = child_sp->brother->stat.expanded;
      d2 = ((float) child_sp->constr.bound - x) / mida_n;
      printf("{{%g,%d},",x,y);
      for (k = 1; k <= mida_n; k++) {
	x += d2;
	y += *(mida_bin + k - 1);
        printf("{%g,%d},",x,y);
      }
    }
  }
#endif
}


#ifndef MIDA_BFSNODE
#define MIDA_BFSNODE 100	/** original is 2000 **/
#endif


mida_bfs_search (sp, bfs_sp, msgp)
search_ *sp, *bfs_sp;
search_message_ *msgp;
{
  search_ending_ termination;
  node_ *node;
  int i, index, mida_n = sp->alg.mida->mida_n;
  domain nodelb;

  bfs_sp->strategy = BFS;
  bfs_sp->constr.time = MIDA_BFSNODE;
  sp->alg.mida->bfsbin = (domain *) malloc(MIDA_BFSNODE * sizeof(domain));

  termination = bfs_primitive (bfs_sp, sp, msgp);

  merge_stat_to_parent(bfs_sp, sp);
  merge_approx_to_parent(bfs_sp, sp);
  merge_sol_to_parent(bfs_sp, sp);

  if (termination == SEARCH_IS_ABORTED) {
    /* Put bfs_bin into mida_bin */
    for (i = 0; i < mida_n; i++) *(sp->alg.mida->bin + i) = 0;
    for (i = 0; i < MIDA_BFSNODE; i++) {
      nodelb = *(sp->alg.mida->bfsbin + i);
      if (nodelb == sp->root_lb) index = 0;
      else
	index = ceil((nodelb - sp->root_lb) * mida_n /
		     ((float) (sp->alg.mida->threshold - sp->root_lb))) - 1;

      if ((index < 0) || (index > mida_n))
	printf("MIDA ERROR: index = %d, nodelb = %g\n", index, (float) nodelb);
      else
	(*(sp->alg.mida->bin + index))++;
    }

    for (node = btree_delete(bfs_sp); node; node = btree_delete(bfs_sp))
      free_node(node);
  }

#ifdef DEBUG
  /* Print debug information if necessary */
  if (cmd.debug >= 1) {
    printf ("MIDA'BFS search: ");
    printf ("approx(global_upb=%g)=%g, last_lb=%g, expand=%d\n",
	(float) sp->glub, (float) expected_opt(sp),
	(float) sp->alg.mida->threshold, bfs_sp->stat.expanded);
  }
#endif DEBUG

  free (sp->alg.mida->bfsbin);
  sp->alg.mida->data.opt_value = bfs_sp->constr.bound = sp->alg.mida->threshold;
  sp->alg.mida->iter++;
  bfs_sp->strategy = sp->strategy;
}
