
/*+ --------------------------------------------------- + 
  +		NEW_IDA*_CR Search Algorithms		+
  + --------------------------------------------------- +*/

#define	PRINT_ITER_PROFILE	printf \
("iter=%d, approx=%g, accu=%g, exp_opt(glub=%g)=%g, gllb=%g, threshold=%g, time=%d\n", \
iter, child_sp->approx.approx, approx2accu (child_sp->approx.approx), \
(float) child_sp->glub, (float) expected_opt(child_sp), \
(float) child_sp->gllb, \
(float) child_sp->constr.bound, get_time(child_sp));

void set_new_ida_cr_device ();

void new_ida_cr_algorithm (sp)
  search_ *sp;
{
  search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal; /** not important, only as a placeholder **/
  int iter = 0;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  set_search_message (&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search (sp);

  while (++iter) {
    set_new_ida_cr_device (child_sp, iter);
    put_child_constr (child_sp, sp, 1.0, 1.0, 1.0);

    idpd_init();
    msg.feasible = NO;
    switch (sp->strategy) {
    case DFS:  term_signal = dfs_primitive (child_sp, sp, &msg); break;
    case GDFS: term_signal = gdfs_primitive (child_sp, sp, &msg); break;
    case BAND: term_signal = band_primitive (child_sp, sp, &msg); break;
    case FBS:  term_signal = first_band_primitive (child_sp, sp, &msg); break;
    default:   error ("ida_algorithm: no such feasible primitive"); break;
    }

    eval_final_approx (child_sp);
#ifdef DEBUG
if (cmd.debug >= 1) { PRINT_ITER_PROFILE }
#endif
    merge_stat_to_parent (child_sp, sp);
    merge_sol_to_parent (child_sp, sp);
    if (get_approx (child_sp) <= sp->approx.achieved)
      merge_approx_to_parent (child_sp, sp);
    merge_threshold_to_parent (child_sp, sp);

    if (msg.feasible == YES || get_time (sp) >= sp->constr.time) break;
    if (end_of_opt_sol_search (sp) == YES) break;

    (sp->child = child_sp = create_brother_search (child_sp))->parent = sp;
  }
}


void set_new_ida_cr_device (sp, iter)
  search_ *sp;
  int iter;
{
  int i;
  long time, count;
  domain x, r, base, y, z;
  float x1, y1, z1, log_x1, log_y1, log_z1, rho;
  double log();

  if (iter == 1)
  {
       sp->constr.bound = ROOT->lowb+1;
  } else {
      x = sp->brother->alg.new_ida_cr->x;
      r = sp->brother->alg.new_ida_cr->r;
      base = sp->brother->glub - x;
      time = sp->brother->stat.expanded * (r - 1);

      for (i = 0, count = 0; i < NEW_IDA_CR_NUM_BIN; i++)
      {
	count += sp->brother->alg.new_ida_cr->bin[i];
	if (count >= time) break;
      }

      if (count < time)
      {
	for (i = NEW_IDA_CR_NUM_BIN - 1; i >= 0; i--)
	{
	  if (sp->brother->alg.new_ida_cr->bin[i]) break;
	}

	if (sp->brother->alg.new_ida_cr->mida && iter >= 6)
	{
	    /* x was defined */
	    x1 = sp->brother->stat.expanded;

	    y = x + base * ((float) (++i)) / NEW_IDA_CR_NUM_BIN;
	    y1 = x1 + count;

	    /* to predict z */
	    z1 = x1 + time;

	    log_x1 = (float) log ((double) x1);
	    log_y1 = (float) log ((double) y1);
	    log_z1 = (float) log ((double) z1);

	    rho = (log_y1 - log_x1) / (y - x);

	    z = x + (log_z1 - log_x1) / rho;
            sp->constr.bound = z;
	} else {
            sp->constr.bound = x + base * ((float) (++i)) / NEW_IDA_CR_NUM_BIN;
	}
      } else {
        sp->constr.bound = x + base * ((float) (++i)) / NEW_IDA_CR_NUM_BIN;
      }
#ifdef DEBUG
        if (cmd.debug >= 1)
          printf ("set_ida_cr_device: not sufficient nodes\n");
#endif
  }
#ifdef DEBUG
        if (cmd.debug >= 1)
          printf ("set_ida_cr_device: time=%d, count=%d, bin number=%d\n",
                        time, count, i);
#endif

  sp->alg.new_ida_cr->x = sp->constr.bound;

  Next_IDA_Threshold = problem.huge;
}


void new_ida_cr_bin_update (sp, node)
search_ *sp;
node_ *node;
{
    domain x = sp->alg.new_ida_cr->x;
    domain diff = node->lowb - x;
    domain base = sp->glub - x;
    int which;

    if (diff > 0 && diff <= base)
    {
	which = diff * NEW_IDA_CR_NUM_BIN / base;
	(sp->alg.new_ida_cr->bin[which])++;
#ifdef DEBUG
        if (cmd.debug >= 4)
        {
          printf ("node->lowb=%f, bin=%d\n", ((float) node->lowb), which);
        }
#endif
    }
}

