/**
 ** 		Parallel Processing of Searches
 **/

void para_algorithm (sp, pe, num_pe)
  search_ *sp;
  search_ *pe[];
  int num_pe;
{ search_message_ msg;
  yesno_ break_signal;
  int i;
  search_ *first_sp;
  node_ *rootlet;
  long *ngen_arr = NULL;
  long max_ngen;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  /** start the first bfs **/
  first_sp = create_child_search (sp);
  set_search_message (&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  idpd_init ();
  put_child_constr (first_sp, sp, 1.0, 1.0, 1.0);
  switch (cmd.first)
  {
    case -1:
      first_sp->strategy = BFS;
      first_sp->algorithm = DEFAULT;
      first_sp->open.btree = NULL;
      pfirst_primitive (first_sp, sp, &msg, 1);
      break;
    default:
      first_sp->strategy = BFS;
      first_sp->algorithm = DEFAULT;
      first_sp->open.btree = NULL;
      pfirst_primitive (first_sp, sp, &msg, num_pe);
      break;
  }
  merge_sol_to_parent (first_sp, sp);
  merge_stat_to_parent (first_sp, sp);

  /** create an array of ngen **/
  ngen_arr = (long *) malloc (num_pe * sizeof (long));

  /** create a search process for each pe **/
  for (i = 0; i < num_pe; i++)
  {
    pe[i] = create_child_search(sp);
    pe[i]->strategy = cmd.strategy;
    pe[i]->algorithm = cmd.algorithm;
    pe[i]->open.list = NULL;
    pe[i]->incumbent = sp->incumbent;
    pe[i]->glub = sp->glub;
    pe[i]->gllb = sp->gllb;
    if ((rootlet = btree_delete (first_sp))) insert (rootlet, pe[i]);
    put_child_constr (pe[i], sp, 1.0, 1.0, 1.0);
    pe[i]->approx.approx = sp->approx.approx;
  }

  set_search_message (&msg, RESUME, NULL, YES, cmd.pf.rt, cmd.pf.st);

  break_signal = NO;
  while (para_termination (pe, num_pe) == NO)
  {
    /** start parallel search **/
    for (i = 0; i < num_pe; i++)
      break_signal = para_primitive (pe[i], sp, &msg, sp->gllb, ngen_arr+i);

#ifdef TIME_IS_GEN
    max_ngen = 0;
    for (i = 0; i < num_pe; i++)
      if (max_ngen < *(ngen_arr+i)) max_ngen = *(ngen_arr+i);
    sp->stat.num_op += max_ngen;
#else
    /** it is assumed that the virtual time is the number of nodes expanded **/
    (sp->stat.num_op)++;
#endif

    /** bookkeeping **/
    para_merge_solution (sp, pe, num_pe);

    if (break_signal == YES) break;

    /** load_balancing **/
    para_load_balancing (pe, num_pe);
  }
  para_merge_stat (sp, pe, num_pe);
  eval_final_approx(sp);
}


yesno_ para_primitive (pep, sp, msgp, gllb, ngenp)
  search_ *pep;
  search_ *sp;
  search_message_ *msgp;
  domain gllb;
  long *ngenp;
{ yesno_ break_signal = NO;
  search_ending_ retval;
  long ngen;

  switch (cmd.algorithm) {
  case PBFS:
    if (pbfs_primitive (pep, sp, msgp, gllb, ngenp) == SEARCH_IS_ABORTED)
      break_signal = YES;
    break;
  case PGDFS:
    if (pgdfs_primitive (pep, sp, msgp, gllb, ngenp) == SEARCH_IS_ABORTED)
      break_signal = YES;
    break;
  default:
    error ("para_primitive: no such parallel search primitive");
    break;
  }
  return break_signal;
}
