
/** RTS Heuristics:
 **
 ** approx:
 **	type-I approx (LG: linear-gradient heuristic):
 **		linear-gradient approximation.
 **	type-II approx (GG: geometric-gradient heuristic):
 **	type-III approx (FR: first-order-regression heuristic):
 **	 	 first-order regression approximation.
 **
 ** accuracy:
 **	type-I accuracy (LG: linear-gradient heuristic):
 **		linear-gradient accuracy.
 **	type-II accuracy (FR: first-order-regression heuristic):
 **		first-order regression accuracy.
 **	type-III accuracy (GG: geometric-gradient heuristic):
 **
 ** thresholding:
 **	naive th (LG: linear-gradient heuristic):
 **		linear-gradient thresholding.
 **		 (GG: geometric-gradient heuristic):
 **	static th: (used concurrently with approx/accu)
 **		type-I static th (PR: pruning-ratio heuristic):
 **			cut ratio btwn upper bounds.
 **		type-II static th (MG: moving-grid heuristic):
 **			k-th grid between init and final opt bounds.
 **	predictive th:
 **		type-I predictive th (FR: first-order-regression heuristic):
 **			threshold vs time.
 **		type-II predictive th (LD: lower-bound-distribution heuristic):
 **			lower bound's bins vs time.
 **/

/** RTS Versions:
 **
 ** # RTS[approx(LG)]
 ** urts :param 6 no a1 - - 6.0 4.0
 ** 
 ** # RTS[approx(GG)]
 ** urts :param 6 no a2 - - 20.0 4.0
 ** 
 ** # RTS[approx(FR)]
 ** urts :param 6 yes a3 - - 6.0 4.0
 ** 
 ** # RTS[th(LG)]
 ** urts :param 6 no t n - 12.0 4.0
 ** 
 ** # RTS[th(FR)]
 ** urts :param 6 yes t p 1 12.0 4.0
 ** 
 ** # RTS[th(LD)]  LD -- lower-bound distribution heuristics   (bin)
 ** urts :param 6 yes t p 2 12.0 4.0
 ** 
 ** # RTS[approx(LG),th(PR)]  PR -- pruning-ratio heuristics  (FR-I)
 ** brts :param 5 no s 1 6.0 4.0
 ** 
 ** # RTS[approx(LG),th(MG)]  MG -- moving-grid heuristics
 ** brts :param 5 no s 2 6.0 4.0
 ** 
 ** # RTS[approx(LG,AI),th(PR,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hrts :param 5 yes s 1 6.0 4.0
 ** 
 ** # RTS[approx(LG,AI),th(MG,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hrts :param 5 yes s 2 6.0 4.0
 ** 
 ** # RTS[approx(GG),th(PR)]  PR -- pruning-ratio heuristics  (FR-I)
 ** brts :param 5 no s2 1 20.0 4.0	; s2 's 2 means a_type == TYPE_II
 ** 
 ** # RTS[approx(GG),th(MG)]  MG -- moving-grid heuristics
 ** brts :param 5 no s2 2 20.0 4.0
 ** 
 ** # RTS[approx(GG,AI),th(PR,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hrts :param 5 yes s2 1 20.0 4.0
 ** 
 ** # RTS[approx(GG,AI),th(MG,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hrts :param 5 yes s2 2 20.0 4.0
 **
 ** # RTS[accu(LG)]
 ** uRTS :param 6 no a - - 6.0 4.0
 ** 
 ** # RTS[accu(GG)]
 ** uRTS :param 6 no a - - 20.0 4.0
 ** 
 ** # RTS[accu(FR)]
 ** uRTS :param 6 yes a2 - - 6.0 4.0
 ** 
 ** # RTS[accu(LG),th(PR)]  PR -- pruning-ratio heuristics  (FR-I)
 ** bRTS :param 5 no s 1 6.0 4.0
 ** 
 ** # RTS[accu(LG),th(MG)]  MG -- moving-grid heuristics
 ** bRTS :param 5 no s 2 6.0 4.0
 ** 
 ** # RTS[accu(LG,AI),th(PR,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hRTS :param 5 yes s 1 6.0 4.0
 ** 
 ** # RTS[accu(LG,AI),th(MG,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hRTS :param 5 yes s 2 6.0 4.0
 **
 ** # RTS[accu(GG),th(PR)]  PR -- pruning-ratio heuristics  (FR-I)
 ** bRTS :param 5 no s2 1 20.0 4.0
 ** 
 ** # RTS[accu(GG),th(MG)]  MG -- moving-grid heuristics
 ** bRTS :param 5 no s2 2 20.0 4.0
 ** 
 ** # RTS[accu(GG,AI),th(PR,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hRTS :param 5 yes s2 1 20.0 4.0
 ** 
 ** # RTS[accu(GG,AI),th(MG,LI)]  AI -- All Iterations & LI -- Last Iteration
 ** hRTS :param 5 yes s2 2 20.0 4.0
 **/


#define	PRINT_ITER_PROFILE	printf \
("iter=%d, approx=%g, accu=%g, exp_opt(glub=%g)=%g, gllb=%g, threshold=%g, time=%d\n", \
iter, child_sp->approx.approx, approx2accu (child_sp->approx.approx), \
(float) child_sp->glub, (float) expected_opt(child_sp), \
(float) child_sp->gllb, \
(float) child_sp->constr.bound, get_time(child_sp));


/** internal functional predeclarations **/
void set_urts_device (), set_urts_approx (), set_urts_threshold ();
void set_brts_device (), set_brts_approx (), set_brts_threshold ();
void set_hrts_device (), set_hrts_approx ();
void set_lg_approx (), set_gg_approx (), set_fr_approx (), set_qg_approx ();
void set_hrts_threshold ();
void set_rts_threshold (), set_naive_threshold (), set_static_threshold ();
void set_lg_threshold (), set_gg_threshold (), set_qg_threshold ();
void set_pr_threshold (), set_mg_threshold ();
void set_predictive_threshold (), set_fr_threshold ();
void set_ld_threshold (), reset_th_environ (), reset_th_bin ();
void set_th_entry (), enter_th_bin ();

void set_fe_approx (), set_ff_approx ();
void set_fe_approx_aux (), set_ff_approx_aux ();
void set_fe_threshold (), set_ff_threshold ();
void set_fe_threshold_aux (), set_ff_threshold_aux ();
void tcr_bin_update ();
yesno_ get_tcr_th_vs_time ();

float get_g_factor (), get_r_factor ();
domain get_predicted_glub ();
yesno_ check_span (), get_bin_vs_time (), get_th_vs_time ();
yesno_ get_glub_vs_time (), get_approx_vs_time (), get_glub_vs_approx ();
yesno_ get_accu_vs_time (), is_null_device ();
iter_attr_ get_iteration_attribute (); 
set_th_ *get_set_th ();


/** ------------------------------------------------------------ **/

/* 		uRTS*			*/

void urts_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal;
  float adjust = 1.0;
  float r;
  iter_attr_ attr = OTHER_ITER;
  yesno_ break_signal = NO;
  int iter;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  set_search_message(&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search(sp);
  break_signal = NO;

  r = get_r_factor (sp);
  if (r == 0.0) r = 1.0;

  for (iter = 1; 1; iter++) {
    RTS_IRD_iter = iter;
#ifndef NO_INHERIT
    inherit_sol_from_parent(child_sp, sp);
#endif
    if (child_sp->alg.urts->last_pred == YES)
    {
      attr = get_iteration_attribute (child_sp, iter);
/**
      switch (attr)
      {
        case LAST_ITER:     adjust = 1.0;     break;
        case LAST_2ND_ITER: adjust = 1.0 / r; break;
        case OTHER_ITER:    adjust = 1.0;     break;
        default: error("urts_algorithm: no such iteration attribute"); break;
      }
**/
    }

    set_urts_device(child_sp, iter, attr);
    if (is_null_device (child_sp) == YES) adjust = 1.0;
    put_child_constr(child_sp, sp, adjust, 1.0, 1.0);
    idpd_init();

    switch (sp->strategy)
    {
    case DFS:  term_signal = dfs_primitive(child_sp, sp, &msg);   break;
    case GDFS: term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    case BAND: term_signal = band_primitive(child_sp, sp, &msg);  break;
    default:   term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    }

    eval_final_approx(child_sp);
#ifdef DEBUG
if (cmd.debug >= 1) { PRINT_ITER_PROFILE }
#endif
    merge_stat_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (get_approx(child_sp) <= sp->approx.achieved)
      merge_approx_to_parent(child_sp, sp);
    merge_threshold_to_parent(child_sp, sp);

    switch (sp->alg.urts->alg) {
      case UNARY_APPROX: 
	if (child_sp->approx.approx <= 0.0) break_signal = YES;
	break;
      case UNARY_TH:
	if (msg.feasible == YES) break_signal = YES;
	break;
      default:	error("urts_algorithm: no such pruning device");  break;
    }
    if (break_signal == YES) break;
    if (get_time(sp) >= sp->constr.time) break;

    if (end_of_opt_sol_search (sp) == YES) break;

    msg.feasible = NO;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
  }
}

/** ------------------------------------------------------------ **/

void set_urts_device (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    switch (sp->alg.urts->alg)
    {
    case UNARY_APPROX:	set_urts_approx (sp, iter, attr);	break;
    case UNARY_TH:	set_urts_threshold (sp, iter, attr);	break;
    default:	error("set_urts_device: no such pruning device");  break;
    }
}

/** ------------------------------------------------------------ **/

void set_urts_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{ 
    switch (attr)
    {
    case OTHER_ITER:
	switch (sp->alg.urts->a_type)
	{
	case TYPE_I:   set_lg_approx (sp, iter); break;
	case TYPE_II:  set_gg_approx (sp, iter); break;
	case TYPE_III: set_fr_approx (sp, iter, attr); break;
	case TYPE_IV:  set_qg_approx (sp, iter); break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error ("set_urts_approx: no such type"); break;
	}
	break;

    case LAST_ITER:
    case LAST_2ND_ITER:
	switch (sp->alg.urts->a_type)
	{
	case TYPE_I:   set_fr_approx (sp, iter, attr); break;
	case TYPE_II:  set_fr_approx (sp, iter, attr); break;
	case TYPE_III: set_fr_approx (sp, iter, attr); break;
	case TYPE_IV:  set_fr_approx (sp, iter, attr); break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error ("set_urts_approx: no such type"); break;
	}
	break;

    default:
	error ("set_urts_approx: no such attr");
	break;
    }
}

/** ------------------------------------------------------------ **/

void set_urts_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    urts_ *alg = sp->alg.urts;
    domain ub = sp->parent->root_ub;
    domain lb = sp->parent->root_lb;
    float g = get_g_factor (sp);
 
    if (iter == 1)
    {	/* set init urts threshold */
	sp->constr.bound = alg->th = lb + g * (ub - lb);
    } else {
	/* set intermediate-iteration urts threshold */
	set_rts_threshold (sp, iter, attr);
    }

    reset_th_environ (sp);

}

/** ------------------------------------------------------------ **/

void urts_entry_aux (sp, node)
search_ *sp;
node_ *node;
{
    switch (sp->alg.urts->alg)
    {
    case UNARY_APPROX:
	switch (sp->alg.urts->a_type)
	{
	case TYPE_V:
	case TYPE_VI:
		tcr_bin_update (sp, node);
		break;
	default: break;
	}
	break;

    case UNARY_TH:
	switch ((get_set_th(sp))->alg)
	{
	case PREDICTIVE_TH:
	    switch ((get_set_th(sp))->type)
	    {
	    case TYPE_III:
	    case TYPE_IV:
		tcr_bin_update (sp, node);
		break;
	    default: break;
	    }
	default: break;
	}

    default: break;
    }
}

/** ------------------------------------------------------------ **/

void urts_entry (sp, node)
search_ *sp;
node_ *node;
{
    switch (sp->alg.urts->alg)
    {
    case UNARY_APPROX:
	break;

    case UNARY_TH:
	set_th_entry (sp, node);
	break;

    default:
	error ("urts_entry: no such pruning device");
	break;
    }
}

/** ------------------------------------------------------------ **/

/* 		bRTS*			*/
void brts_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal;
  float adjust = 1.0;
  float r;
  iter_attr_ attr = OTHER_ITER;
  int iter;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  set_search_message(&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search(sp);
 
  r = get_r_factor (sp);
  if (r == 0.0) r = 1.0;

  for (iter = 1; 1; iter++) {
    RTS_IRD_iter = iter;
#ifndef NO_INHERIT
    inherit_sol_from_parent(child_sp, sp);
#endif
    if (child_sp->alg.brts->last_pred == YES)
    {
      attr = get_iteration_attribute(child_sp, iter);
/**
      switch (attr)
      {
        case LAST_ITER:     adjust = 1.0;     break;
        case LAST_2ND_ITER: adjust = 1.0 / r; break;
        case OTHER_ITER:    adjust = 1.0;     break;
        default: error("brts_algorithm: no such iteration attribute"); break;
      }
**/
    }

    set_brts_device(child_sp, iter, attr);
    if (is_null_device (child_sp) == YES) adjust = 1.0;
    put_child_constr(child_sp, sp, adjust, 1.0, 1.0);
    idpd_init();

    switch (sp->strategy)
    {
    case DFS:  term_signal = dfs_primitive(child_sp, sp, &msg);   break;
    case GDFS: term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    case BAND: term_signal = band_primitive(child_sp, sp, &msg);  break;
    default:   term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    }

    eval_final_approx(child_sp);
#ifdef DEBUG
if (cmd.debug >= 1) { PRINT_ITER_PROFILE }
#endif

    merge_stat_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (get_approx(child_sp) <= sp->approx.achieved)
      merge_approx_to_parent(child_sp, sp);
    merge_threshold_to_parent(child_sp, sp);

    if (child_sp->constr.bound == problem.huge)
    {
      if (child_sp->approx.approx <= 0.0) break;
    }
    else if (msg.feasible == YES) break;
    if (get_time(sp) >= sp->constr.time) break;

    if (end_of_opt_sol_search (sp) == YES) break;

    msg.feasible = NO;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
  }
}

/** ------------------------------------------------------------ **/

void set_brts_device (sp, iter, attr)
  search_ *sp;
  int iter;
  iter_attr_ attr;
{
    /* Ideally, approx and threshold should be set iteratively,
     * until both converge.
     * However, it may not converge and here it is emulated by
     * setting approx first and then setting threshold.
     */
    set_brts_approx(sp, iter, attr);
    set_brts_threshold(sp, iter, attr);
}

/** ------------------------------------------------------------ **/

void set_brts_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    switch (attr)
    {
    case OTHER_ITER:
        switch (sp->alg.brts->a_type)
	{
        case TYPE_I:   set_lg_approx (sp, iter);  break;
        case TYPE_II:  set_gg_approx (sp, iter);  break;
        case TYPE_III: set_fr_approx (sp, iter, attr);  break;
        case TYPE_IV:  set_qg_approx (sp, iter);  break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error("set_brts_approx: no such type"); break;
	}
	break;

    case LAST_ITER:
    case LAST_2ND_ITER:
	switch (sp->alg.brts->a_type)
	{
	case TYPE_I:   set_fr_approx (sp, iter, attr); break;
	case TYPE_II:  set_fr_approx (sp, iter, attr); break;
	case TYPE_III: set_fr_approx (sp, iter, attr); break;
	case TYPE_IV:  set_fr_approx (sp, iter, attr); break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error ("set_brts_approx: no such type"); break;
	}
	break;

    default:
	error ("set_brts_approx: no such attr");
	break;
    }
}

/** ------------------------------------------------------------ **/

void set_brts_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    brts_ *alg = sp->alg.brts;
    domain ub = sp->parent->root_ub;
    domain lb = sp->parent->root_lb;
    float g = get_g_factor (sp);
 
    if (iter == 1)
    {	/* set init brts threshold */
	sp->constr.bound = alg->th = lb + g * (ub - lb);
    } else {
	/* set intermediate-iteration urts threshold */
	set_rts_threshold (sp, iter, attr);
    }

    /** if approximation degree is 0, then there is no thresholding **/
    if (sp->approx.approx == 0.0)
    {
	sp->constr.bound = problem.huge;
	put_threshold (sp, sp->constr.bound);
    }

    reset_th_environ (sp);
}

/** ------------------------------------------------------------ **/

void brts_entry_aux (sp, node)
search_ *sp;
node_ *node;
{
    switch (sp->alg.brts->a_type)
    {
    case TYPE_V:
    case TYPE_VI:
	tcr_bin_update (sp, node);
	break;
    default: break;
    }
}

/** ------------------------------------------------------------ **/

/**
void brts_entry (sp, node)
  search_ *sp;
  node_ *node;
{ set_th_entry (sp, node); }
**/

/** ------------------------------------------------------------ **/

/* 		hRTS*			*/

void hrts_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  search_ending_ term_signal;
  float adjust = 1.0;
  float r;
  iter_attr_ attr = OTHER_ITER;
  int iter;

  sp->approx.achieved = sp->approx.root_approx;
  sp->gllb = sp->root_lb;
  sp->glub = sp->root_ub;

  set_search_message(&msg, FRESH_START, NULL, YES, cmd.pf.rt, cmd.pf.st);
  child_sp = create_child_search(sp);

  r = get_r_factor (sp);
  if (r == 0.0) r = 1.0;

  for (iter = 1; 1; iter++)
  {
    RTS_IRD_iter = iter;
#ifndef NO_INHERIT
    inherit_sol_from_parent(child_sp, sp);
#endif
    if (child_sp->alg.hrts->last_pred == YES)
    {
      attr = get_iteration_attribute(child_sp, iter);
/**
      switch (attr)
      {
      case LAST_ITER:     adjust = 1.0;     break;
      case LAST_2ND_ITER: adjust = 1.0 / r; break;
      case OTHER_ITER:    adjust = 1.0;     break;
      default: error("hrts_algorithm: no such iteration attribute"); break;
      }
**/
    }

    set_hrts_device(child_sp, iter, attr);
    if (is_null_device (child_sp) == YES) adjust = 1.0;
    put_child_constr(child_sp, sp, adjust, 1.0, 1.0);
    idpd_init();

    switch (sp->strategy)
    {
    case DFS:  term_signal = dfs_primitive(child_sp, sp, &msg);   break;
    case GDFS: term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    case BAND: term_signal = band_primitive(child_sp, sp, &msg);  break;
    default:   term_signal = gdfs_primitive(child_sp, sp, &msg);  break;
    }

    eval_final_approx(child_sp);
#ifdef DEBUG
if (cmd.debug >= 1) { PRINT_ITER_PROFILE }
#endif
    merge_stat_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (get_approx(child_sp) <= sp->approx.achieved)
      merge_approx_to_parent(child_sp, sp);
    merge_threshold_to_parent(child_sp, sp);
    if (child_sp->constr.bound == problem.huge) {
      if (child_sp->approx.approx <= 0.0) break;
    } else {
      if (msg.feasible == YES) break;
    }
    if (get_time(sp) >= sp->constr.time) break;

    if (end_of_opt_sol_search (sp) == YES) break;

    msg.feasible = NO;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
  }
}

/** ------------------------------------------------------------ **/

void set_hrts_device (sp, iter, attr)
  search_ *sp;
  int iter;
  iter_attr_ attr;
{
  /* Only the search of last iteration needs both approx and threshold,
   * all searches of other iterations needs approx only.
   */
  set_hrts_approx(sp, iter, attr);

  if (attr == LAST_ITER || attr == LAST_2ND_ITER)
  {
    set_hrts_threshold(sp, iter, attr);

    /** if approximation degree is 0, then there is no thresholding **/
    if (sp->approx.approx == 0.0)
    {
      sp->constr.bound = problem.huge;
      put_threshold (sp, sp->constr.bound);
    }
  }
}

/** ------------------------------------------------------------ **/

void set_hrts_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    switch (attr)
    {
    case OTHER_ITER:
        switch (sp->alg.hrts->a_type)
	{
        case TYPE_I:   set_lg_approx (sp, iter);  break;
        case TYPE_II:  set_gg_approx (sp, iter);  break;
        case TYPE_III: set_fr_approx (sp, iter, attr);  break;
        case TYPE_IV:  set_qg_approx (sp, iter);  break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error("set_hrts_approx: no such type"); break;
	}
	break;

    case LAST_ITER:
    case LAST_2ND_ITER:
	switch (sp->alg.hrts->a_type)
	{
	case TYPE_I:   set_fr_approx (sp, iter, attr); break;
	case TYPE_II:  set_fr_approx (sp, iter, attr); break;
	case TYPE_III: set_fr_approx (sp, iter, attr); break;
	case TYPE_IV:  set_fr_approx (sp, iter, attr); break;
	case TYPE_V:   set_fe_approx (sp, iter, attr); break;
	case TYPE_VI:  set_ff_approx (sp, iter, attr); break;
	default: error ("set_hrts_approx: no such type"); break;
	}
	break;

    default:
	error ("set_hrts_approx: no such attr");
	break;
    }
}


/** ------------------------------------------------------------ **/

void set_hrts_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    hrts_ *alg = sp->alg.hrts;
    domain ub = sp->parent->root_ub;
    domain lb = sp->parent->root_lb;
    float g = get_g_factor (sp);
 
    if (iter == 1)
    {	/* set init hrts threshold **/
	sp->constr.bound = alg->th = lb + g * (ub - lb);
    } else {
	set_rts_threshold (sp, iter, attr);
    }

    reset_th_environ (sp);
}

/** ------------------------------------------------------------ **/

void hrts_entry_aux (sp, node)
search_ *sp;
node_ *node;
{
    switch (sp->alg.hrts->a_type)
    {
    case TYPE_V:
    case TYPE_VI:
	tcr_bin_update (sp, node);
	break;
    default: break;
    }
}

/** ------------------------------------------------------------ **/

/**
void hrts_entry (sp, node)
  search_ *sp;
  node_ *node;
{ set_th_entry (sp, node); }
**/

 

/*+ --------------------------------------------------- + 
  + 	  RTS* Approximation Setting Routines		+
  + algorithms:						+
  +   . naive approximation (Type I):			+
  +     linear gradient, like sTCGD.			+
  +   . regression approximation (Type II):		+
  +	regressed gradient, like pTCGD.			+
  +   * last approximation: regression gradient.	+
  + --------------------------------------------------- +*/


/** LG: linear-gradient approximation **/
void set_lg_approx (sp, iter)
  search_ *sp;
  int iter;
{ float sched, margin_scale;
  float x_root = sp->parent->approx.x_root_approx;
  float g = get_g_factor (sp);

  if (cmd.xon == YES) {
    x_root = approx2accu (x_root);
    sched = x_root + (1.0 - x_root) * iter * g;
  } else {
    sched = x_root * (1.0 - iter * g);
  }
  if (cmd.xon == YES) sched = accu2approx (sched);
  sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}

/** ------------------------------------------------------------ **/

/** GG: geometric-gradient approximation **/
void set_gg_approx (sp, iter)
  search_ *sp;
  int iter;
{ float sched, margin_scale;
  float x_root = sp->parent->approx.x_root_approx;
  float g = get_g_factor (sp);
  int factor = 1;

  if (iter > 2) {
    for (--iter; iter > 0; --iter) factor *= 2;
    iter = factor;
  }

  if (cmd.xon == YES) {
    x_root = approx2accu (x_root);
    sched = x_root + (1.0 - x_root) * iter * g;
  } else {
    sched = x_root * (1.0 - iter * g);
  }
  if (cmd.xon == YES) sched = accu2approx (sched);
  sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}

/** ------------------------------------------------------------ **/

/** FE: first-order-extrapolation approximation **/
void set_fe_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    float sched, r;
    long time = sp->constr.time;

    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
	r = ((float) time) / ((float) get_time (sp->brother));
    else
	r = get_r_factor (sp);

    set_fe_approx_aux (sp, iter, r);

    /* special handler for discrete problems */
    if (sp->brother)
    {
	if (sp->constr.bound < sp->brother->constr.bound + MIN_DELTA_TH)
	    sp->constr.bound = sp->brother->constr.bound + MIN_DELTA_TH;
    }

    if (cmd.xon == YES)
    {
	sched = ((float) sp->constr.bound) / ((float) sp->parent->glub);
	sched = accu2approx (sched);
    }
    else
    {
	sched = ((float) sp->parent->glub) / ((float) sp->constr.bound) - 1.0;
    }

    sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}

/** ------------------------------------------------------------ **/

/** FF: first-order-fit approximation **/
void set_ff_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    float sched, r;
    long time = sp->constr.time;

    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
	r = ((float) time) / ((float) get_time (sp->brother));
    else
	r = get_r_factor (sp);

    set_ff_approx_aux (sp, iter, r);

    /* special handler for discrete problems */
    if (sp->brother)
    {
	if (sp->constr.bound < sp->brother->constr.bound + MIN_DELTA_TH)
	    sp->constr.bound = sp->brother->constr.bound + MIN_DELTA_TH;
    }

    if (cmd.xon == YES)
    {
	sched = ((float) sp->constr.bound) / ((float) sp->parent->glub);
	sched = accu2approx (sched);
    } else {
	sched = ((float) sp->parent->glub) / ((float) sp->constr.bound) - 1.0;
    }

    sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}

/** ------------------------------------------------------------ **/

/** FR: first-order-regression approximation **/
void set_fr_approx (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    span_ span = FULL_SPAN;
    float sched, a;
    search_ *brother = sp->brother;
    double beta[2];
    float x_root = sp->parent->approx.x_root_approx;
    float r = get_r_factor (sp);
    float g = get_g_factor (sp);
    double t;

    if (cmd.xon == YES)
    {
	x_root = approx2accu (x_root);
	if (brother)
	    sched = approx2accu (brother->approx.approx) + (1.0 - x_root) * g;
	else
	    sched = x_root + (1.0 - x_root) * iter * g;
    } else {
	if (brother)
	    sched = brother->approx.approx - x_root * g;
	else
	    sched = x_root * (1.0 - iter * g);
    }

    if (cmd.xon == YES)
    {
	if (get_accu_vs_time (sp, beta, span) == YES)
	{
  	    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
		t = (double) (sp->constr.time);
	    else
		t = (double) (r * get_time (sp->brother));

	    sched = beta[0] + beta[1] * log10 (t);
	    a = get_approx (sp->parent);
	    a = approx2accu (a);
	    if (sched <= a) sched = a + (1.0 - x_root) * g;
        }
	sched = accu2approx (sched);

    } else {

	if (get_approx_vs_time(sp, beta, span) == YES)
	{
  	    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
		t = (double) (sp->constr.time);
	    else
		t = (double) (r * get_time (sp->brother));

	    sched = beta[0] + beta[1] * log10 (t);
	    a = get_approx (sp->parent);
	    if (sched >= a) sched = a - x_root * g;
	}
    }

    sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}

/** ------------------------------------------------------------ **/

/** QG: quadratic-gradient approximation **/
void set_qg_approx (sp, iter)
  search_ *sp;
  int iter;
{ float sched, margin_scale;
  float x_root = sp->parent->approx.x_root_approx;
  float g = get_g_factor (sp);

  iter *= iter;

  if (cmd.xon == YES) {
    x_root = approx2accu (x_root);
    sched = x_root + (1.0 - x_root) * iter * g;
  } else {
    sched = x_root * (1.0 - iter * g);
  }
  if (cmd.xon == YES) sched = accu2approx (sched);
  sp->approx.approx = (sched >= 0.0) ? sched : 0.0;
}


/*+ --------------------------------------------------- + 
  + 	    RTS* Threshold Setting Routines		+
  + algorithms:						+
  +   . naive algorithm.				+
  +   . static algorithm: type I & type II		+
  +   . predictive algorithm: type I & type II		+
  + --------------------------------------------------- +*/


void set_rts_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    domain prev_th, this_th;

    switch ((get_set_th(sp))->alg)
    {
    case NAIVE_TH:	set_naive_threshold (sp, iter, attr); 		break;
    case STATIC_TH:	set_static_threshold (sp, iter, attr); 		break;
    case PREDICTIVE_TH: set_predictive_threshold (sp, iter, attr);	break;
    default: error ("set_rts_threshold: no such threshold setting");	break;
    }

    /** check whether backward thresholding happens **/
    if (sp->brother && sp->algorithm == uRTS)
    {
	prev_th = get_threshold (sp->brother);
	if (prev_th != problem.huge)
	{
	    this_th = get_threshold (sp);
	    if (this_th < prev_th + MIN_DELTA_TH)
	    {
		this_th = prev_th + MIN_DELTA_TH;
		sp->constr.bound = this_th;
		put_threshold (sp, this_th);
	    }
	}
    }
}

/** ------------------------------------------------------------ **/

void set_naive_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    switch (attr)
    {
    case OTHER_ITER:
	switch ((get_set_th (sp))->type)
	{
	case TYPE_I:   set_lg_threshold (sp, iter, attr); break;
	case TYPE_II:  set_gg_threshold (sp, iter, attr); break;
	case TYPE_III: set_qg_threshold (sp, iter, attr); break;
	default: error ("set_naive_threshold: no such type");  break;
	}
	break;

    case LAST_ITER:
    case LAST_2ND_ITER:
	switch ((get_set_th (sp))->type)
	{
	case TYPE_I:   set_fr_threshold (sp, iter, attr); break;
	case TYPE_II:  set_fr_threshold (sp, iter, attr); break;
	case TYPE_III: set_fr_threshold (sp, iter, attr); break;
	default: error ("set_naive_threshold: no such type");  break;
	}
	break;

    default:
	error ("set_naive_threshold: no such attr");
	break;
    }
}

/** ------------------------------------------------------------ **/

/** LG: linear-gradient heuristic **/
void set_lg_threshold (sp, iter, attr)
search_ *sp;
int iter;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    float g = get_g_factor (sp);
    long time = sp->constr.time;
    search_ *brother;
    domain th;
    float scale, glub[2];
    int count;

    sp->constr.bound = lb + iter * g * (ub - lb);

#ifdef RTS_DIRECT_JUMP
    /** check whether there is no change in upper bound,
     ** if there is no change, it implies that good suboptimum is found. **/
    if (cmd.nparam >= 7)
    {
	sscanf (*(cmd.param+6), "%f", &scale);
        glub[0] = glub[1] = huge_float;
        count = 0;

        for (brother = sp->brother; brother; brother = brother->brother)
	{
#ifdef RTS_ENHANCE
            if (is_feasible_time (get_time (brother)) && brother->brother)
#else
            if (is_feasible_time (get_time (brother)))
#endif
	    {
		glub[count++] = (float) (brother->glub);
		if (count >= 2) break;
	    }
	}

	if (glub[1] != huge_float)
	{
	    if (glub[1] - glub[0] < glub[0] * scale)
	    {
		/** if the upper bounds changes only a little bit,
		** then it is likely that upb is very close to optimum. **/
		sp->constr.bound = glub[0];
	    }
	}
    }
#endif

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** GG: geometric-gradient heuristic **/
void set_gg_threshold (sp, iter, attr)
search_ *sp;
int iter;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    float g = get_g_factor (sp);
    long time = sp->constr.time;
    search_ *brother;
    domain th;
    float scale, glub[2];
    int count;
    int factor = 1;

    /** calc the geometric factor **/
    if (iter > 2)
    {
	for (--iter; iter > 0; --iter) factor *= 2;
	iter = factor;
    }

    /** if everything is too tough, simply use it **/
    sp->constr.bound = lb + iter * g * (ub - lb);

#ifdef RTS_DIRECT_JUMP
    /** check whether there is no change in upper bound,
     ** if there is no change, it implies that good suboptimum is found. **/
    if (cmd.nparam >= 7)
    {
	sscanf (*(cmd.param+6), "%f", &scale);
	glub[0] = glub[1] = huge_float;
	count = 0;

	for (brother = sp->brother; brother; brother = brother->brother)
	{
#ifdef RTS_ENHANCE
	    if (is_feasible_time (get_time (brother)) && brother->brother)
#else
	    if (is_feasible_time (get_time (brother)))
#endif
	    {
		glub[count++] = (float) (brother->glub);
		if (count >= 2) break;
	    }
	}

	if (glub[1] != huge_float)
	{
	    if (glub[1] - glub[0] < glub[0] * scale)
	    {
		/** if the upper bounds changes only a little bit,
		 ** then it is likely that upb is very close to optimum. **/
		sp->constr.bound = glub[0];
	    }
	}
    }
#endif

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** QG: quadratic-gradient heuristic **/
void set_qg_threshold (sp, iter, attr)
search_ *sp;
int iter;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    float g = get_g_factor (sp);
    long time = sp->constr.time;
    search_ *brother;
    domain th;
    float scale, glub[2];
    int count;

    iter *= iter;

    /** if everything is too tough, simply use it **/
    sp->constr.bound = lb + iter * g * (ub - lb);

#ifdef RTS_DIRECT_JUMP
    /** check whether there is no change in upper bound,
     ** if there is no change, it implies that good suboptimum is found. **/
    if (cmd.nparam >= 7)
    {
	sscanf (*(cmd.param+6), "%f", &scale);
	glub[0] = glub[1] = huge_float;
	count = 0;

	for (brother = sp->brother; brother; brother = brother->brother)
	{
#ifdef RTS_ENHANCE
	    if (is_feasible_time (get_time (brother)) && brother->brother)
#else
	    if (is_feasible_time (get_time (brother)))
#endif
	    {
		glub[count++] = (float) (brother->glub);
		if (count >= 2) break;
	    }
	}

	if (glub[1] != huge_float)
	{
	    if (glub[1] - glub[0] < glub[0] * scale)
	    {
		/** if the upper bounds changes only a little bit,
		 ** then it is likely that upb is very close to optimum. **/
		sp->constr.bound = glub[0];
	    }
	}
    }
#endif

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** FE: first-order-extrapolation heuristic **/
void set_fe_threshold (sp, iter, attr)
search_ *sp;
int iter;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    long time = sp->constr.time;
    search_ *brother;
    domain th;
    float r, scale, glub[2];
    int count;

    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
	r = ((float) time) / ((float) get_time (sp->brother));
    else
	r = get_r_factor (sp);

    set_fe_threshold_aux (sp, iter, r);

#ifdef RTS_DIRECT_JUMP
    /** check whether there is no change in upper bound,
     ** if there is no change, it implies that good suboptimum is found. **/
    if (cmd.nparam >= 7)
    {
	sscanf (*(cmd.param+6), "%f", &scale);
	glub[0] = glub[1] = huge_float;
	count = 0;

	for (brother = sp->brother; brother; brother = brother->brother)
	{
#ifdef RTS_ENHANCE
	    if (is_feasible_time (get_time (brother)) && brother->brother)
#else
	    if (is_feasible_time (get_time (brother)))
#endif
	    {
		glub[count++] = (float) (brother->glub);
		if (count >= 2) break;
	    }
	}

	if (glub[1] != huge_float)
	{
	    if (glub[1] - glub[0] < glub[0] * scale)
	    {
		/** if the upper bounds changes only a little bit,
		 ** then it is likely that upb is very close to optimum. **/
		sp->constr.bound = glub[0];
	    }
	}
    }
#endif

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** FF: first-order-fit heuristic **/
void set_ff_threshold (sp, iter, attr)
search_ *sp;
int iter;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    long time = sp->constr.time;
    search_ *brother;
    domain th;
    float r, scale, glub[2];
    int count;

    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
	r = ((float) time) / ((float) get_time (sp->brother));
    else
	r = get_r_factor (sp);

    set_ff_threshold_aux (sp, iter, r);

#ifdef RTS_DIRECT_JUMP
    /** check whether there is no change in upper bound,
     ** if there is no change, it implies that good suboptimum is found. **/
    if (cmd.nparam >= 7)
    {
	sscanf (*(cmd.param+6), "%f", &scale);
	glub[0] = glub[1] = huge_float;
	count = 0;

	for (brother = sp->brother; brother; brother = brother->brother)
	{
#ifdef RTS_ENHANCE
	    if (is_feasible_time (get_time (brother)) && brother->brother)
#else
	    if (is_feasible_time (get_time (brother)))
#endif
	    {
		glub[count++] = (float) (brother->glub);
		if (count >= 2) break;
	    }
	}

	if (glub[1] != huge_float)
	{
	    if (glub[1] - glub[0] < glub[0] * scale)
	    {
		/** if the upper bounds changes only a little bit,
		 ** then it is likely that upb is very close to optimum. **/
		sp->constr.bound = glub[0];
	    }
	}
    }
#endif

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** used concurrently with approx/accu **/
void set_static_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    if (iter > 2)
    {
	if (is_feasible_time (get_time (sp->brother)))
	{

	    switch ((get_set_th (sp))->type)
	    {
	    case TYPE_I:   set_pr_threshold (sp, iter, attr);  break;
	    case TYPE_II:  set_mg_threshold (sp, iter, attr); break;
	    default: 	   error ("set_static_threshold: no such type");  break;
	    }
	}
    }
}

/** ------------------------------------------------------------ **/

/** PR: pruning-ratio heuristics; used concurrently with approx/accu **/
void set_pr_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    float this = (float) get_predicted_glub (sp, iter, attr);
    float prev = (float) sp->brother->glub;
    float r = get_r_factor (sp);
    float shed;
    double num, den;

    /** in case, everything is too tough **/
    sp->constr.bound = ((float) sp->parent->glub) / (1.0 + sp->approx.approx);
    put_threshold (sp, sp->constr.bound);

    /** infeasible prediction of glub **/
    if (sp->brother->constr.bound != problem.huge)
	if (this < sp->brother->constr.bound)
	    return;

    /** main body **/
    num = log10 ((double) (1.0 - cmd.cut_ratio + cmd.cut_ratio * r));
    den = log10 ((double) r);

    shed = (prev + (this - prev) * num / den) / (1.0 + sp->approx.approx);

    if (sp->constr.bound > shed) sp->constr.bound = shed;
    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** MG: moving-grid heuristics; used concurrently with approx/accu **/
void set_mg_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    search_ *brother = sp->brother;
    domain prev_th = get_threshold (sp->brother);
    domain prev_glub = sp->brother->glub;
    float g = get_g_factor (sp);
    float up, low, delta;

    /** main body **/
    low = tofloat (prev_glub) / (1.0 + brother->approx.approx);
    up = tofloat (prev_glub) / (1.0 + sp->approx.approx);
    delta = up - low;

    if (prev_th < low) sp->constr.bound = low + iter * g * delta;
    else sp->constr.bound = prev_th + g * delta;

    if (sp->constr.bound > up) sp->constr.bound = up;
    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

void set_predictive_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    switch ((get_set_th(sp))->type)
    {
    case TYPE_I:   set_fr_threshold (sp, iter, attr); break;
    case TYPE_II:  set_ld_threshold (sp, iter, attr); break;
    case TYPE_III: set_fe_threshold (sp, iter, attr); break;
    case TYPE_IV:  set_ff_threshold (sp, iter, attr); break;
    default: 	   error ("set_predictive_threshold: no such type"); break;
    }
}

/** ------------------------------------------------------------ **/

/** FR: first-order-regression heuristics **/
/** th vs time **/
void set_fr_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    float g = get_g_factor (sp);
    long this_time, time;
    double beta[2];

    /** if everything is too tough, simply use it **/
    sp->constr.bound = lb + iter * g * (ub - lb);

    /** if it is the last one, then use prediction **/
    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
    {
	if (get_th_vs_time (sp, beta, PARTIAL_SPAN) == YES)
	{
	    time = sp->constr.time;
	    sp->constr.bound = beta[0] + beta[1] * log10 ((double) time);
	    if (sp->constr.bound <= sp->brother->constr.bound)
		sp->constr.bound = sp->brother->constr.bound + g * (ub - lb);
	}
    }

    else if (get_th_vs_time (sp, beta, FULL_SPAN) == YES)
    {
	this_time = get_r_factor (sp) * get_time (sp->brother);
	sp->constr.bound = beta[0] + beta[1] * log10 ((double) this_time);
	if (sp->constr.bound <= sp->brother->constr.bound)
	    sp->constr.bound = sp->brother->constr.bound + g * (ub - lb);
    }

    put_threshold (sp, sp->constr.bound);
}

/** ------------------------------------------------------------ **/

/** LD: lower-bound distribution heuristics **/
/** lowb's bin vs time **/
void set_ld_threshold (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    domain lb = sp->parent->root_lb;
    domain ub = sp->parent->root_ub;
    long this_time, time;
    double beta[2];
    float g = get_g_factor (sp);

    /** if everything is too tough, then simply use it **/
    sp->constr.bound = lb + iter * g * (ub - lb);

    /** if it is the last one, then use prediction **/
    if (attr == LAST_ITER || attr == LAST_2ND_ITER)
    {
	if (get_bin_vs_time (sp, beta, PARTIAL_SPAN) == YES)
	{
	    time = sp->constr.time;
	    sp->constr.bound = beta[0] + beta[1] * log10 ((double) time);
	    if (sp->constr.bound <= sp->brother->constr.bound)
		sp->constr.bound = sp->brother->constr.bound + g * (ub - lb);
	}
    }

    else if (get_bin_vs_time (sp, beta, FULL_SPAN) == YES)
    {
	this_time = get_r_factor (sp) * get_time (sp->brother);
	sp->constr.bound = beta[0] + beta[1] * log10 ((double) this_time);
	if (sp->constr.bound <= sp->brother->constr.bound)
	    sp->constr.bound = sp->brother->constr.bound + g * (ub - lb);
    }

    put_threshold (sp, sp->constr.bound);
}



/*+ --------------------------------------------------- + 
  +		RTS*	Library Routines		+
  + --------------------------------------------------- +*/ 


set_th_ *get_set_th (sp)
search_ *sp;
{
    switch (sp->algorithm)
    {
    case uRTS:	return sp->alg.urts->set_th;	break;
    case bRTS:	return sp->alg.brts->set_th;	break;
    case hRTS:	return sp->alg.hrts->set_th;	break;
    case uIRD:	return sp->alg.uird->set_th;	break;
    case bIRD:	return sp->alg.bird->set_th;	break;
    default:	error ("get_set_th: no such algorithm"); break;
    }

    return NULL;
}

/** ------------------------------------------------------------ **/

domain get_predicted_glub (sp, iter, attr)
search_ *sp;
int iter;
iter_attr_ attr;
{
    span_ span;
    float step, this_time;
    double beta[2];
    domain prev, prev_prev;

    span = (attr == LAST_ITER || attr == LAST_2ND_ITER) ? PARTIAL_SPAN :
							  FULL_SPAN;
 
    if (get_glub_vs_time (sp, beta, span) == YES)
    {
	this_time = get_r_factor(sp) * get_time(sp->brother);
	return ((domain) (beta[0] + beta[1] * log10 ((double) this_time)));
    }

    /* if situation is too tough to handle, simply use naive glub */
    if (iter == 1)
    {
	step = get_g_factor(sp) * (sp->parent->root_ub - sp->parent->root_lb);
	this_time = sp->parent->root_ub - step;

    } else {
	prev = sp->brother->glub;
	if (sp->brother->brother) prev_prev = sp->brother->brother->glub;
	else prev_prev = sp->parent->root_ub;
	this_time = prev - (prev_prev - prev);
    }

    return ((domain) this_time);
}

/** ------------------------------------------------------------ **/

yesno_ get_bin_vs_time (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother = sp->brother;
  domain th, exp_opt, *calibrate;
  set_th_ *set_th;
  long *bin, num;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0, x, y, det;
  double sum;
  int nbin, n, i, last_bin;
  
  if (brother == NULL) return NO;

#ifdef RTS_ENHANCE
  if (is_feasible_time (get_time (brother)) && brother->brother)
#else
  if (is_feasible_time (get_time (brother)))
#endif
  {
    th = get_threshold (brother);
    exp_opt = (domain) (brother->glub / (1.0 + get_approx(brother)));
    th = min (th, exp_opt);

    set_th = get_set_th (brother);
    nbin = set_th->nbin;
    bin = set_th->bin;
    calibrate = set_th->calibrate;

    if (cmd.ignore_last_bin == YES)
    {
	/* locate the last bin */
	for (i = nbin - 1; i >= 0; i--)
	{
	    if (*(bin+i) > 0)
	    {
		last_bin = i;
		break;
	    }
	}
    }

    /* first-order regression over bins, x is #nodes, y is lowb val */
    sum = 0.0;
    for (n = i = 0; i < nbin && *(calibrate+i) <= th; i++)
    {
	if (cmd.ignore_last_bin == YES)
	{
	    if (i == last_bin) break;
	}

	num = *(bin+i);
	if (num < MinValidBin) continue;
	n++;
	x = log10 ((double) num);
	y = (double) *(calibrate+i);
	sum += x;
	x_sum += sum;
	y_sum += y;
	x2_sum += (sum * sum);
	xy_sum += (sum * y);
    }

#ifdef DEBUG
    if (cmd.debug >= 1)
      printf ("get_bin_vs_time: n=%d\n", ((int) n));
#endif
    if (n <= 1) return NO;
    det = todouble (n) * x2_sum - x_sum * x_sum;
    if (tofloat (det) == 0.0) return NO;
    beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
    beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

    if (tofloat (beta[1]) == 0.0) return NO;
    return YES;
  }

  return NO;
}

/** ------------------------------------------------------------ **/

yesno_ get_th_vs_time (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0, x, y, det;
  long time;
  int n = 0;

  /* skip infeasible regression */
  if (check_span (sp, span) == NO) return NO;

  /* first-order regression over bins, x is #nodes, y is lowb val */
  for (brother = sp->brother; brother; brother = brother->brother) {
    time = get_time (brother);
#ifdef RTS_ENHANCE
    if (is_feasible_time (time) && brother->brother) {
#else
    if (is_feasible_time (time)) {
#endif
      n++;
      x = log10 ((double) time);
      y = (double) get_threshold (brother);
      x_sum += x;
      y_sum += y;
      x2_sum += (x * x);
      xy_sum += (x * y);
    } else break;
  }

  det = todouble (n) * x2_sum - x_sum * x_sum;
  if (tofloat (det) == 0.0) return NO;
  beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
  beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

  if (tofloat (beta[1]) == 0.0) return NO;
  return YES;
}

/** ------------------------------------------------------------ **/

yesno_ get_glub_vs_time (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0, x, y, det;
  long time;
  int n = 0;

  /* skip infeasible regression */
  if (check_span (sp, span) == NO) return NO;

  /* first-order regression over bins, x is #nodes, y is lowb val */
  for (brother = sp->brother; brother; brother = brother->brother) {
    time = get_time (brother);  
#ifdef RTS_ENHANCE
    if (is_feasible_time (time) && brother->brother) {
#else
    if (is_feasible_time (time)) {
#endif
      n++;
      x = log10 ((double) time);
      y = (double) brother->glub;
      x_sum += x;
      y_sum += y;
      x2_sum += (x * x);
      xy_sum += (x * y);
    } else break;
  }

  det = todouble (n) * x2_sum - x_sum * x_sum;
  if (tofloat (det) == 0.0) return NO;
  beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
  beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

  if (tofloat (beta[1]) == 0.0) return NO;
  return YES;
}

/** ------------------------------------------------------------ **/

yesno_ get_approx_vs_time (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0;
  double x, y, det;
  long time;
  int n = 0;

  /* skip infeasible regression */
  if (check_span (sp, span) == NO) return NO;

  /* first-order regression over bins, x is #nodes, y is lowb val */
  for (brother = sp->brother; brother; brother = brother->brother) {
    time = get_time (brother);
#ifdef RTS_ENHANCE
    if (is_feasible_time (time) && brother->brother) {
#else
    if (is_feasible_time (time)) {
#endif
      n++;
      x = log10 ((double) time);
      y = (double) get_approx (brother);
      x_sum += x;
      y_sum += y;
      x2_sum += (x * x);
      xy_sum += (x * y);
    } else break;
  }

  det = todouble (n) * x2_sum - x_sum * x_sum;
  if (tofloat (det) == 0.0) return NO;
  beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
  beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

  if (tofloat (beta[1]) == 0.0) return NO;
  return YES;
}

/** ------------------------------------------------------------ **/

yesno_ get_accu_vs_time (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0;
  double x, y, det;
  long time;
  int n = 0;
  float a;

  /* skip infeasible regression */
  if (check_span (sp, span) == NO) return NO;

  /* first-order regression over bins, x is #nodes, y is lowb val */
  for (brother = sp->brother; brother; brother = brother->brother) {
    time = get_time (brother);
#ifdef RTS_ENHANCE
    if (is_feasible_time (time) && brother->brother) {
#else
    if (is_feasible_time (time)) {
#endif
      n++;
      a = get_approx (brother);
      a = approx2accu (a);
      x = log10 ((double) time);
      y = todouble (a);
      x_sum += x;
      y_sum += y;
      x2_sum += (x * x);
      xy_sum += (x * y);
    } else break;
  }

  det = todouble (n) * x2_sum - x_sum * x_sum;
  if (tofloat (det) == 0.0) return NO;
  beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
  beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

  if (tofloat (beta[1]) == 0.0) return NO;
  return YES;
}

/** ------------------------------------------------------------ **/

yesno_ get_glub_vs_approx (sp, beta, span)
  search_ *sp;
  double beta[];
  span_ span;
{ search_ *brother;
  double x_sum = 0.0, y_sum = 0.0, x2_sum = 0.0, xy_sum = 0.0, x, y, det;
  int n = 0;

  /* skip infeasible regression */
  if (check_span (sp, span) == NO) return NO;

  /* first-order regression over bins, x is #nodes, y is lowb val */
  for (brother = sp->brother; brother; brother = brother->brother)
#ifdef RTS_ENHANCE
    if (is_feasible_time (get_time (brother)) && brother->brother) {
#else
    if (is_feasible_time (get_time (brother))) {
#endif
      n++;
      x = (double) brother->approx.approx;
      y = (double) brother->glub;
      x_sum += x;
      y_sum += y;
      x2_sum += (x * x);
      xy_sum += (x * y);
    } else break;

  det = todouble (n) * x2_sum - x_sum * x_sum;
  if (tofloat (det) == 0.0) return NO;
  beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
  beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

  if (tofloat (beta[1]) == 0.0) return NO;
  return YES;
}

/** ------------------------------------------------------------ **/

yesno_ check_span (sp, span)
  search_ *sp;
  span_ span;
{ search_ *brother;
  int count = 0;

  for (brother = sp->brother; brother; brother = brother->brother)
#ifdef RTS_ENHANCE
    if (is_feasible_time (get_time (brother)) && brother->brother) count++;
#else
    if (is_feasible_time (get_time (brother))) count++;
#endif
    else break;

  if (span == FULL_SPAN) return (count >= 2 ? YES : NO);
  else return (count >= 3 ? YES : NO);
}

/** ------------------------------------------------------------ **/

void reset_th_environ (sp)
search_ *sp;
{
    set_th_ *set_th = get_set_th (sp);

    /* reset environ for predicting threshold */
    if (set_th->alg == PREDICTIVE_TH && set_th->type == TYPE_II)
	reset_th_bin (sp);
}

/** ------------------------------------------------------------ **/

void reset_th_bin (sp)
search_ *sp;
{
    set_th_ *set_th = get_set_th (sp);
    int nbin = set_th->nbin, i;
    long *bin = set_th->bin;
    domain *calibrate = set_th->calibrate;
    domain abs_lb = sp->parent->root_lb;
    domain abs_ub = sp->parent->root_ub;
    float step;

    /* reset bin and set the calibration of bins */
    step = (float) (abs_ub - abs_lb);
    step /= tofloat (nbin);
    for (i = 0; i < nbin; i++)
    {
	*(bin+i) = 0;
	*(calibrate+i) = abs_lb + (domain) (step * i);
    }
}

/** ------------------------------------------------------------ **/

void enter_th_bin (sp, lowb)
  search_ *sp;
  domain lowb;
{ set_th_ *set_th = get_set_th(sp);
  int nbin = set_th->nbin, i;
  domain *calibrate = set_th->calibrate;

  for (i = 0; i < nbin && *(calibrate+i) < lowb; i++) ;
  if (i == nbin) i--;
  ++(*(set_th->bin+i));
}

/** ------------------------------------------------------------ **/

void set_th_entry (sp, node)
search_ *sp;
node_ *node;
{
    set_th_ *set_th = get_set_th (sp);

    if (set_th->alg == PREDICTIVE_TH && set_th->type == TYPE_II)
	enter_th_bin (sp, node->lowb);
}

/** ------------------------------------------------------------ **/

float get_g_factor (sp)
  search_ *sp;
{ float g;
  switch (sp->algorithm) {
    case uRTS:  g = sp->alg.urts->g_factor;	break;
    case bRTS:  g = sp->alg.brts->g_factor;	break;
    case hRTS:  g = sp->alg.hrts->g_factor;	break;
    case uIRD:  g = sp->alg.uird->g_factor;	break;
    case bIRD:  g = sp->alg.bird->g_factor;	break;
    default:  error("get_g_factor: no such algorithm"); break;
  }

#ifdef GRADIENT_STEP
  if (g == 0.0 || g == huge_float) g = 0.1;
  else g = 1.0 / g;
#else
  if (g == 0.0 || g == huge_float) g = 0.1;
#endif
  return g;
}

/** ------------------------------------------------------------ **/

float get_r_factor (sp)
search_ *sp;
{
    search_ *p;
    float r_sum = 0.0, r;
    int count;

    switch (sp->algorithm)
    {
    case uRTS:  r =  sp->alg.urts->r_factor;	break;
    case bRTS:  r =  sp->alg.brts->r_factor;	break;
    case hRTS:  r =  sp->alg.hrts->r_factor;	break;
    case uIRD:  r =  sp->alg.uird->r_factor;	break;
    case bIRD:  r =  sp->alg.bird->r_factor;	break;
    default:  error("get_r_factor: no such algorithm"); break;
    }

    if (r == huge_float)
    {	/* growth ratio is not specified */
	if ((p = sp->brother))
	{
      	    for (count = 0; p->brother; p = p->brother, count++)
		r_sum += (((float) get_time (p)) /
				((float) get_time (p->brother)));
	}

	r = count ? r_sum / ((float) count) : 2;
    }

    return r;
}

/** ------------------------------------------------------------ **/

iter_attr_ get_iteration_attribute (sp, iter)
search_ *sp;
int iter;
{
    long prev_time, this_time, curr_time, Tconstr;
    float r;

    if (iter > 1)
    {
	r = get_r_factor (sp);
	prev_time = get_time (sp->brother);
	this_time = r * prev_time;
	curr_time = get_time (sp->parent);
	Tconstr = sp->parent->constr.time;

	if (curr_time + this_time >= Tconstr) 		return LAST_ITER;

	if (curr_time + (1+r) * this_time >= Tconstr)	return LAST_2ND_ITER;
    }

    return OTHER_ITER;
}

