
/* Time-Constrained A* search algorithms.
 * TCA=<sTCA,pTCA,dTCA> 
 */

/* sTCA: stca_algorithm(search_ *sp) */
void stca_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  int n = 0;

  sp->approx.achieved = sp->approx.root_approx;
  child_sp = create_child_search(sp);
  put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
  set_search_message(&msg, FRESH_START, NULL, NO, cmd.pf.rt, cmd.pf.st);
  while (1) {
    child_sp->approx.approx = get_stca_next_approx(sp, ++n);
    if (cmd.xon == YES)
      child_sp->approx.accu = approx2accu (child_sp->approx.approx);
    if (bfs_primitive(child_sp, sp, &msg) == SEARCH_IS_ABORTED) break;
    eval_final_approx(child_sp);
    merge_stat_to_parent(child_sp, sp);
    merge_approx_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (child_sp->approx.approx <= 0.0) break;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
    put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
    inherit_sol_from_parent(child_sp, sp);

    if (end_of_opt_sol_search (sp) == YES) break;
  }
  merge_stat_to_parent(child_sp, sp);
  merge_sol_to_parent(child_sp, sp);
  if (get_approx(child_sp) < sp->approx.achieved) {
    merge_approx_to_parent(child_sp, sp);
  }
}


float get_stca_next_approx (sp, n)
/* a = a0 * (1 - n * g); a = (a < 0) ? 0 : a; */
  search_ *sp;
  int n;
{ float new_a;

  if (cmd.xon == YES) {
    /* accuracy-driven */
    new_a = sp->approx.root_accu *
				(1.0 - tofloat (n) * sp->alg.stca->g_factor);
    new_a = accu2approx (new_a);
  } else {
    /* approx-driven */
    new_a = sp->approx.root_approx *
				(1.0 - tofloat (n) * sp->alg.stca->g_factor);
  }
  if (new_a < 0.0) new_a = 0.0;
  return new_a; 
}


/* pTCA: ptca_algorithm(search_ *sp) */
void ptca_algorithm (sp)
  search_ *sp; 
{ constr_ save_constr;
  search_message_ msg;
  ptca_ *ptcap = sp->alg.ptca;
  float s;

  sp->approx.achieved = sp->approx.root_approx;
  s = ptcap->s_factor;
  init_regress_struct(&(ptcap->regress));
  set_search_message(&msg, FRESH_START, &(ptcap->regress), NO,
			cmd.pf.rt, cmd.pf.st);
  save_constr = sp->constr;
  /* nTCA & profiling parts */
  set_stop_constr(sp, s, s, s);
  bfs_primitive(sp, NULL, &msg);
  if (end_of_opt_sol_search (sp) == YES) {
    eval_final_approx(sp);
    return;
  }
  /* prediction part */
  alg_pf_predict(sp, &(ptcap->regress));
  sp->approx.approx = sp->approx.predicted * ptcap->c_factor;
  /* solution part */
  sp->constr = save_constr;
  msg.style = RESUME;
  msg.rp = NULL;
  bfs_primitive(sp, NULL, &msg);
  eval_final_approx(sp);
}


/* dTCA: dtca_algorithm(search_ *sp) & dtca_entry(search_ *sp) */
void dtca_algorithm (sp)
  search_ *sp; 
{ constr_ save_constr;
  search_message_ msg;
  dtca_ *dtcap = sp->alg.dtca;
  float s;

  sp->approx.achieved = sp->approx.root_approx;
  s = dtcap->s_factor;
  init_regress_struct(&(dtcap->regress));
  save_constr = sp->constr;
  set_stop_constr(sp, s, s, s);
  set_search_message(&msg, FRESH_START, &(dtcap->regress), NO,
		     cmd.pf.rt, cmd.pf.st);
  bfs_primitive(sp, NULL, &msg);
  if (end_of_opt_sol_search (sp) == YES) {
    eval_final_approx(sp);
    return;
  }

  alg_pf_predict(sp, &(dtcap->regress));
  sp->approx.approx = sp->approx.predicted * dtcap->c_factor;
  sp->constr = save_constr;
  msg.style = RESUME; 
  msg.alg_entry = YES;
  bfs_primitive(sp, NULL, &msg);
  eval_final_approx(sp);
}


void dtca_entry(sp)
  search_ *sp;
{ float temp;

  alg_pf_predict(sp, &(sp->alg.dtca->regress));
  temp = sp->approx.predicted * sp->alg.dtca->c_factor;
  /* approx can only go up and can not go down */
  if (sp->approx.approx < temp) sp->approx.approx = temp;
}

