
/* Time-Constrained GDFS algorithms.
 * TCGD=<sTCGD,pTCGD>
 */

/* sTCGD: stcgd_algorithm(search_ *sp) */
void stcgd_algorithm (sp)
  search_ *sp;
{ search_ *child_sp;
  search_message_ msg;
  int n = 0;

  sp->approx.achieved = sp->approx.root_approx;
  child_sp = create_child_search(sp);
  put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
  set_search_message(&msg, FRESH_START, NULL, NO, cmd.pf.rt, cmd.pf.st);
  while (1) {
    child_sp->approx.approx = get_stcgd_next_approx(sp, ++n);
    if (cmd.xon == YES)
      child_sp->approx.accu = approx2accu (child_sp->approx.approx);
    if (gdfs_primitive(child_sp, sp, &msg) == SEARCH_IS_ABORTED) break;
    eval_final_approx(child_sp);
    merge_stat_to_parent(child_sp, sp);
    merge_approx_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (child_sp->approx.approx <= 0.0) break;
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
    put_child_constr(child_sp, sp, 1.0, 1.0, 1.0);
    inherit_sol_from_parent(child_sp, sp);

    if (end_of_opt_sol_search (sp) == YES) break;
  }
  merge_stat_to_parent(child_sp, sp);
  merge_sol_to_parent(child_sp, sp);
  if (get_approx(child_sp) < sp->approx.achieved) {
    merge_approx_to_parent(child_sp, sp);
  }
}


float get_stcgd_next_approx (sp, n)
/* a = a0 * (1 - n * g); a = (a < 0) ? 0 : a; */
  search_ *sp;
  int n;
{ float new_a;

  if (cmd.xon == YES) {
    /* accuracy-driven */
    new_a = sp->approx.root_accu *
                                (1.0 - tofloat (n) * sp->alg.stcgd->g_factor);
    new_a = accu2approx (new_a);
  } else {
    /* approx-driven */
    new_a = sp->approx.root_approx *
                                (1.0 - tofloat (n) * sp->alg.stcgd->g_factor);
  }
  if (new_a < 0.0) new_a = 0.0;
  return new_a;
}


/* pTCGD: ptcgd_algorithm(search_ *sp) */
void ptcgd_algorithm (sp)
  search_ *sp; 
{ search_ *child_sp;
  constr_ save_constr;
  search_message_ msg;
  ptcgd_ *ptcgdp = sp->alg.ptcgd;
  float s;
  int n = 0;

  sp->approx.achieved = sp->approx.root_approx;
  s = ptcgdp->s_factor;
  init_regress_struct(&(ptcgdp->regress));
  set_search_message(&msg, FRESH_START, NULL, NO, cmd.pf.rt, cmd.pf.st);
  /* create search struct for sTCGD */
  child_sp = create_child_search(sp);
  put_child_constr(child_sp, sp, s);
  /* sTCGD & profiling part */
  while (1) {
    child_sp->approx.approx = get_stcgd_next_approx(sp, ++n);
    if (cmd.xon == YES)
      child_sp->approx.accu = approx2accu (child_sp->approx.approx);
    if (gdfs_primitive(child_sp, sp, &msg) == SEARCH_IS_ABORTED) break;
    eval_final_approx(child_sp);
    merge_stat_to_parent(child_sp, sp);
    merge_approx_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
    if (child_sp->approx.approx <= 0.0) break;
    alg_pf(child_sp, &(sp->alg.ptcgd->regress));
    save_constr = child_sp->constr;   
    (sp->child = child_sp = create_brother_search(child_sp))->parent = sp;
    child_sp->constr = save_constr;
    inherit_sol_from_parent(child_sp, sp);

    if (end_of_opt_sol_search (sp) == YES) break;
  }
  merge_stat_to_parent(child_sp, sp);
  if (get_approx(child_sp) < sp->approx.achieved) {
    merge_approx_to_parent(child_sp, sp);
    merge_sol_to_parent(child_sp, sp);
  }
  /* prediction part */
  alg_pf_predict(sp, &(ptcgdp->regress));
  sp->approx.approx = sp->approx.predicted * ptcgdp->c_factor;
  /* solution part */
  gdfs_primitive(sp, NULL, &msg);
  eval_final_approx(sp);
}

