/** core routines for "true control of re-expansion" **/

yesno_ get_tcr_th_vs_time ();
double log10 ();

/** ------------------------------------------------------------ **/

void set_fe_threshold_aux (sp, iter, r)
search_ *sp;
int iter;
float r;
{
    int i;
    long time, count;
    domain x, base, y, z;
    float x1, y1, z1, log_x1, log_y1, log_z1, rho;

    if (iter == 1)
    {
	sp->constr.bound = ROOT->lowb+1;

    } else {

	x = sp->brother->alg.urts->tcr->x;
	base = sp->brother->glub - x;
	time = get_time (sp->brother) * (r - 1);

	for (i = 0, count = 0; i < TCR_NUM_BIN; i++)
	{
	    count += sp->brother->alg.urts->tcr->bin[i];
	    if (count >= time) break;
	}

	if (count < time)
	{
	    for (i = TCR_NUM_BIN - 1; i >= 0; i--)
	    {
		if (sp->brother->alg.urts->tcr->bin[i]) break;
	    }

	    /** ? value; ?1 time **/
	    x1 = get_time (sp->brother);
	    if (iter >= 3 && is_feasible_time (x1))
	    {
		y = x + base * ((float) (++i)) / TCR_NUM_BIN;
		y1 = x1 + count;

		/* to predict z */
		z1 = x1 + time;

		log_x1 = (float) log10 ((double) x1);
		log_y1 = (float) log10 ((double) y1);
		log_z1 = (float) log10 ((double) z1);

		rho = (log_y1 - log_x1) / (y - x);

		z = x + (log_z1 - log_x1) / rho;
		sp->constr.bound = z;

	    } else {

		sp->constr.bound = x + base * ((float) (++i)) / TCR_NUM_BIN;
	    }

	} else {

            sp->constr.bound = x + base * ((float) (++i)) / TCR_NUM_BIN;
	}

#ifdef DEBUG
	if (cmd.debug >= 1)
	    printf ("set_fe_threshold: not sufficient nodes\n");
#endif
    }
#ifdef DEBUG
    if (cmd.debug >= 1)
	printf ("set_fe_threshold: time=%d, count=%d, bin number=%d\n",
								time, count, i);
#endif

    sp->alg.urts->tcr->x = sp->constr.bound;

    Next_IDA_Threshold = problem.huge;
}

/** ------------------------------------------------------------ **/

void set_ff_threshold_aux (sp, iter, r)
search_ *sp;
int iter;
float r;
{
    span_ span = FULL_SPAN;
    int i;
    long time, count;
    domain x, base, y, z;
    float x1, y1, z1, log_x1, log_y1, log_z1, rho;
    double beta[2];

    if (iter == 1)
    {
       sp->constr.bound = ROOT->lowb+1;

    } else {

	x = sp->brother->alg.urts->tcr->x;
	base = sp->brother->glub - x;
	time = get_time (sp->brother) * (r - 1);

	for (i = 0, count = 0; i < TCR_NUM_BIN; i++)
	{
	    count += sp->brother->alg.urts->tcr->bin[i];
	    if (count >= time) break;
	}

	if (count < time)
	{
	    for (i = TCR_NUM_BIN - 1; i >= 0; i--)
	    {
		if (sp->brother->alg.urts->tcr->bin[i]) break;
	    }

	    /** ? value; ?1 time **/
	    x1 = get_time (sp->brother);
	    if (iter >= 3 && is_feasible_time (x1))
	    {
		y = x + base * ((float) (++i)) / TCR_NUM_BIN;
		y1 = x1 + count;
		z1 = x1 + time;

		if (get_tcr_th_vs_time (sp, beta, span, y1, y) == YES)
		{
	    	    sp->constr.bound = beta[0] + beta[1] * log10 ((double) z1);
		    if (sp->constr.bound <= x) sp->constr.bound = y;

		} else {

			/** using extrapolation **/

			log_x1 = (float) log10 ((double) x1);
			log_y1 = (float) log10 ((double) y1);
			log_z1 = (float) log10 ((double) z1);
			rho = (log_y1 - log_x1) / (y - x);

			z = x + (log_z1 - log_x1) / rho;
			sp->constr.bound = z;
		}

	    } else {

            sp->constr.bound = x + base * ((float) (++i)) / TCR_NUM_BIN;
	    }

	} else {

	    sp->constr.bound = x + base * ((float) (++i)) / TCR_NUM_BIN;
	}

#ifdef DEBUG
	if (cmd.debug >= 1)
	    printf ("set_fe_threshold: not sufficient nodes\n");
#endif
    }
#ifdef DEBUG
    if (cmd.debug >= 1)
	printf ("set_fe_threshold: time=%d, count=%d, bin number=%d\n",
								time, count, i);
#endif

    sp->alg.urts->tcr->x = sp->constr.bound;

    Next_IDA_Threshold = problem.huge;
}

/** ------------------------------------------------------------ **/

void set_fe_approx_aux (sp, iter, r)
search_ *sp;
int iter;
float r;
{
    set_fe_threshold_aux (sp, iter, r);
}

/** ------------------------------------------------------------ **/

void set_ff_approx_aux (sp, iter, r)
search_ *sp;
int iter;
float r;
{
    set_ff_threshold_aux (sp, iter, r);
}

/** ------------------------------------------------------------ **/

void tcr_bin_update (sp, node)
search_ *sp;
node_ *node;
{
    domain x = sp->alg.urts->tcr->x;
    domain diff = node->lowb - x;
    domain base = sp->glub - x;
    int which;

    if (diff > 0 && diff <= base)
    {
	which = diff * TCR_NUM_BIN / base;
	(sp->alg.urts->tcr->bin[which])++;
#ifdef DEBUG
        if (cmd.debug >= 4)
        {
          printf ("node->lowb=%f, bin=%d\n", ((float) node->lowb), which);
        }
#endif
    }
}

/** ------------------------------------------------------------ **/

yesno_ get_tcr_th_vs_time (sp, beta, span, curr_time, curr_th)
search_ *sp;
double beta[];
span_ span;
long curr_time;
domain curr_th;
{
    search_ *brother;
    int n = 1;
    double x = log10 ((double) curr_time);
    double y = (double) curr_th;
    double x_sum = x;
    double y_sum = y;
    double x2_sum = x * x;
    double xy_sum = x * y;
    double det;
    long time;

    /* skip infeasible regression */
    if (check_span (sp, span) == NO) return NO;

    /* first-order regression over bins, x is #nodes, y is lowb val */
    for (brother = sp->brother; brother; brother = brother->brother)
    {
	time = get_time (brother);
#ifdef RTS_ENHANCE
	if (is_feasible_time (time) && brother->brother)
#else
	if (is_feasible_time (time))
#endif
	{
	    n++;
	    x = log10 ((double) time);
	    y = (double) get_threshold (brother);
	    x_sum += x;
	    y_sum += y;
	    x2_sum += (x * x);
	    xy_sum += (x * y);
	} else break;
    }

    det = todouble (n) * x2_sum - x_sum * x_sum;
    if (tofloat (det) == 0.0) return NO;
    beta[0] = (x2_sum * y_sum - x_sum * xy_sum) / det;
    beta[1] = (- x_sum * y_sum + todouble (n) * xy_sum) / det;

    if (tofloat (beta[1]) == 0.0) return NO;
    return YES;
}

