
#ifndef	__wise_define_h_
#define __wise_define_h_ 

#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	<math.h>
#include	<sys/types.h>
#include	<sys/times.h>


#ifndef RANDOM_RADIX
#define	RANDOM_RADIX		((float) (0x7fffffff))
#endif

#ifndef LPG_LARGE_CONST
#define LPG_LARGE_CONST		1000
#endif


/* const definitions */
#define	NIL			('\0')
#define	DONT_CARE		(- 100)


#define IDA_CR_NUM_BIN          1000
#define NEW_IDA_CR_NUM_BIN      100
#define TCR_NUM_BIN      	100


/* predicting & profiling */
#define	PROFILING_STEP		5
#define	REGRESSION_ORDER	1


/* problem-dependent definitions, if they are forgotten to be defined */
#ifndef MaxProblemSize
#define MaxProblemSize		200
#endif

#ifndef AllocationSize
#define AllocationSize		256
#endif

#ifndef MaxNumberOfProblems
#define	MaxNumberOfProblems	100
#endif

/**
#ifndef NumberOfBins
#define	NumberOfBins		50
#endif
**/
#ifndef NumberOfBins
#define	NumberOfBins		20
#endif

#ifndef MinFeasibleTime
#define MinFeasibleTime		(problem.size)
#endif

#ifndef FirstMaxDepth
#define FirstMaxDepth		256
#endif


#define is_feasible_time(t)	(t >= MinFeasibleTime)


/* algorithm specific definition */

#ifndef MinValidBin
#define MinValidBin		4
#endif


/* type conversion shorthand */
#define	toshort(a)	((short) (a))
#define	toint(a)	((int) (a))
#define	tolong(a)	((long) (a))
#define	tofloat(a)	((float) (a))
#define	todouble(a)	((double) (a))

#define	min(a,b)	((a > b) ? b : a)
#define	max(a,b)	((a > b) ? a : b)
#define in(a,b,c)	((a >= b) && (a <= c))

#define	streql(a,b)	(! strcmp(a,b))

#define accu2approx(a)	((a == 0.0) ? huge_float : 1.0 / a - 1.0)
#define approx2accu(e)	((e == huge_float) ? 0.0 : 1.0 / (e + 1.0))


/* what the search for */
enum _task_ { LEARNING = 0, CSP = 1 };
typedef enum _task_ 		task_;


/* search style */
enum _search_style_ { RESUME = 0, FRESH_START = 1 };
typedef enum _search_style_	search_style_;


/* question answer */
enum _yesno_ { NO = 0, YES = 1 };
typedef enum _yesno_		yesno_;


enum _boolean_ { FALSE = 0, TRUE = 1};
typedef enum _boolean_		boolean_;


/** threshold is defined absolutely or relatively **/
typedef enum { ABSOLUTE = 0, RELATIVE = 1 } metric;


/* search ending info */
enum _search_ending_ {
	SEARCH_IS_ABORTED = 0,
	SEARCH_IS_COMPLETED = 1,
	SEARCH_IS_IDLE = 2
};
typedef enum _search_ending_	search_ending_;


/* domain data types */
enum _datatype_ { INT = 1, LONG = 2, FLOAT = 3, DOUBLE = 4};
typedef enum _datatype_		datatype_;


/* search strategies */
enum _strategy_ { GBB, BFS, DFS, GDFS, BAND, FBS };
typedef enum _strategy_		strategy_;


/* search algorithms */
enum _algorithm_ {	DEFAULT,
			LW,
			sTCA, pTCA, dTCA,
			sTCGD, pTCGD,
			uRTS, bRTS, hRTS,
			uIRD, bIRD,
			IRA, IDA, DFS_star,
			MIDA, IDA_CR, NEW_IDA_CR,
			iFBS,
			gDFS_iFBS,
			PBFS, PDFS, PGDFS, PGBB, PBAND };
typedef enum _algorithm_	algorithm_;


/* definitions for counting system */
enum _count_ { REAL, VIRTUAL };
typedef enum _count_		count_;


/* profiling signals  */
enum _psignal_ { RUN_TIME = -10, SPACE_VS_TIME = -20 };
typedef enum _psignal_		psignal_;


/* which child to be expanded */
enum _child_ { ALL_CHILDREN = -1, NEXT_CHILD, NEXT_N_CHILDREN };
typedef enum _child_		child_;


/* enumerative type */
enum _enum_type_ { TYPE_NULL = 0, TYPE_I = 1, TYPE_II = 2, TYPE_III = 3,
		   TYPE_IV = 4, TYPE_V = 5, TYPE_VI = 6, TYPE_VII = 7,
		   TYPE_VIII = 8, TYPE_IX = 9, TYPE_X = 10 };
typedef enum _enum_type_	enum_type_;


/* search prob. info */
struct _problem_ {
	int	domain;
	int	size; 		/* problem size */
	int	seed;		/* random seed */
	domain	opt_value;	/* the optimal value */
	domain	huge;		/* huge number in the domain */
	task_	task;		/* task: search or learning */
};
typedef struct _problem_	problem_;


struct _constr_ {
	long	time;		/* time constraint */
	long	space;		/* space constraint */
	float	cst;		/* cst constraint */
	domain	bound;		/* bound of sol val, useful in ida threshold */
};
typedef struct _constr_		constr_;


struct _profile_ {
	yesno_	rt;
	yesno_	st;
	float	last_approx;
	float	last_accu;
	float	step;		/* run-time profile step */
};
typedef struct _profile_	profile_;


struct _factor_ {
	float	g;		/* stca, stcgd */
	float	s;		/* ptca, dtca, ptcgd, dtcgd */
	float	c;		/* ptca, dtca, ptcgd, dtcgd */
};
typedef struct _factor_		factor_;


typedef enum {	FIXED_BW,
		LINEAR_BW,
		EXP_BW,
		DYNAMIC_BW,	/** change bw dynamically **/
		GDFS_dBS_BW,	/** GDFS+dBS **/
		GDFS_fBS_BW	/** GDFS+fBS **/
				} bw_fx_;


struct _cmd_ {
	int	num_pe;		/* num of processing elements */
	int	comm_idle;	/* comm overhead in terms of num_op */
	int	first;		/* style of starting search */
	int	iter;		/* num of sample runs */
	int	debug;		/* debugging mode */
	float	cut_ratio;	/* cut-line ratio */
	yesno_	ens;		/** essential-node search **/
	yesno_	first_sol;	/** first-solution search **/
	yesno_	opt_sol;	/** optimal-solution search **/
	yesno_	xon;		/* accuracy is on */
	yesno_	report;		/* special report, eg ida */
	yesno_	ignore_last_bin;/* ignore the last bin */
	count_	count;		/* counting system */
	strategy_ strategy;	/* search strategy */
	algorithm_ algorithm;	/* algorithm */
	int	nbin;		/* num of bins */
	float	approx;		/* hard approx */
	float	accu;		/* hard accuracy */
	constr_	constr;		/* hard constr */
	metric	bound_metric;	/** relative or absolute **/
	metric	approx_metric;	/** relative or absolute **/
	float	rel_a;		/** relative approx degree, 0--1 **/
	domain	bound_base;	/** relative base, usually opt sol val **/
	profile_ pf;		/* profile command */
	factor_	factor;		/* parameters for algorithm */
	bw_fx_	bw_fx;		/** bandwidth function name **/
	domain	opt_sol_val;	/** opt-sol-val for optimal-solution search **/
	float	sol_precision;	/** solution precision allowable **/
	int	nparam;		/* num of parameters */
	char	**param;	/* parameter list */
	int	nauxparam;	/* num of aux parameters */
	char	**auxparam;	/* aux parameter list for strategy information */
};
typedef struct _cmd_		cmd_;


struct _io_ {
	char	*stat;		/* stat file for learning */
	char	*summary;	/* summary of statistics */
	char	*graph;		/* graph ready format */
	char	*rt_pf;		/* run-time profile */
	char	*st_pf;		/* space-vs-time profile */
	char	*report;	/* special report, eg ida */
};
typedef struct _io_		io_;


struct _stat_ {
	long	num_op;		/* # operations in parallel search */
	long	generated;	/* # nodes generated */ 	
	long	expanded;	/* # nodes expanded */ 	
	long	feasible;	/* # nodes feasible */
	long	infeasible;	/* # nodes infeasible */
	long	active;		/* # nodes active */
	long	bounded;	/* # nodes bounded */
	long	dominated;	/* # nodes dominated */
	long	bounding;	/* # nodes bounded by new node */
	long	dominating;	/* # nodes dominated by new node */
	long	hard_bounded;	/* # nodes hard bounded, for ida */
	/* pruned == bounded + dominated + bounding + dominating */
	long	time;		/* real timer */
	long	max_active;	/* max space usage */
	float	v_cst; 		/* virtual cst */
	float	r_cst;		/* real cst */
	long	unix_utime;	/* user time provided by unix */
	long	unix_stime;	/* system time provided by unix */
};
typedef struct _stat_		stat_;


struct _approx_ {
	float	approx;
	float	accu;
	float	run_time;
	float	predicted;
	float	root_approx;
	float	root_accu;
	float	x_root_approx;	/* external root's approx degree */
	float	achieved;	/* approx achieved already */
};
typedef struct _approx_		approx_;


/*
	NODE STRUCTURE node_
	a node structure node_ consists of three parts:
	(a) PI (problem independent):
	    permanent and predefined in B&B shell.
	(b) PDSI (problem independent but size dependent):
	    user-defined in 'node.h'
	(c) PDSD (problem and size dependent):
	    user-defined in 'node.h' with an array of sizes for all regions.
	    all PDSD regions are accessed via pointers in PDSI.
*/

enum _nodetype_ { INACTIVE, ACTIVE_SINGLE, ACTIVE_COMPOUND };
typedef enum _nodetype_		nodetype_;


#define	PDSI_	struct	PDSI__
#define	PDSD_	struct	PDSD__


struct _node_ {
	struct _node_	*parent;
	struct _node_	*next;		/* used in the search tree */
	struct _node_	*brother;	/* used in children list */

	struct _node_	*left;		/* used in b_plus tree */
	struct _node_	*right;		/* used in b_plus tree */
	struct _node_	*up;		/* used in b_plus tree */

	int	entity;		/* entity id, ie, city id currently visited */
	int	depth;		/* depth of this node */
	int	nsprout;	/* number of live children */
	int	ndecomp;	/* number of docomposed */
	nodetype_ type;		/* type of search node,
				   non-zero active, 0 inactive;
				   specifically, 2: compound; 1: ordinary */
	domain  g_cost;		/* g_cost is the current node's actual cost */
	domain  lowb, upb;	/* node's lower bound and upper bound */

	PDSI_PART		/* PDSI part which is declared in 'node.h' */
};
typedef struct _node_		node_;


struct _node_conf_ {
	/* all sizes are in terms of integer */
	int	pdsd_size;	/* size of prob-dep, size-dep part */
	int	node_size;	/* size of a entire node */
	/* node_size == (pdsd_size + sizeof(node_)) */
};
typedef struct _node_conf_	node_conf_;


struct _regress_ {
	double	x_sum;	/* sum of x's */
	double	x2_sum;	/* sum of x squares */
	double	y_sum;	/* sum of y's */
	double	xy_sum;	/* sum of xy products */
	double	n; 	/* number of points */
	double	beta[REGRESSION_ORDER + 1];	/* regression coeff's */
};
typedef struct _regress_	regress_;


struct _stca_ {
	float	g_factor;
};
typedef struct _stca_		stca_;


struct _stcgd_ {
	float	g_factor;
};
typedef struct _stcgd_		stcgd_;


struct _ptca_ {
	regress_	regress;
	float	s_factor;
	float	c_factor;
};
typedef struct _ptca_		ptca_;


struct _ptcgd_ {
	regress_	regress;
	float	g_factor;
	float	s_factor;
	float	c_factor;
};
typedef struct _ptcgd_		ptcgd_;


struct _dtca_ {
	regress_	regress;
	float	s_factor;
	float	c_factor;
};
typedef struct _dtca_		dtca_;


/** begin, Arthur's, 4/10/92 **/
typedef struct _mida_stat_ {
	int	b_cost;
	int	mida_cost;
	domain	opt_value;
	float	max_exp_factor;
	float	sqr_exp_factor;
	float	tot_exp_factor;
	float	min_exp_factor;
} mida_stat_;

typedef struct _mida_ {
	float	g_factor;
	int	mida_n;			/* Number of histogram bin */
	int	*bin;			/* Histogram bin */
#ifdef UPAST
	int	*oldbin;		/* Another histogram bin */
#endif
	domain	*bfsbin;		/* Bin for initial BFS */
	int	iter;			/* # of iteration so far */
	domain	threshold;		/* Current threshold value */
	mida_stat_	data;		/* Data for ida search */
} mida_;
/** end, Arthur's, 4/10/92 **/


typedef struct _ida_cr_ {
	float	r;			/* growth rate */
	domain	x;			/* starting bin value */
	int	bin[IDA_CR_NUM_BIN];
} ida_cr_;


typedef struct _new_ida_cr_ {
	float	r;			/* growth rate */
	domain	x;			/* starting bin value */
	int	mida;			/* 1: allow prediction, 0 no. */
	int	bin[NEW_IDA_CR_NUM_BIN];
} new_ida_cr_;


enum _unary_alg_ { UNARY_APPROX = 1, UNARY_TH = 2 };
typedef enum _unary_alg_	unary_alg_;


enum _th_alg_ { NAIVE_TH = 1, STATIC_TH = 2, PREDICTIVE_TH = 3 };
typedef enum _th_alg_		th_alg_;


enum _iter_attr_ { LAST_ITER = 1, LAST_2ND_ITER = 2, OTHER_ITER = 3 };
typedef enum _iter_attr_	iter_attr_;


enum _span_ { FULL_SPAN = 1, PARTIAL_SPAN = 2 };
typedef enum _span_		span_;


struct _set_th_ {
	th_alg_		alg;
	enum_type_	type;
	int		nbin;
	long		*bin;
	domain		*calibrate;
};
typedef struct _set_th_		set_th_;


typedef struct _tcr_ {			/* true control of reexpansion */
	domain	x;			/* starting bin value */
	int	bin[TCR_NUM_BIN];
} tcr_;


struct _urts_ {
	yesno_		last_pred;
	domain		th;
	enum_type_	a_type;		/* approx type */
	float		g_factor;	/* step factor */
	float		r_factor;	/* growth rate */
	unary_alg_	alg;
	set_th_		*set_th;
	tcr_		*tcr;
};
typedef struct _urts_		urts_;


struct _brts_ {
	yesno_		last_pred;
	domain		th;
	enum_type_	a_type;		/* approx type */
	float		g_factor;	/* step factor */
	float		r_factor;	/* growth rate */
	set_th_		*set_th;
	tcr_		*tcr;
};
typedef struct _brts_		brts_;


struct _hrts_ {
	yesno_		last_pred;
	domain		th;
	enum_type_	a_type;		/* approx type */
	float		g_factor;	/* step factor */
	float		r_factor;	/* growth rate */
	set_th_		*set_th;
	tcr_		*tcr;
};
typedef struct _hrts_		hrts_;


struct _ird_basic_ {
	yesno_		shortcut;	/* jump if upper not change too much */
	float		base;		/* starting approx or threshold */
	float		margin;		/* shortcut margin */
};
typedef struct _ird_basic_	ird_basic_;


struct _uird_ {
	ird_basic_	basic;
	domain		th;
	enum_type_	a_type;		/* approx type */
	set_th_		*set_th;
	float		g_factor;	/* step factor */
	float		r_factor;	/* growth rate */
	unary_alg_	alg;
};
typedef struct _uird_		uird_;


struct _bird_ {
	ird_basic_	basic;
	domain		th;
	enum_type_	a_type;		/* approx type */
	set_th_		*set_th;
	float		g_factor;	/* step factor */
	float		r_factor;	/* growth rate */
};
typedef struct _bird_		bird_;


struct _search_message_ {
	search_style_	style;
	regress_	*rp;
	yesno_		alg_entry;
	yesno_		rt_pf;
	yesno_		st_pf;
	yesno_		feasible;
};
typedef struct _search_message_	search_message_;


typedef struct {
	node_		*young;		/** list of youngs of this depth **/
	int		count;		/** num of adults of this depth **/
	int		width;		/** width of this depth **/
} tree_;


typedef struct {
	int		ceiling_depth;	/** lower bound of depth **/
	int		floor_depth;	/** upper bound of depth **/
	int		init_width;	/** init_width of band **/
	int		bw_bound;	/** max bw allowed in dBS **/
	float		growth_rate;	/** growth rate in first-band search **/
	tree_		*tree;		/** band search's array of youngs **/
} band_;


struct _search_
{
	strategy_	strategy;
	algorithm_	algorithm;
	domain		glub;		/* global least upper bound */
	domain		gllb;		/* global least lower bound */
	domain		root_lb;	/* root's lower bound */
	domain		root_ub;	/* root's upper bound */
 	domain		x_root_lb;	/* external lower bound */
 	domain		x_root_ub;	/* external upper bound */
	stat_		stat;		/* statistics */
	constr_		constr;		/* constraints */
	approx_		approx;		/* approx */
	solution_	*incumbent;	/* incumbent */
	band_		band;		/** band descriptor in band search **/
	domain		next_ida_th;	/** for IDA* only **/
	union active_list_
	{
	  node_		*bptree;	/* bfs b_plus tree */
	  node_		*btree;		/* bfs virtual b+ tree, & band search */
          node_		*list;		/* open list, for gdfs */
	  node_		*stack;		/* activation stack, for dfs */
	} open;
        union alg_struct_
	{
	  void		*def;		/* no associated structure is needed */
	  void		*lw;
	  stca_		*stca;
	  ptca_		*ptca;
	  dtca_		*dtca;
	  stcgd_	*stcgd;
	  ptcgd_	*ptcgd;
	  urts_		*urts;
	  brts_		*brts;
	  hrts_		*hrts;
	  uird_		*uird;
	  bird_		*bird;
	  mida_		*mida;
	  ida_cr_	*ida_cr;
	  new_ida_cr_	*new_ida_cr;
	} alg;
	struct _search_	*child;		/* any sub-search */
	struct _search_	*brother;	/* search of a level */
	struct _search_	*parent;	/* parent search */
	int		idle_clk;	/* idle timer for parallel processing */
};
typedef struct _search_		search_;


#endif __wise_define_h_

