
problem_	problem;
cmd_		cmd;
io_		io;
node_		*pool_manager = NULL, *savep = NULL, *ROOT = NULL;
node_conf_	node_conf;
struct tms	start_time, last_time, break_time;

int		SIZE = 0;	/** for HARE **/
int		huge_int = HUGE_INT;
long		huge_long = HUGE_LONG;
float		huge_float = HUGE_FLOAT;
double		huge_double = HUGE_DOUBLE;

long		MaxDepth = FirstMaxDepth;
long		NumNodesTaken = 0;
int		RTS_IRD_iter = 0;
domain		Next_IDA_Threshold = (domain) 0;
yesno_		No_Upper_Bound = NO;

yesno_		_is_first_sol_found_ = NO;

stat_		keep_stat[MaxNumberOfProblems];


void init_problem_struct (p)
  problem_ *p;
{
  switch (p->domain = Domain) {
    case INT:    p->huge = HUGE_INT;    break;
    case LONG:   p->huge = HUGE_LONG;   break;
    case FLOAT:  p->huge = HUGE_FLOAT;  break;
    case DOUBLE: p->huge = HUGE_DOUBLE; break;
    default:     error("no such generic data type\n"); break;
  }
  p->opt_value = p->huge;
  p->seed = 1;
  p->task = CSP;
}


void init_cmd_struct (p)
  cmd_ *p;
{
  p->num_pe = 1;
  p->comm_idle = 0;
  p->first = 0;
  p->iter = 1;
  p->debug = 0;
  p->cut_ratio = 0.5;
  p->ens = NO;
  p->first_sol = NO;
  p->opt_sol = NO;
  p->opt_sol_val = problem.huge;
  p->sol_precision = 0.005;
  p->xon = NO;
  p->report = NO;
  p->ignore_last_bin = NO;
  p->count = VIRTUAL;
  p->strategy = BFS;
  p->algorithm = DEFAULT;
  p->nbin = NumberOfBins;
  p->approx = 0.0;
  p->accu = 1.0;
  p->constr.time = HUGE_LONG;
  p->constr.space = HUGE_LONG;
  p->constr.cst = HUGE_FLOAT;
  p->constr.bound = problem.huge;
  p->bound_metric = p->approx_metric = ABSOLUTE;
  p->rel_a = 0.0;
  p->bound_base = (domain) problem.huge;
  init_profile_struct(&(p->pf));
  p->bw_fx = FIXED_BW;
  p->nparam = 0;
  p->param = NULL;
  p->nauxparam = 0;
  p->auxparam = NULL;
}


void init_factor_struct (p)
  factor_ *p;
{
  p->g = 0.25;
  p->s = 0.15;
  p->c = 0.8;
}


void init_io_struct (iop)
  io_ *iop;
{
  iop->stat = "stat";		/* old out */
  iop->summary = "summary";	/* old constr.out */
  iop->graph = "graph-ready";
  iop->rt_pf = "run-time";
  iop->st_pf = "space-time";
  iop->report = "report";
}


void init_stat_struct (p)
  stat_ *p;
{
  p->num_op = 0;
  p->generated = 0;
  p->expanded = 0;
  p->feasible = 0;
  p->infeasible = 0;
  p->active = 0;
  p->bounded = 0;
  p->dominated = 0;
  p->bounding = 0;
  p->dominating = 0;
  p->hard_bounded = 0;
  p->max_active = 0;
  p->time = 0;
  p->v_cst = 0.0;
  p->r_cst = 0.0;
  p->unix_utime = 0;
  p->unix_stime = 0;
}


void init_constr_struct (p)
  constr_ *p;
{
  p->time = cmd.constr.time;
  p->space = cmd.constr.space;
  p->cst = cmd.constr.cst;
  p->bound = cmd.constr.bound;
}


void init_profile_struct (p)
  profile_ *p;
{
  p->rt = NO;
  p->st = NO;
  p->step = 0.01;
  p->last_approx = HUGE_FLOAT;
  p->last_accu = 0.0;
}


void init_approx_struct (p)
  approx_ *p;
{
  p->approx = cmd.approx;
  p->accu = cmd.accu;
  p->predicted = 0.0;
  p->run_time= HUGE_FLOAT;
  p->root_approx = HUGE_FLOAT;
  p->x_root_approx = HUGE_FLOAT;
  p->achieved = HUGE_FLOAT;
}


void init_fbs_struct (sp)
    search_ *sp;
{   band_ *bp = &(sp->band);
    int i, num;

    cmd.bw_fx = FIXED_BW;
    bp->ceiling_depth = 0;
    bp->floor_depth = 0;
    bp->init_width = get_int_cmd_line_param (HUGE_INT, 0, 0);
    bp->growth_rate = get_float_cmd_line_param (2.0, 1, 0);

    MaxDepth = FirstMaxDepth;
    num = MaxDepth + 1;
    sp->band.tree = (tree_ *) malloc (num * sizeof (tree_));
    for (i = 0; i < num; i++)
	init_tree_struct (sp->band.tree, i, sp->band.init_width);
}


void init_band_struct (sp)
    search_ *sp;
{   band_ *bp = &(sp->band);
    int i, num;

    switch (sp->algorithm)
    {

    case uRTS:
    case bRTS:
    case hRTS:

      bp->ceiling_depth = 0;
      bp->floor_depth = 0;
      bp->init_width = get_int_cmd_line_auxparam (HUGE_INT, 0, 0);
      bp->bw_bound = get_int_cmd_line_auxparam (HUGE_INT, 2, 0);
      bp->growth_rate = 1.0;  /** not used in band search **/

      if (cmd.nauxparam <= 1) cmd.bw_fx = FIXED_BW;
      else {
	switch (**(cmd.auxparam+1))
        {

	case 'l':
	case 'L': cmd.bw_fx = LINEAR_BW;	break;

	case 'e':
	case 'E': cmd.bw_fx = EXP_BW;		break;

	case 'd':
	case 'D': cmd.bw_fx = DYNAMIC_BW;	break;

	case 'a': /** adaptive **/
	case 'A': cmd.bw_fx = GDFS_dBS_BW;	break;

	case 'h': /** hybrid **/
	case 'H': cmd.bw_fx = GDFS_fBS_BW;	break;

	case 'f':
	case 'F':
	case '-':
	default:  cmd.bw_fx = FIXED_BW;		break;
	}
    }

    break;

    default:
      bp->ceiling_depth = 0;
      bp->floor_depth = 0;
      bp->init_width = get_int_cmd_line_param (HUGE_INT, 0, 0);
      bp->bw_bound = get_int_cmd_line_param (HUGE_INT, 2, 0);
      bp->growth_rate = 1.0;  /** not used in band search **/

      if (cmd.nparam <= 1) cmd.bw_fx = FIXED_BW;
      else {
	switch (**(cmd.param+1))
        {

	case 'l':
	case 'L': cmd.bw_fx = LINEAR_BW;	break;

	case 'e':
	case 'E': cmd.bw_fx = EXP_BW;		break;

	case 'd':
	case 'D': cmd.bw_fx = DYNAMIC_BW;	break;

	case 'a': /** adaptive **/
	case 'A': cmd.bw_fx = GDFS_dBS_BW;	break;

	case 'h': /** hybrid **/
	case 'H': cmd.bw_fx = GDFS_fBS_BW;	break;

	case 'f':
	case 'F':
	case '-':
	default:  cmd.bw_fx = FIXED_BW;		break;
	}
    }

    break;
    } /* switch */

    MaxDepth = FirstMaxDepth;
    num = MaxDepth + 1;
    sp->band.tree = (tree_ *) malloc (num * sizeof (tree_));
    for (i = 0; i < num; i++)
	init_tree_struct (sp->band.tree, i, sp->band.init_width);
}


void init_tree_struct (tp, depth, init_width)
    tree_ *tp;
    int depth, init_width;
{
    (tp+depth)->young = NULL;
    (tp+depth)->count = 0;
    (tp+depth)->width = bandwidth_function (depth, init_width);
}


void init_all_bandwidths (sp)
    search_ *sp;
{   int i, num;

    num = MaxDepth + 1;
    for (i = 0; i < num; i++)
	init_tree_struct (sp->band.tree, i, sp->band.init_width);
}


void init_search_struct (sp, init_open_ptr, init_alg_ptr)
  search_ *sp;
  int init_open_ptr, init_alg_ptr;
{
  sp->strategy = cmd.strategy;
  sp->algorithm = cmd.algorithm;
  init_stat_struct(&(sp->stat));
  init_constr_struct(&(sp->constr));
  init_approx_struct(&(sp->approx));
  if (init_open_ptr) sp->open.bptree = NULL;
  if (init_alg_ptr) sp->alg.def = NULL;
  sp->child = NULL;
  sp->brother = NULL;
  sp->parent = NULL;
  sp->idle_clk = 0;

  sp->incumbent = NULL;
  sp->glub = problem.huge;
  sp->gllb = - problem.huge;

  switch (sp->strategy) {
  case BAND: init_band_struct (sp);	break;
  case FBS:  init_fbs_struct (sp);	break;
  default:   sp->band.tree = NULL;	break;
  }
}


void init_regress_struct (rp)
  regress_ *rp;
{ int i;

  rp->x_sum = 0.0;
  rp->x2_sum = 0.0;
  rp->y_sum = 0.0;
  rp->xy_sum = 0.0;
  rp->n = 0.0;
  for (i = 0; i <= REGRESSION_ORDER; i++) rp->beta[i] = 0.0;
}


void init_stca_struct (p)
  stca_ *p;
{ p->g_factor = cmd.factor.g; }


void init_ptca_struct (p)
  ptca_ *p;
{
  p->s_factor = cmd.factor.s;
  p->c_factor = cmd.factor.c;
  init_regress_struct(&(p->regress));
}


void init_dtca_struct (p)
  dtca_ *p;
{
  p->s_factor = cmd.factor.s;
  p->c_factor = cmd.factor.c;
  init_regress_struct(&(p->regress));
}


void init_stcgd_struct (p)
  stcgd_ *p;
{ p->g_factor = cmd.factor.g; }


void init_ptcgd_struct (p)
  ptcgd_ *p;
{
  p->g_factor = cmd.factor.g;
  p->s_factor = cmd.factor.s;
  p->c_factor = cmd.factor.c;
  init_regress_struct(&(p->regress));
}


/** begin, Arthur's, 4/10/92 **/
#define MIDA_DEF_GFACT	3.0
#define MIDA_DEF_NBIN	20

void init_mida_struct (p)
mida_ *p;
{ float get_float_cmd_line_param ();

  p->g_factor = get_float_cmd_line_param (MIDA_DEF_GFACT, 0, 0);
  p->mida_n = get_int_cmd_line_param (MIDA_DEF_NBIN, 1, 0);
  p->bin = NULL;
  p->bfsbin = NULL;
#ifdef UPAST
  p->oldbin = NULL;
#endif
  p->threshold = p->data.opt_value = (domain) 0;
  p->iter = p->data.b_cost =  p->data.mida_cost = 0;
  p->data.max_exp_factor = p->data.tot_exp_factor = p->data.sqr_exp_factor = 0.0;
  p->data.min_exp_factor = HUGE_FLOAT;
} 
/** end, Arthur's, 4/10/92 **/


void init_ida_cr_struct (p)
ida_cr_ *p;
{
    int i;
    float get_float_cmd_line_param();

    p->r = get_float_cmd_line_param(2.0, 0, 0);
    p->x = (domain) 0;
    for (i = 0; i < IDA_CR_NUM_BIN; i++) p->bin[i] = 0;
}


void init_new_ida_cr_struct (p)
new_ida_cr_ *p;
{
    int i;
    float get_float_cmd_line_param();

    p->r = get_float_cmd_line_param(2.0, 0, 0);
    p->x = (domain) 0;
    p->mida = get_int_cmd_line_param (0, 1, 0);
    for (i = 0; i < NEW_IDA_CR_NUM_BIN; i++) p->bin[i] = 0;
}


void init_set_th_struct (p, idx)
  set_th_ *p;
  int idx;
{
  switch (**(cmd.param + idx++))
  {
    case '-':
    case 'n': p->alg = NAIVE_TH; 	break;
    case 's': p->alg = STATIC_TH; 	break;
    case 'p': p->alg = PREDICTIVE_TH; 	break;
    default: error("init_set_th_struct: no such algorithm"); break;
  }

  switch (**(cmd.param + idx))
  {
    case '-':
    case '1': p->type = TYPE_I;   break;
    case '2': p->type = TYPE_II;  break;
    case '3': p->type = TYPE_III; break;
    case '4': p->type = TYPE_IV;  break;
    default: error("init_set_th_struct: no such type"); break;
  }

  p->nbin = cmd.nbin;
  p->bin = NULL;
  p->calibrate = NULL;
}


void create_th_bin (set_th)
set_th_ *set_th;
{
    int i;

    set_th->bin = (long *) malloc (set_th->nbin * sizeof (long));
    set_th->calibrate = (domain *) malloc (set_th->nbin * sizeof (domain));
    for (i = 0; i < set_th->nbin; i++)
    {
	*(set_th->bin + i) = (long) 0;
	*(set_th->calibrate + i) = (domain) 0;
    }
}


float get_float_cmd_line_param (def_val, loc, offset)
  float def_val;
  int loc, offset;
{ float val;

  if (loc >= cmd.nparam) return def_val;
  if (*(*(cmd.param + loc) + offset) == '-') return def_val;
  sscanf((*(cmd.param + loc) + offset), "%f", &val);
  return val;
}


int get_int_cmd_line_param (def_val, loc, offset)
  int def_val;
  int loc, offset;
{ int val;

  if (loc >= cmd.nparam) return def_val;
  if (*(*(cmd.param + loc) + offset) == '-') return def_val;
  sscanf((*(cmd.param + loc) + offset), "%d", &val);
  return val;
}


float get_float_cmd_line_auxparam (def_val, loc, offset)
  float def_val;
  int loc, offset;
{
  float val;

  if (loc >= cmd.nparam) return def_val;
  if (*(*(cmd.param + loc) + offset) == '-') return def_val;
  sscanf((*(cmd.param + loc) + offset), "%f", &val);
  return val;
}


int get_int_cmd_line_auxparam (def_val, loc, offset)
  int def_val;
  int loc, offset;
{
  int val;

  if (loc >= cmd.nparam) return def_val;
  if (*(*(cmd.param + loc) + offset) == '-') return def_val;
  sscanf((*(cmd.param + loc) + offset), "%d", &val);
  return val;
}


void init_tcr_struct (p)
tcr_ *p;
{
    int i;

    p->x = (domain) 0;
    for (i = 0; i < TCR_NUM_BIN; i++) p->bin[i] = 0;
}


void init_urts_struct (p)
  urts_ *p;
{
  p->last_pred = (**(cmd.param) == 'y') ? YES : NO;
  p->th = problem.huge;
  switch (*(*(cmd.param+1) + 1)) {
    case '1': p->a_type = TYPE_I;   break;
    case '2': p->a_type = TYPE_II;  break;
    case '3': p->a_type = TYPE_III; break;
    case '4': p->a_type = TYPE_IV;  break;
    case '5': p->a_type = TYPE_V;   break;
    case '6': p->a_type = TYPE_VI;  break;
    default:  p->a_type = TYPE_I;   break;
  }
  p->g_factor = get_float_cmd_line_param (HUGE_FLOAT, 4, 0);
  p->r_factor = get_float_cmd_line_param (HUGE_FLOAT, 5, 0);
  p->alg = (**(cmd.param+1) == 'a') ? UNARY_APPROX : UNARY_TH;

  p->set_th = (set_th_ *) malloc (sizeof (set_th_));
  init_set_th_struct (p->set_th, 2);
  create_th_bin (p->set_th);

  p->tcr = (tcr_ *) malloc (sizeof (tcr_));
  init_tcr_struct (p->tcr);
}


void init_brts_struct (p)
  brts_ *p;
{
  p->last_pred = (**(cmd.param) == 'y') ? YES : NO;
  p->th = problem.huge;
  switch (*(*(cmd.param+1) + 1)) {
    case '1': p->a_type = TYPE_I;   break;
    case '2': p->a_type = TYPE_II;  break;
    case '3': p->a_type = TYPE_III; break;
    case '4': p->a_type = TYPE_IV;  break;
    default:  p->a_type = TYPE_I;   break;
  }
  p->g_factor = get_float_cmd_line_param(HUGE_FLOAT, 3, 0);
  p->r_factor = get_float_cmd_line_param(HUGE_FLOAT, 4, 0);
  p->set_th = (set_th_ *) malloc(sizeof(set_th_));
  init_set_th_struct(p->set_th, 1);
  if (cmd.nparam >= 6) p->set_th->nbin = atoi(*(cmd.param + 5));
  create_th_bin(p->set_th);
}


void init_hrts_struct (p)
  hrts_ *p;
{
  p->last_pred = (**(cmd.param) == 'y') ? YES : NO;
  p->th = problem.huge;
  switch (*(*(cmd.param+1) + 1)) {
    case '1': p->a_type = TYPE_I;   break;
    case '2': p->a_type = TYPE_II;  break;
    case '3': p->a_type = TYPE_III; break;
    case '4': p->a_type = TYPE_IV;  break;
    default:  p->a_type = TYPE_I;   break;
  }
  p->g_factor = get_float_cmd_line_param(HUGE_FLOAT, 3, 0);
  p->r_factor = get_float_cmd_line_param(HUGE_FLOAT, 4, 0);
  p->set_th = (set_th_ *) malloc(sizeof(set_th_));
  init_set_th_struct(p->set_th, 1);
  if (cmd.nparam >= 6) p->set_th->nbin = atoi(*(cmd.param + 5));
  create_th_bin(p->set_th);
}


void init_uird_struct (p)
  uird_ *p;
{
  if ((p->basic.shortcut = (**(cmd.param) == 'y') ? YES : NO) == YES)
    p->basic.margin = get_float_cmd_line_param(0.0, 0, 1);
  p->th = problem.huge;
  switch (*(*(cmd.param+1) + 1)) {
    case '1': p->a_type = TYPE_I;   break;
    case '2': p->a_type = TYPE_II;  break;
    case '3': p->a_type = TYPE_III; break;
    case '4': p->a_type = TYPE_IV;  break;
    default:  p->a_type = TYPE_I;   break;
  }
  p->g_factor = get_float_cmd_line_param(HUGE_FLOAT, 4, 0);
  p->r_factor = get_float_cmd_line_param(HUGE_FLOAT, 5, 0);
  p->basic.base = get_float_cmd_line_param(0.0, 6, 0);
  p->alg = (**(cmd.param+1) == 'a') ? UNARY_APPROX : UNARY_TH;
  p->set_th = (set_th_ *) malloc(sizeof(set_th_));
  init_set_th_struct(p->set_th, 2);
  if (cmd.nparam >= 8) p->set_th->nbin = atoi(*(cmd.param + 7));
  create_th_bin(p->set_th);
}


void init_bird_struct (p)
  bird_ *p;
{
  if ((p->basic.shortcut = (**(cmd.param) == 'y') ? YES : NO) == YES)
    p->basic.margin = get_float_cmd_line_param(0.0, 0, 1);
  p->th = problem.huge;
  switch (*(*(cmd.param+1) + 1)) {
    case '1': p->a_type = TYPE_I;   break;
    case '2': p->a_type = TYPE_II;  break;
    case '3': p->a_type = TYPE_III; break;
    case '4': p->a_type = TYPE_IV;  break;
    default:  p->a_type = TYPE_I;   break;
  }
  p->g_factor = get_float_cmd_line_param(HUGE_FLOAT, 3, 0);
  p->r_factor = get_float_cmd_line_param(HUGE_FLOAT, 4, 0);
  p->basic.base = get_float_cmd_line_param(0.0, 5, 0);
  p->set_th = (set_th_ *) malloc(sizeof(set_th_));
  init_set_th_struct(p->set_th, 1);
  if (cmd.nparam >= 7) p->set_th->nbin = atoi(*(cmd.param + 6));
  create_th_bin(p->set_th);
}


void init_node_struct (p, parentp)
  node_ *p, *parentp;
{
  p->parent = parentp;    
  p->next = p->brother = NULL;
  p->left = p->right = p->up = NULL;
  p->nsprout = 0;
  p->type = 1;
  p->ndecomp = 0;
  if (parentp) { (parentp->nsprout)++; p->depth = parentp->depth + 1; }
  else p->depth = 0;
} 


