/*+ ------------------------------------------------------------------- +
  +	commmand line syntax						+
  + 									+
  +  search [:virtual | :real]						+
  +	    [:report [report-file]]					+
  +         [:search bfs [default | lw | stca | ptca | dtca] |		+
  +          :search gdfs [default | lw | stcgd | ptcgd | mida] |	+
  +          :search dfs [default | lw | ida] |				+
  +          :prob iter size seed					+
  +         [:constr time [space [cst [bound]]]]			+
  +         [:approx approx]						+
  +	    [:accu accuracy]						+
  +         [:dbg dbg]							+
  +         [:io stat [summary [graph-ready [rt_pf [st_pf]]]]]		+
  +         [:pf rt [st]]						+
  +         [:param n p1 p2 p3 ... pn] ([:param g|s [c]])		+
  +         [:auxparam n p1 p2 p3 ... pn] 				+
  +	    [:pe n]							+
  +	    [:comm_idle n]						+
  +	    [:first n]	(-1 single; otherwise bfs)			+
  +         [:nbin n]							+
  +         [:rel_a_metric a]						+
  +         [:ens]							+
  +         [:ignore_last_bin]						+
  +         [:first_sol]						+
  +         [:opt_sol opt-sol sol-precision]				+
  + ------------------------------------------------------------------- +*/

#define	is_not_default		(*(argv[index]) != '-') 
#define	inc_index(abort_cond)	if (++index >= argc) { \
                                  if (abort_cond) error(errmsg); \
				  else break; \
				} \
				if (*(argv[index]) == ':') \
                                  if (abort_cond) error(errmsg); \
				  else continue;

void cmd_line (argc, argv)
  int argc;
  char *argv[];
{ int index = 1, is_prob_spec = 0, i;
  enum_type_ t1 = TYPE_NULL, t2 = TYPE_NULL;
  float f_val;
  char *errmsg = "not sufficient arguments\n";
#ifdef DEBUG
if (cmd.debug >= 1) {
  printf("enter cmd_line processor\n");
  for(index = 0; index < argc; index++) printf("%s ", argv[index]);
  printf("\n");
  index = 1;
}
#endif

  if (argc <= 1) error(errmsg);

  while (index < argc) {

    if (streql(argv[index],":real")) {
      cmd.count = REAL;
      inc_index(0);
    } /* :real */

    else if (streql(argv[index],":virtual")) {
      cmd.count = VIRTUAL; 
      inc_index(0);
    } /* :virtual */

    else if (streql(argv[index],":task")) {
      inc_index(1);
      if (streql(argv[index],"csp")) problem.task = CSP;
      else if (streql(argv[index],"learn")) problem.task = LEARNING;
      else error("no such search task\n");
      inc_index(0);
    }

    else if (streql(argv[index],":report")) {
      cmd.report = 1;
      inc_index(0);
      if (is_not_default) io.report = argv[index];
      inc_index(0);
    } /* :report */

    else if (streql(argv[index],":ens")) {
      cmd.ens = YES;
      inc_index(0);
    } /* :ens */

    else if (streql(argv[index],":ignore_last_bin")) {
      cmd.ignore_last_bin = YES;
      inc_index(0);
    } /* :ignore_last_bin */

    else if (streql(argv[index],":first_sol")) {
      cmd.first_sol = YES;
      inc_index(0);
    } /* :first_sol */

    else if (streql(argv[index],":opt_sol")) {
      cmd.opt_sol = YES;
      inc_index(1);
      sscanf (argv[index], "%f", &f_val);
      cmd.opt_sol_val = (domain) f_val;
      inc_index(0);
      if (is_not_default) {
        sscanf (argv[index], "%f", &f_val);
        cmd.sol_precision = f_val;
      }
      inc_index(0);
    } /* :opt_sol */

    else if (streql(argv[index],":search")) {
      inc_index(1);
      if (streql(argv[index],"bfs")) cmd.strategy = BFS;
      else if (streql(argv[index],"dfs")) cmd.strategy = DFS;
      else if (streql(argv[index],"gdfs")) cmd.strategy = GDFS;
      else if (streql(argv[index],"gbb")) cmd.strategy = GBB;
      else if (streql(argv[index],"band")) cmd.strategy = BAND;
      else if (streql(argv[index],"fbs")) cmd.strategy = FBS;
      else error("no such search strategy\n");
      inc_index(0);

      /* TCA(approx) */
      if (streql(argv[index],"stca")) cmd.algorithm = sTCA;
      else if (streql(argv[index],"ptca")) cmd.algorithm = pTCA;
      else if (streql(argv[index],"dtca")) cmd.algorithm = dTCA;

      /* TCA(accu) */
      if (streql(argv[index],"sTCA"))
	{ cmd.algorithm = sTCA; cmd.xon = YES; }
      else if (streql(argv[index],"pTCA"))
	{ cmd.algorithm = pTCA; cmd.xon = YES; }
      else if (streql(argv[index],"dTCA"))
	{ cmd.algorithm = dTCA; cmd.xon = YES; }

      /* TCGD(approx) */
      else if (streql(argv[index],"stcgd")) cmd.algorithm = sTCGD;
      else if (streql(argv[index],"ptcgd")) cmd.algorithm = pTCGD;

      /* TCGD(accu) */
      else if (streql(argv[index],"sTCGD"))
	{ cmd.algorithm = sTCGD; cmd.xon = YES; }
      else if (streql(argv[index],"pTCGD"))
	{ cmd.algorithm = pTCGD; cmd.xon = YES; }

      /* uRTS(approx) */
      else if (streql(argv[index],"urts")) cmd.algorithm = uRTS;

      /* uRTS(accu) */
      else if (streql(argv[index],"uRTS"))
	{ cmd.algorithm = uRTS; cmd.xon = YES; }

      /* bRTS(approx) */
      else if (streql(argv[index],"brts")) cmd.algorithm = bRTS;

      /* bRTS(accu) */
      else if (streql(argv[index],"bRTS"))
	{ cmd.algorithm = bRTS; cmd.xon = YES; }

      /* hRTS(approx) */
      else if (streql(argv[index],"hrts")) cmd.algorithm = hRTS;

      /* hRTS(accu) */
      else if (streql(argv[index],"hRTS")) 
	{ cmd.algorithm = hRTS; cmd.xon = YES; }

      /* uIRD(approx) */
      else if (streql (argv[index],"uird")) cmd.algorithm = uIRD;

      /* uIRD(accu) */
      else if (streql (argv[index],"uIRD"))
	{ cmd.algorithm = uIRD; cmd.xon = YES; }

      /* bIRD(approx) */
      else if (streql (argv[index],"bird")) cmd.algorithm = bIRD;

      /* bIRD(accu) */
      else if (streql (argv[index],"bIRD"))
	{ cmd.algorithm = bIRD; cmd.xon = YES; }

      /* LW(approx) */
      else if (streql (argv[index],"lw")) cmd.algorithm = LW;

      /* LW(accu) */
      else if (streql (argv[index],"LW"))
	{ cmd.algorithm = LW; cmd.xon = YES; }

      /* IRA*(approx) */
      else if (streql (argv[index],"ira")) cmd.algorithm = IRA;

      /* IRA*(accu) */
      else if (streql (argv[index],"IRA"))
	{ cmd.algorithm = IRA; cmd.xon = YES; }

      /* IDA*(approx) */
      else if (streql (argv[index],"ida")) cmd.algorithm = IDA;

      /* IDA*(accu) */
      else if (streql (argv[index],"IDA"))
	{ cmd.algorithm = IDA; cmd.xon = YES; }

      /* MIDA*(approx) */
      else if (streql (argv[index],"mida")) cmd.algorithm = MIDA;

      /* IDA*_CR(approx) */
      else if (streql (argv[index],"ida_cr")) cmd.algorithm = IDA_CR;

      /* NEW_IDA*_CR(approx) */
      else if (streql (argv[index],"new_ida_cr")) cmd.algorithm = NEW_IDA_CR;

      /* iFBS (approx) */
      else if (streql (argv[index],"ifbs")) cmd.algorithm = iFBS;

      /* gDFS_iFBS (approx) */
      else if (streql (argv[index],"gdfs_ifbs")) cmd.algorithm = gDFS_iFBS;

      /* MIDA*(accu) */
      else if (streql (argv[index],"MIDA"))
	{ cmd.algorithm = MIDA; cmd.xon = YES; }

      /* IDA*_CR(accu) */
      else if (streql (argv[index],"IDA_CR"))
	{ cmd.algorithm = IDA_CR; cmd.xon = YES; }

      /* NEW_IDA*_CR(accu) */
      else if (streql (argv[index],"NEW_IDA_CR"))
	{ cmd.algorithm = NEW_IDA_CR; cmd.xon = YES; }

      /* iFBS (accu) */
      else if (streql (argv[index],"iFBS"))
	{ cmd.algorithm = iFBS; cmd.xon = YES; }

      /* gDFS_iFBS (accu) */
      else if (streql (argv[index],"gDFS_iFBS"))
	{ cmd.algorithm = gDFS_iFBS; cmd.xon = YES; }

      /* DFS*(approx) */
      else if (streql (argv[index],"dfs")) cmd.algorithm = DFS_star;

      /* DFS*(accu) */
      else if (streql (argv[index],"DFS"))
	{ cmd.algorithm = DFS_star; cmd.xon = YES; }

      /* PBFS(approx) */
      else if (streql (argv[index],"pbfs")) cmd.algorithm = PBFS;

      /* PBAND(approx) */
      else if (streql (argv[index],"pband")) cmd.algorithm = PBAND;

      /* PBFS(accu) */
      else if (streql (argv[index],"PBFS"))
	{ cmd.algorithm = PBFS; cmd.xon = YES; }

      /* PBAND(accu) */
      else if (streql (argv[index],"PBAND"))
	{ cmd.algorithm = PBAND; cmd.xon = YES; }

      /* PDFS(approx) */
      else if (streql (argv[index],"pdfs")) cmd.algorithm = PDFS;

      /* PDFS(accu) */
      else if (streql (argv[index],"PDFS"))
	{ cmd.algorithm = PDFS; cmd.xon = YES; }

      /* PGDFS(approx) */
      else if (streql (argv[index],"pgdfs")) cmd.algorithm = PGDFS;

      /* PGDFS(accu) */
      else if (streql (argv[index],"PGDFS"))
	{ cmd.algorithm = PGDFS; cmd.xon = YES; }

      /* PGBB(approx) */
      else if (streql (argv[index],"pgbb")) cmd.algorithm = PGBB;

      /* PGBB(accu) */
      else if (streql (argv[index],"PGBB"))
	{ cmd.algorithm = PGBB; cmd.xon = YES; }

      /* DEFAULT(approx) */
      else if (streql (argv[index],"def")) cmd.algorithm = DEFAULT;
      else if (streql (argv[index],"default")) cmd.algorithm = DEFAULT;

      /* DEFAULT(accu) */
      else if (streql (argv[index],"DEF"))
	{ cmd.algorithm = DEFAULT; cmd.xon = YES; }
      else if (streql (argv[index],"DEFAULT"))
	{ cmd.algorithm = DEFAULT; cmd.xon = YES; }

      else error("no such search algorithm\n");
      inc_index(0);
    } /* :search */

    else if (streql (argv[index],":prob")) {
      inc_index(1);
      cmd.iter = atoi (argv[index]);
      inc_index(1);
      problem.size = SIZE = atoi (argv[index]);
      inc_index(1)
      if (problem.size > MaxProblemSize) {
        printf("MaxProblemSize=%d, problem.size=%d\n",
	        MaxProblemSize, problem.size);
        error("problem size too large\n");
      }
      problem.seed = atoi(argv[index]);
      is_prob_spec = 1;
      inc_index(0);
    } /* :prob */

    else if (streql(argv[index],":io")) {
      inc_index(1);
      if (is_not_default) io.stat = argv[index];
      inc_index(0);
      if (is_not_default) io.summary = argv[index];
      inc_index(0);
      if (is_not_default) io.graph = argv[index];
      inc_index(0);
      if (is_not_default) io.rt_pf = argv[index];
      inc_index(0);
      if (is_not_default) io.st_pf = argv[index];
      inc_index(0);
    } /* :io */

    else if (streql(argv[index],":approx")) {
      inc_index(1);
      if (is_not_default) sscanf(argv[index], "%f", &(cmd.approx));
      inc_index(0);
    } /* :approx */

    else if (streql(argv[index],":accu")) {
      inc_index(1);
      if (is_not_default) sscanf(argv[index], "%f", &(cmd.accu));
      inc_index(0);
    } /* :accu */

    else if (streql(argv[index],":dbg")) {
      inc_index(1);
      if (is_not_default) cmd.debug = atoi(argv[index]);
      inc_index(0);
    } /* :dbg */

    else if (streql(argv[index],":rel_a_metric")) {
      cmd.approx_metric = RELATIVE;
      inc_index(1);
      if (is_not_default) {
	sscanf (argv[index], "%f", &f_val);
	cmd.rel_a = f_val;
      }
      inc_index(0);
    }

    else if (streql(argv[index],":constr")) {
      inc_index(1);
      if (is_not_default) cmd.constr.time = atoi(argv[index]);
      inc_index(0);
      if (is_not_default) cmd.constr.space = atoi(argv[index]);
      inc_index(0);
      if (is_not_default) {
	sscanf (argv[index], "%f", &f_val);
	cmd.constr.cst = f_val;
      }
      inc_index(0);
      if (is_not_default) {
	sscanf (argv[index], "%f", &f_val);
	cmd.constr.bound = (domain) f_val;
      }
      inc_index(0);
      if (is_not_default) {
	if (*(argv[index]) == 'r' || *(argv[index]) == 'R') {
	  cmd.bound_metric = RELATIVE;
	  if (isdigit (*(argv[index]+1)) || ispunct (*(argv[index]+1))) {
	    sscanf (argv[index]+1, "%f", &f_val);
	    cmd.bound_base = (domain) f_val;
	  }
	}
      }
      inc_index(0);
    } /* :constr */

    else if (streql(argv[index],":pf")) {  /* profile */
      inc_index(1);
      if (is_not_default) {
        cmd.pf.rt = YES;
        sscanf(argv[index], "%f", &(cmd.pf.step));
      }
      inc_index(0);
      if (is_not_default) cmd.pf.st = YES;
      inc_index(0);
    } /* :pf */

    else if (streql (argv[index],":pe")) {	/* num of pe's */
      inc_index(1);
      if (is_not_default)
        cmd.num_pe = atoi (argv[index]);
      inc_index(0);
    } /* :pe */

    else if (streql (argv[index],":comm_idle")) {	/* comm overhead */
      inc_index(1);
      if (is_not_default)
        cmd.comm_idle = atoi (argv[index]);
      inc_index(0);
    } /* :comm_idle */

    else if (streql (argv[index],":first")) {	/* starting search */
      inc_index(1);
      if (is_not_default)
        cmd.first = atoi (argv[index]);
      inc_index(0);
    } /* :first */

    else if (streql (argv[index],":nbin")) {	/* num of bins */
      inc_index(1);
      if (is_not_default)
        cmd.nbin = atoi (argv[index]);
      inc_index(0);
    } /* :nbin */

    else if (streql (argv[index],":param")) {	/* algorithm parameters */
      inc_index(1);
      if (is_not_default) {
	cmd.nparam = atoi (argv[index]);
	cmd.param = (char **) malloc(cmd.nparam * sizeof(char *));
      }
      for (i = 0; i < cmd.nparam; i++) {
	inc_index(1);
	*(cmd.param + i) = argv[index];
      }
      inc_index(0);
    } /* :param */

    else if (streql (argv[index],":auxparam")) {	/* strategy parameters */
      inc_index(1);
      if (is_not_default) {
	cmd.nauxparam = atoi (argv[index]);
	cmd.auxparam = (char **) malloc(cmd.nauxparam * sizeof(char *));
      }
      for (i = 0; i < cmd.nauxparam; i++) {
	inc_index(1);
	*(cmd.auxparam + i) = argv[index];
      }
      inc_index(0);
    } /* :auxparam */

    else error("no such option in cmd line\n");
  } /* while loop */

  if (cmd.accu != 1.0) cmd.approx = accu2approx (cmd.accu);

#ifdef DEBUG
if (cmd.debug >= 1) {
  printf("summary of command and search info:\n");
  printf("counting system=%s\n", dbg_count[toint(cmd.count)]);
  printf("task is %s\n", dbg_task[toint(problem.task)]);
  printf("search strategy=%s\n", dbg_strategy[toint(cmd.strategy)]);
  printf("search algorithm=%s\n", dbg_algorithm[toint(cmd.algorithm)]);
  printf("nsamples=%d\n", cmd.iter);
  printf("problem size=%d\n", problem.size);
  printf("random seed=%d\n", problem.seed);
  printf("stat file=%s\n", io.stat);
  printf("summary file=%s\n", io.summary);
  printf("graph-ready file=%s\n", io.graph);
  printf("run-time file=%s\n", io.rt_pf);
  printf("space-vs-time file=%s\n", io.st_pf);
  printf("report file=%s\n", io.report);
  printf("approx=%g\n", cmd.approx);
  printf("accu=%g\n", cmd.accu);
  printf("debugging mode=%d\n", cmd.debug);
  printf("time constr=");
  if (cmd.constr.time == huge_long) printf("unlimited\n");
  else printf("%d\n", cmd.constr.time);
  printf("space constr=");
  if (cmd.constr.space == huge_long) printf("unlimited\n");
  else printf("%d\n", cmd.constr.space);
  printf("cst constr=");
  if (cmd.constr.cst == huge_float) printf("unlimited\n");
  else printf("%g\n", cmd.constr.cst);
  printf("profile run-time=%s\n", (cmd.pf.rt == YES) ? "YES" : "NO");
  if (cmd.pf.rt == YES) printf("profiling step=%f\n", cmd.pf.step);
  printf("profile space-vs-time=%s\n", (cmd.pf.st == YES) ? "YES" : "NO");
}
#endif

  if (! is_prob_spec) error("problem is not specified\n");

  /* specify some special parameters for Chu */
  switch(cmd.algorithm) {
    case sTCA:  /* :param 1 g */
    case sTCGD: /* :param 1 g */
		sscanf(*(cmd.param), "%f", &(cmd.factor.g)); break;
    case pTCA:  /* :param 2 s c */
    case dTCA:  /* :param 2 s c */
    case pTCGD: /* :param 2 s c */
		sscanf(*(cmd.param), "%f", &(cmd.factor.s));
		sscanf(*(cmd.param+1), "%f", &(cmd.factor.c)); break;
    default: 	break;
  }

#ifdef DEBUG
if (cmd.debug >= 1) {
  printf("g_factor=%g\n", cmd.factor.g);
  printf("s_factor=%g\n", cmd.factor.s);
  printf("c_factor=%g\n", cmd.factor.c);
  printf("num_parameters=%d\n", cmd.nparam);
  for (i = 0; i < cmd.nparam; i++) printf("param[%d]=%s\n", i, *(cmd.param+i));
  printf("num_aux_parameters=%d\n", cmd.nauxparam);
  for (i = 0; i < cmd.nauxparam; i++) printf("aux_param[%d]=%s\n", i, *(cmd.auxparam+i));
  printf("exit cmd_line processor\n");
}
#endif
}

