
void free_environ (sp, not_top_level)
  search_ *sp;
  int not_top_level;
{ search_ *p;

  if (sp) {

    if (sp->child) free_environ (sp->child, 1);
    if (sp->brother) free_environ (sp->brother, 1);

    free_open(sp);

    if (not_top_level) {
      free_alg_struct(sp);
      free(sp);
    }
  }
}


void free_alg_struct (sp)
  search_ *sp;   
{ 
  if (sp->alg.def)
    switch (sp->algorithm)
    {
      case DEFAULT: 			sp->alg.def = NULL;   break;
      case LW:                    	sp->alg.lw = NULL;    break;
      case sTCA:  free (sp->alg.stca);	sp->alg.stca = NULL;  break;
      case pTCA:  free (sp->alg.ptca);  sp->alg.ptca = NULL;  break;
      case dTCA:  free (sp->alg.dtca);  sp->alg.dtca = NULL;  break;
      case sTCGD: free (sp->alg.stcgd); sp->alg.stcgd = NULL; break;
      case pTCGD: free (sp->alg.ptcgd); sp->alg.ptcgd = NULL; break;
      case MIDA:  free (sp->alg.mida);  sp->alg.mida = NULL;  break;
      case IDA_CR: free (sp->alg.ida_cr);  sp->alg.ida_cr = NULL;  break;
      case NEW_IDA_CR: free (sp->alg.new_ida_cr);  sp->alg.new_ida_cr = NULL;  break;
      case uRTS:  free (sp->alg.urts);  sp->alg.urts = NULL;  break;
      case bRTS:  free (sp->alg.brts);	sp->alg.brts = NULL;  break;
      case hRTS:  free (sp->alg.hrts);	sp->alg.hrts = NULL;  break;
      case uIRD:  free (sp->alg.uird);	sp->alg.uird = NULL;  break;
      case bIRD:  free (sp->alg.bird);	sp->alg.bird = NULL;  break;
      default:    error ("free_alg_struct: no such algorithm\n"); break;
    }
}


void free_open (sp)
  search_ *sp;
{ int i;

  switch (sp->strategy)
  {
    case GBB:  free_list(sp->open.list);   sp->open.list = NULL;  break;
    case BFS:  free_bptree(sp->open.bptree); sp->open.bptree = NULL; break;
    case DFS:  free_stack(sp->open.stack); sp->open.stack = NULL; break;
    case GDFS: free_list(sp->open.list);   sp->open.list = NULL;  break;
    case FBS:  /** first-band search **/
    case BAND: free_btree(sp->open.btree); sp->open.btree = NULL;
	       for (i = sp->band.ceiling_depth; i <= sp->band.floor_depth; i++) {
		 free_btree ((sp->band.tree+i)->young);
		 (sp->band.tree+i)->young = NULL;
	       }
	       break;
    default:   free_list(sp->open.list);   sp->open.list = NULL;  break;
  }
}


void free_list (list)
  node_ *list;
{ node_ *p, *p0;

  for (p = list; p; p = p0) {
    p0 = p->next;
    free_node(p);
  }
}


void free_btree (tree)
  node_ *tree;
{ node_ *p, *p0;

  for (p = tree; p; p = p0) {
    p0 = p->next;
    free_node(p);
  }
}


void free_stack (stk)
  node_ *stk;
{ node_ *p, *p0;

  for (p = stk; p; p = p0) {
    p0 = p->next;
    free_node(p);
  }
}


#ifndef DELAY_FREE_NODE
void free_node (node)
  node_ *node;
{ node_ *parentp;

  /* bad-node test */
  if (node == NULL) return;

  /* reset activeness of node */
  node->type = INACTIVE;

  /* free this node */
  dispose_search_node_to_pool(node);
}

#else

void free_node (node)
  node_ *node;
{ node_ *parentp;

  /* bad-node test */
  if (node == NULL) return;

  /* reset activeness of node */
  node->type = INACTIVE;

  if (node->nsprout <= 0) {
    while (node) {

      /* remember parent */
      parentp = node->parent;

#ifdef DEBUG
if (cmd.debug >= 3) printf("free node=%d\n", node);
#endif

      /* free this node */
      dispose_search_node_to_pool(node);

      /* update parent's nsprout
       * further if it is zero, then free parent also
       */
      if (parentp)
        if (--(parentp->nsprout) > 0) node = NULL;
        else node = parentp;
      else node = NULL;
    }
  }
#ifdef DEBUG
  else
    if (cmd.debug >= 3)
      printf("not pruned due to nsprout=%d\n", node->nsprout);
#endif
}
#endif

