
search_ *init_environ (sp)
  search_ *sp;
{ search_ *new_sp;
  domain base;

  _is_first_sol_found_ = NO;

  new_sp = (sp) ? sp : ((search_ *) malloc(sizeof(search_)));

  if (sp) {
    init_search_struct(new_sp, 1, 0);
    init_alg_struct(new_sp);
  }
  else {
    init_search_struct(new_sp, 1, 1);
    attach_alg_struct(new_sp);
    clear_files();
  }

  idpd_init();
  examine_root(new_sp);
  if (cmd.bound_metric == RELATIVE) {
    base = (cmd.bound_base == problem.huge) ? ROOT->upb : cmd.bound_base;
    new_sp->constr.bound = ROOT->lowb +
		(base - ROOT->lowb) * cmd.constr.bound / 100.0;
  }
  if (cmd.approx_metric == RELATIVE) {
    new_sp->approx.approx = cmd.rel_a * new_sp->approx.root_approx;
  }
  return new_sp;
}


void clear_files ()
{ char cmd[100];

  fclose(fopen(io.stat,"w"));
  fclose(fopen(io.summary,"w"));
  fclose(fopen(io.rt_pf,"w"));
  fclose(fopen(io.st_pf,"w"));
  fclose(fopen(io.report,"w"));
  
  sprintf(cmd, "/bin/rm -f %s %s %s %s %s",
               io.stat, io.summary, io.rt_pf, io.st_pf, io.report);
  system(cmd);
}


void attach_alg_struct (sp)
  search_ *sp;
{ 
  switch (sp->algorithm)
  {
    case sTCA:	sp->alg.stca = (stca_ *) malloc (sizeof (stca_));    	break;
    case pTCA:	sp->alg.ptca = (ptca_ *) malloc (sizeof (ptca_));   	break;
    case dTCA:	sp->alg.dtca = (dtca_ *) malloc (sizeof (dtca_));   	break;
    case sTCGD:	sp->alg.stcgd = (stcgd_ *) malloc (sizeof (stcgd_));	break;
    case pTCGD:	sp->alg.ptcgd = (ptcgd_ *) malloc (sizeof (ptcgd_));	break;
    case MIDA:	sp->alg.mida = (mida_ *) malloc (sizeof (mida_));	break;
    case IDA_CR: sp->alg.ida_cr = (ida_cr_ *) malloc (sizeof (ida_cr_)); break;
    case NEW_IDA_CR: sp->alg.new_ida_cr = (new_ida_cr_ *) malloc (sizeof (new_ida_cr_)); break;
    case uRTS:	sp->alg.urts = (urts_ *) malloc (sizeof (urts_));	break;
    case bRTS:	sp->alg.brts = (brts_ *) malloc (sizeof (brts_));	break;
    case hRTS:	sp->alg.hrts = (hrts_ *) malloc (sizeof (hrts_));	break;
    case uIRD:	sp->alg.uird = (uird_ *) malloc (sizeof (uird_));	break;
    case bIRD:	sp->alg.bird = (bird_ *) malloc (sizeof (bird_));	break;
    default:     break;
  }
  init_alg_struct (sp);
}


void init_alg_struct (sp)
  search_ *sp;
{
  switch (sp->algorithm)
  {
    case sTCA:	init_stca_struct (sp->alg.stca);	break;
    case pTCA:	init_ptca_struct (sp->alg.ptca);	break;
    case dTCA:	init_dtca_struct (sp->alg.dtca);	break;
    case sTCGD:	init_stcgd_struct (sp->alg.stcgd);	break;
    case pTCGD:	init_ptcgd_struct (sp->alg.ptcgd);	break;
    case MIDA:	init_mida_struct (sp->alg.mida);	break;
    case IDA_CR: init_ida_cr_struct (sp->alg.ida_cr);	break;
    case NEW_IDA_CR: init_new_ida_cr_struct (sp->alg.new_ida_cr);	break;
    case uRTS:	init_urts_struct (sp->alg.urts);	break;
    case bRTS:	init_brts_struct (sp->alg.brts);	break;
    case hRTS:	init_hrts_struct (sp->alg.hrts);	break;
    case uIRD:	init_uird_struct (sp->alg.uird);	break;
    case bIRD:	init_bird_struct (sp->alg.bird);	break;
    default:	break;
  }
}


void examine_root (sp)
  search_ *sp;
{ node_ *root;
  solution_ *solp;

  /* set init threshold */
  ROOT = root = root_generator();
  evaluate_lower_bound(root);
  solp = evaluate_upper_bound(root, get_sol_buf());
  if (sp->incumbent == NULL) sp->incumbent = solp;
  else free_sol_buf(solp);
  sp->approx.x_root_approx = sp->approx.root_approx
		= calc_rt_approx(root->lowb, root->upb, NO, problem.huge);
  sp->gllb = sp->x_root_lb = sp->root_lb = root->lowb;
  sp->glub = sp->x_root_ub = sp->root_ub = root->upb;
  sp->approx.root_accu = approx2accu (sp->approx.root_approx);

#ifdef DEBUG
if (cmd.debug >= 1)
  printf("root: lowb=%g, upb=%g, root_approx=%g\n",
         ((float) sp->root_lb), ((float) sp->root_ub), sp->approx.root_approx);
#endif
}


node_ *create_root (sp)
  search_ *sp;
{ node_ *root;
  solution_ *solp;

#ifdef DEBUG
if (cmd.debug >= 2) printf("generate root\n");
#endif

  root = root_generator();

  evaluate_lower_bound(root);
  solp = evaluate_upper_bound(root, get_sol_buf());

  sp->root_lb = root->lowb;
  sp->root_ub = root->upb;

  sp->gllb = sp->root_lb;   /** always starting from the root **/

  if (sp->incumbent == NULL) {
    sp->incumbent = solp;
    sp->x_root_lb = root->lowb;
    sp->glub = sp->x_root_ub = root->upb;
  }
  else free_sol_buf (solp);

#ifdef DEBUG
if (cmd.debug >= 2) debug_node("root", root, sp);
#endif

  return root;
}

