
yesno_ is_null_device (sp)
  search_ *sp;
{
  if (sp->approx.approx > 0.0) return NO;
  if (sp->constr.bound < problem.huge) return NO;
 
  return YES;
}


domain expected_opt (sp)
  search_ *sp;
{ float temp;

  temp = tofloat(sp->glub) / (1.0 + sp->approx.approx);
  switch (problem.domain) {
    case INT:
    case LONG:   return ((domain) ceiling(temp)); break;
    case FLOAT:
    case DOUBLE:
    default:     break;
  }
  return ((domain) temp);
}


float calc_rt_approx (lowb, upb, th_flag, th)
  domain lowb, upb;
  yesno_ th_flag;
  domain th;
{ float rt, rt_th, up = tofloat(upb), low = tofloat(lowb);

  /* normal run-time approx */
  rt = (low == ((domain) 0) || lowb == problem.huge)
					? HUGE_FLOAT : (up - low) / low;
  rt = max(rt,0.0);

  /* threshold-induced run-time approx */
  if (th_flag == YES) {
    if (th == problem.huge) rt_th = 0.0;
    else rt_th = (th == ((domain) 0)) ? HUGE_FLOAT : (up - th) / th;
    rt = max(rt,rt_th);
  }
 
  return rt;
}


float get_rt_approx (sp)
  search_ *sp;
{ domain th = problem.huge;
  yesno_ th_flag = NO;

  if (is_threshold_related(sp) == YES) {
    th = get_threshold(sp);
    th_flag = YES;
  }
  return calc_rt_approx(sp->gllb, sp->glub, th_flag, th);
}


void eval_rt_approx(sp)
  search_ *sp;
{ domain th = problem.huge;
  yesno_ th_flag = NO;

  if (is_threshold_related(sp) == YES) {
    th = get_threshold(sp);
    th_flag = YES;
  }
  sp->approx.run_time = calc_rt_approx(sp->gllb, sp->glub, th_flag, th);
}


float get_approx (sp)
  search_ *sp;
{ return (max(sp->approx.run_time,sp->approx.approx)); }
  

void put_achieved_approx (sp, a)
  search_ *sp;
  float a;
{
  if (sp->approx.achieved > a) sp->approx.achieved = a;
}


void eval_final_approx (sp)
  search_ *sp;
{ 
  eval_rt_approx(sp);
  put_achieved_approx(sp, get_approx(sp));
}


yesno_ is_threshold_related (sp)
  search_ *sp;
{
  if (cmd.ens == YES) return YES;

  switch (sp->algorithm) {
    case uRTS: return ((sp->alg.urts->alg == UNARY_TH) ? YES : NO); break;
    case bRTS: return YES; break;
    case hRTS: return YES; break;
    case uIRD: return ((sp->alg.uird->alg == UNARY_TH) ? YES : NO); break;
    case bIRD: return YES; break;
    case IDA:  return YES; break;
    case MIDA:  return YES; break;
    case IDA_CR:  return YES; break;
    case NEW_IDA_CR:  return YES; break;
    case DFS_star:  return YES; break;
    default:     break;	
  }

  return NO;
}


domain get_threshold (sp)
  search_ *sp;
{
  if (cmd.ens == YES) return sp->constr.bound;

  switch (sp->algorithm) {
    case uRTS:  return sp->alg.urts->th;	break;
    case bRTS:  return sp->alg.brts->th;	break;
    case hRTS:  return sp->alg.hrts->th;	break;
    case uIRD:  return sp->alg.uird->th;	break;
    case bIRD:  return sp->alg.bird->th;	break;
    case IDA:   return sp->constr.bound;	break;
    case MIDA:  return sp->constr.bound;	break;
    case IDA_CR: return sp->constr.bound;	break;
    case NEW_IDA_CR: return sp->constr.bound;	break;
    case DFS_star: return sp->constr.bound;	break;
    default:	  error("get_threshold: no such algorithm"); break;
  }
  return problem.huge;
}


void put_threshold (sp, new_th)
  search_ *sp;
  domain new_th;
{
  switch (sp->algorithm) {
    case uRTS:	sp->alg.urts->th = new_th;	break;
    case bRTS:	sp->alg.brts->th = new_th;	break;
    case hRTS:	sp->alg.hrts->th = new_th;	break;
    case uIRD:	sp->alg.uird->th = new_th;	break;
    case bIRD:	sp->alg.bird->th = new_th;	break;
    case IDA:   sp->constr.bound = new_th;	break;
    case MIDA:  sp->constr.bound = new_th;	break;
    case IDA_CR: sp->constr.bound = new_th;	break;
    case NEW_IDA_CR: sp->constr.bound = new_th;	break;
    case DFS_star:   sp->constr.bound = new_th;	break;
    default:	error("put_threshold: no such algorithm"); break;
  }
}


void merge_threshold_to_parent (child_sp, sp)
  search_ *child_sp, *sp;
{ domain th = get_threshold(child_sp);

  put_threshold(sp, th);
  /** sp->gllb = th; **/
}


/* return 1, if violated;
 * return 0, otherwise.
 */
int is_constr_violated (sp)
  search_ *sp;
{ int signal = 0;

  if ( (get_time(sp) >= sp->constr.time)    ||
       (get_space(sp) >= sp->constr.space)  ||
       (get_cst(sp) >= sp->constr.cst) 		)  signal = 1;
   
#ifdef DEBUG
if (cmd.debug >= 2)
  if (signal) {
    printf("constr is violated: ");
    if (get_time(sp) >= sp->constr.time) printf("time(YES), ");
    else printf("time(NO), ");
    if (get_space(sp) >= sp->constr.space) printf("space(YES), ");
    else printf("space(NO), ");
    if (get_cst(sp) >= sp->constr.cst) printf("cst(YES)\n");
    else printf("cst(NO)\n");
    printf("(%d,%d), (%d,%d), (%f,%f)\n", get_time(sp), sp->constr.time,
           get_space(sp), sp->constr.space, get_cst(sp), sp->constr.cst);
  }
  else printf("no constr is violated\n");
#endif

  return signal;
}


void put_child_constr (childp, parentp, t_percent, s_percent, cst_percent)
  search_ *childp, *parentp;
  float t_percent, s_percent, cst_percent;
{ float time, space;

  if (parentp->constr.time == huge_long) childp->constr.time = HUGE_LONG;
  else {
    time = (float) (parentp->constr.time - get_time(parentp));
    time = time * t_percent;
    childp->constr.time = tolong(time);
  }
  if (parentp->constr.space == huge_long) childp->constr.space = HUGE_LONG;
  else {
    space = (float) (parentp->constr.space - get_space(parentp));
    space = space * s_percent;
    childp->constr.space = tolong(space);
  }
  childp->constr.cst = (parentp->constr.cst == huge_float) ?
          HUGE_FLOAT : (parentp->constr.cst - get_cst(parentp)) * cst_percent;
}


void set_stop_constr (sp, t_percent, s_percent, cst_percent)
  search_ *sp;
  float t_percent, s_percent, cst_percent;
{ float time = tofloat(sp->constr.time), space = tofloat(sp->constr.space);

  time = time * t_percent;    sp->constr.time = tolong(time);
  space = space * s_percent;  sp->constr.space = tolong(space);
  sp->constr.cst = sp->constr.cst * cst_percent;
}

