
#define	TRANSIENT_PHASE		20

int is_in_transient_phase (sp)
  search_ *sp;
{ return((get_time(sp) < TRANSIENT_PHASE)); }


/* if we are doing run-time profile, Real/Virtual_Time_Limit is used as
 * step size for recording run-time approximation
 */
void pf_run_time (sp, offset_sp)
  search_ *sp, *offset_sp;
{ FILE *fp;
  float accu;

  if (is_in_transient_phase(sp)) return;
  eval_rt_approx(sp);

  if (cmd.xon == YES) {
    accu = approx2accu (sp->approx.run_time);
    if (cmd.pf.last_accu - accu >= cmd.pf.step) {
      cmd.pf.last_accu = accu;
      if (accu != 0.0) {
	fp = fopen(io.rt_pf, "a");
        accu = get_offset_rt_approx(sp, offset_sp);
	accu = approx2accu (accu);
	fprintf(fp, "%g  %d  %g  %d\n",
              accu,
              get_offset_time(sp, offset_sp),
              get_offset_cst(sp, offset_sp),
              get_offset_gen(sp, offset_sp));
	fclose(fp);
      }
    }
  } else {
    if ((cmd.pf.last_approx - sp->approx.run_time) >= cmd.pf.step) {
      cmd.pf.last_approx = sp->approx.run_time;
      if (sp->approx.run_time != huge_float) {
	fp = fopen(io.rt_pf, "a");
	fprintf(fp, "%g  %d  %g  %d\n",
              get_offset_rt_approx(sp, offset_sp),
              get_offset_time(sp, offset_sp),
              get_offset_cst(sp, offset_sp),
              get_offset_gen(sp, offset_sp));
	fclose(fp);
      }
    }
  }
}


void pf_space_vs_time (sp, offset_sp)
  search_ *sp, *offset_sp;
{ FILE *fp;

  fp = fopen(io.st_pf, "a");
  fprintf(fp, "%d  %d  %g  %d\n",
          get_offset_time(sp, offset_sp),  get_offset_space(sp, offset_sp),
          get_offset_cst(sp, offset_sp),   get_offset_gen(sp, offset_sp));
  fclose(fp);
}


/* collect run-time profile for some search algorithms */
void alg_rt_pf (sp, rp) 
  search_ *sp; 
  regress_ *rp;
{ long time = get_time(sp);
  float a = sp->approx.run_time;
  double x, y;
 
  if (cmd.xon == YES) {
    a = approx2accu (a);
    if (time == 0 || a == 0.0) return;
  } else {
    if (time == 0 || a == huge_float) return;
  }
  x = log10(todouble(time));
  y = todouble(a);
  rp->n += 1.0;
  rp->x_sum += x;
  rp->x2_sum += (x * x);
  rp->y_sum += y;
  rp->xy_sum += (x * y);
}


/* collect execution profile (search by search) for some search algorithms */
void alg_pf (sp, rp)
  search_ *sp;
  regress_ *rp;
{ long time = get_time(sp);
  float a = get_approx(sp);
  double x, y;
 
  if (cmd.xon == YES) {
    a = approx2accu (a);
    if (time == 0 || a == 0.0) return;
  } else {
    if (time == 0 || a == huge_float) return;
  }
  x = log10(todouble(time));  y = todouble(a);
#ifdef DEBUG
if (cmd.debug >= 1) printf("doing alg pf over searches\n");
#endif
  rp->n += 1.0;
  rp->x_sum += x;
  rp->x2_sum += (x * x);
  rp->y_sum += y;
  rp->xy_sum += (x * y);
}


void alg_pf_predict (sp, rp)
  search_ *sp;
  regress_ *rp;
{ double det;
  float tmp;
 
  det = rp->n * rp->x2_sum - rp->x_sum * rp->x_sum;
  rp->beta[0] = (rp->x2_sum * rp->y_sum - rp->x_sum * rp->xy_sum) / det;
  rp->beta[1] = (- rp->x_sum * rp->y_sum + rp->n * rp->xy_sum) / det;
  tmp = (float) (rp->beta[0] + rp->beta[1] * log10(todouble(sp->constr.time)));
#ifdef DEBUG
if (cmd.debug >= 1) {
  printf("n=%f, x2_sum=%f, x_sum=%f\n", rp->n, rp->x2_sum, rp->x_sum);
  printf("det=%f, beta0=%f, beta1=%f\n", det, rp->beta[0], rp->beta[1]);
  printf("old pred=%g,  new pred=%f\n", sp->approx.predicted, tofloat(tmp));
}
#endif
  
  if (cmd.xon == YES) tmp = accu2approx (tmp);
  sp->approx.predicted = tmp;
  if (sp->approx.predicted < 0.0) sp->approx.predicted = 0.0;
}

