
#define GetPE	((search_ **) malloc (cmd.num_pe * sizeof (search_ *)))

/* the search starter per iteration,
 * the stat structure will be initialized, since it is entirely a new search.
 */
void search (sp)
  search_ *sp; 
{
  switch (sp->algorithm)
  {
    case DEFAULT: default_algorithm (sp);		break;
    case LW:	lawler_wood_algorithm (sp);		break;
    case sTCA:	stca_algorithm (sp);			break;
    case pTCA:	ptca_algorithm (sp);			break;
    case dTCA:	dtca_algorithm (sp);			break;
    case sTCGD:	stcgd_algorithm (sp);			break;
    case pTCGD:	ptcgd_algorithm (sp);			break;
    case MIDA:	mida_algorithm (sp);			break;
    case IDA_CR: ida_cr_algorithm (sp);			break;
    case NEW_IDA_CR: new_ida_cr_algorithm (sp);		break;
    case uRTS:	urts_algorithm (sp);			break;
    case bRTS:	brts_algorithm (sp);			break;
    case hRTS:	hrts_algorithm (sp);			break;
    case uIRD:	uird_algorithm (sp);			break;
    case bIRD:	bird_algorithm (sp);			break;
    case IRA:	ira_algorithm (sp);			break;
    case IDA:	ida_algorithm (sp);			break;
    case DFS_star:	ida_algorithm (sp);		break;
    case iFBS:	ifbs_algorithm (sp);			break;
    case gDFS_iFBS:	gdfs_ifbs_algorithm (sp);	break;
    case PBFS:  para_algorithm (sp, GetPE, cmd.num_pe);	break;
    case PGDFS: para_algorithm (sp, GetPE, cmd.num_pe);	break;
    case PBAND: para_algorithm (sp, GetPE, cmd.num_pe);	break;
    default:	error ("search: no such algorithm\n");	break;
  }
}
#undef GetPE


search_ *create_brother_search (sp)
  search_ *sp;
{ search_ *new_sp;

  new_sp = (search_ *) malloc(sizeof(search_));
  init_search_struct(new_sp, 1, 1);
  new_sp->strategy = sp->strategy;
  new_sp->algorithm = sp->algorithm;

  new_sp->brother = sp;
  new_sp->parent = sp->parent;

  attach_alg_struct(new_sp);
  return new_sp;
}


search_ *create_child_search (sp)
  search_ *sp;
{ search_ *new_sp;

  new_sp = (search_ *) malloc(sizeof(search_));
  init_search_struct(new_sp, 1, 1);
  new_sp->strategy = sp->strategy;
  new_sp->algorithm = sp->algorithm;

  new_sp->brother = sp->child;
  new_sp->parent = sp;
  sp->child = new_sp;

  attach_alg_struct(new_sp);
  return new_sp;
}


/* some book-keeping for certain algorithms */
void alg_entry (sp, node)
  search_ *sp;
  node_ *node;
{
  switch (sp->algorithm) {
    case dTCA:	dtca_entry(sp);		break;
    case MIDA:	mida_bin_update (sp, node); break;
    case uRTS:	urts_entry (sp, node);	break;
/**
    case bRTS:	brts_entry (sp, node);	break;
    case hRTS:	hrts_entry (sp, node);	break;
**/
    default:    				break;
  }
}


void alg_entry_aux (sp, child)
  search_ *sp;
  node_ *child;
{
  switch(sp->algorithm) {
    case MIDA:   mida_entry (sp, child); break;
    case IDA_CR: ida_cr_bin_update (sp, child); break;
    case NEW_IDA_CR: new_ida_cr_bin_update (sp, child); break;
    case uRTS:	urts_entry_aux (sp, child);	break;
    case bRTS:	brts_entry_aux (sp, child);	break;
    case hRTS:	hrts_entry_aux (sp, child);	break;
    default:                    break;
  }
}


#define	is_alg_skip_pruning	(RTS_IRD_iter == 1 && sp->stat.feasible <= 0 && child->lowb < child->upb)

int is_bounded (child, sp)
  node_ *child;
  search_ *sp;
{
#if defined (RTS_ENHANCE) || defined (IRD_ENHANCE)
  if (is_alg_skip_pruning) return 0;
#endif
  return (child->lowb >= expected_opt(sp) ? 1 : 0);
}


int is_hard_bounded (child, sp)
  node_ *child;
  search_ *sp;
{
#if defined (RTS_ENHANCE) || defined (IRD_ENHANCE)
  if (is_alg_skip_pruning) return 0;
#endif
  if (No_Upper_Bound == YES && cmd.ens == NO)
    return (child->lowb > sp->constr.bound ? 1 : 0);
  else return (child->lowb >= sp->constr.bound ? 1 : 0);
}

#undef is_alg_skip_pruning


void dominating (node, sp) node_ *node; search_ *sp; { }


void bounding (sp)
  search_ *sp;
{ int i;
  void young_bounding ();

  switch (sp->strategy) {
    case BFS:  bptree_bounding(sp); break;
    case GDFS: list_bounding(sp);  break;
    case BAND: btree_bounding(sp);
	       for (i = sp->band.ceiling_depth; i <= sp->band.floor_depth; i++)
		  young_bounding (sp, i);
	       break;
    default:			   break;
  }
}



void bptree_bounding (sp)
  search_ *sp;
{
#ifdef BPLUS_TREE
  node_ *p, *q;
  long free_bptree ();

  for (p = sp->open.bptree; p; ) {
    if (is_bounded (p, sp)) {
      if (p == sp->open.bptree) sp->open.bptree = p->left;
      if (p->up) p->up->right = p->left;
      if (p->left) p->left->up = p->up;

      /* re-check its left subtree */
      q = p;
      p = p->left;

      /* rm itself and its right subtree */
      q->left = NULL;
      sp->stat.active -= free_bptree (q);

    } else p = p->right;
  }
#else
  btree_bounding (sp);
#endif
}


long free_bptree (p)
  node_ *p;
{ long count = 0;

  if (p) {
    count = 1;
    if (p->left)  count += free_bptree (p->left);
    if (p->right) count += free_bptree (p->right);
    p->left = p->right = p->up = NULL;
    free_node (p);
  }

  return count;
}


void btree_bounding (sp)
  search_ *sp;
{ node_ *p, *q;
  long count = 0;

  q = p = sp->open.btree;
  while (p) {
    if (is_bounded(p, sp)) {
      count++;
      if (p == sp->open.btree) {
        sp->open.btree = p->next;
        free_node(p);
        q = p = sp->open.btree;
      }
      else {
        q->next = p->next;
        free_node(p);
        p = q->next;
      }
    }
    else {
      p = (q = p)->next;
    }
  }
  sp->stat.active -= count;
}


void list_bounding (sp)
  search_ *sp;
{ node_ *p, *q;
  long count = 0;

  q = p = sp->open.list;
  while (p) {
    if (is_bounded(p, sp)) {
      count++;
      if (p == sp->open.list) {
        sp->open.list = p->next;
        free_node(p);
        q = p = sp->open.list;
      }
      else {
        q->next = p->next;
        free_node(p);
        p = q->next;
      }
    }
    else {
      p = (q = p)->next;
    }
  }

  sp->stat.active -= count;
}


void young_bounding (sp, depth)
  search_ *sp;
  int depth;
{ node_ *p, *q;
  long count = 0;

  q = p = (sp->band.tree+depth)->young;
  while (p) {
    if (is_bounded(p, sp)) {
      count++;
      if (p == (sp->band.tree+depth)->young) {
        (sp->band.tree+depth)->young = p->next;
        free_node(p);
        q = p = (sp->band.tree+depth)->young;
      }
      else {
        q->next = p->next;
        free_node(p);
        p = q->next;
      }
    }
    else {
      p = (q = p)->next;
    }
  }

  sp->stat.active -= count;
}


void set_search_message (msgp, style, rp, alg_entry, rt_pf, st_pf)
  search_message_ *msgp;
  search_style_ style;
  regress_ *rp;
  yesno_ alg_entry, rt_pf, st_pf;
{
  msgp->style = style;
  msgp->rp = rp;
  msgp->alg_entry = alg_entry;
  msgp->rt_pf = rt_pf;
  msgp->st_pf = st_pf;
  msgp->feasible = NO;
}


is_feasible_or_equiv (node)
  node_ *node;
{
  return is_feasible (node);
}


yesno_ end_of_opt_sol_search (sp)
  search_ *sp;
{ float delta;

  if (cmd.opt_sol == NO) return NO;

  delta = (float) (sp->glub - cmd.opt_sol_val); 
  delta = (delta >= 0.0) ? delta : - delta;

  if (delta > cmd.sol_precision) return NO;
  return YES; 
}


yesno_ end_of_first_sol_search (sp)
  search_ *sp;
{
  if (cmd.first_sol == YES && sp->glub != problem.huge) return YES;
  return NO;
}
