
#define young_list(sp,depth)	(sp->band.tree+depth)->young
#define adult_count(sp,depth)	(sp->band.tree+depth)->count
#define bandwidth(sp,depth)	(sp->band.tree+depth)->width


extern node_ *btree_delete ();
void btree_insert ();

node_ *young_delete ();
void young_insert ();
#ifdef SWAP_BAND_NODE
node_ *swap_band_node ();
#endif


node_ *get_band_node (sp)
    search_ *sp;
{   node_ *node;
    tree_ *treep;
    int i, depth;
    long sum = 0;

#ifdef PEEK_ACTIVE_LIST
printf ("NEW STATISTICS\n");
for (i = 0, node = sp->open.btree; node; node = node->next) i++;
printf ("sp->open.btree=%d\n", i);
sum += i;
for (depth = sp->band.ceiling_depth; depth <= sp->band.floor_depth; depth++) {
    for (i = 0, node = young_list (sp, depth); node; node = node->next) i++;
    printf ("depth=%d, young=%d\n", depth, i);
    sum += i;
}
printf ("sp->stat.active=%d, sum=%d\n", sp->stat.active, sum);
#endif

    if (sp->stat.active <= 0) sp->stat.active = 0;
    else sp->stat.active --;

    if ((node = btree_delete (sp))) return node;

    /** no backtracking for first-band search **/
    if (sp->strategy == FBS) return NULL;

    /** update ceiling **/
    for (depth = sp->band.ceiling_depth; depth <= sp->band.floor_depth; depth++)
	if (young_list (sp, depth)) break;
    sp->band.ceiling_depth = depth;

    while (1) {
	/** try this depth, include young nodes into band **/
	treep = sp->band.tree + sp->band.floor_depth;
	treep->count = 0;
	for (i = 0; i < treep->width; i++)
	    if ((node = young_delete (sp, sp->band.floor_depth))) {
		btree_insert (node, sp);
		treep->count ++;
	    }
	    else break;

	/** check if new adult nodes exist **/
	if ((node = btree_delete (sp))) return node;

	/** unfortunately, there is no new adult node **/
	/** check young nodes less deep in the search tree **/
	sp->band.floor_depth--;
	if (sp->band.floor_depth < sp->band.ceiling_depth) {
	    sp->stat.active = 0;
	    return NULL;
	}
    }
}


void put_band_node (children, sp)
    node_ *children;
    search_ *sp;
{   node_ *child, *node, *new_node;
    tree_ *treep;
    int count, depth, num, i;

    if (sp->stat.active < 0) sp->stat.active = 0;
    if (children == NULL) return;

    depth = children->depth;
    if (depth > sp->band.floor_depth) {
	sp->band.floor_depth = depth;
	if (depth > MaxDepth) {
	    num = 2 * MaxDepth + 1;
	    treep = (tree_ *) malloc (num * sizeof (tree_));
	    for (i = 0; i <= MaxDepth; i++) {
		(treep+i)->young = (sp->band.tree+i)->young;
		(treep+i)->count = (sp->band.tree+i)->count;
		(treep+i)->width = (sp->band.tree+i)->width;
	    }
	    for (i = MaxDepth+1; i < num; i++)
		init_tree_struct (treep, i, sp->band.init_width);

	    free (sp->band.tree);
	    sp->band.tree = treep;
	    MaxDepth = (-- num);
	}
    }

    while ((child = children)) {
        sp->stat.active ++; 
	children = children->brother;
	young_insert (child, depth, sp);
    }

    count = adult_count (sp, depth);
    while (count < bandwidth (sp, depth))
	if ((node = young_delete (sp, depth))) {
	    btree_insert (node, sp);
	    count++;
	} else break;
    adult_count (sp, depth) = count;

#ifdef SWAP_BAND_NODE
    while ((node = young_delete (sp, depth))) {
	new_node = swap_band_node (node, sp);
	young_insert (new_node, depth, sp);
	if (node == new_node) break;
    }
#endif
}

#ifdef LOWB_PLUS_GCOST
#define young_sort_key(node)	(node->lowb * LPG_LARGE_CONST - node->g_cost)
#else
#define	young_sort_key(node)	(node->lowb)
#endif

/** insert a node into a b+ tree, ascendingly **/
void young_insert (node, depth, sp)
    node_ *node;
    int depth;
    search_ *sp;
{   node_ *p, *q;
    domain node_key;

    if (young_list (sp, depth)) {
	node_key = young_sort_key (node);
	for (q = p = young_list (sp, depth); p; p = (q = p)->next) {
	    if (node_key < young_sort_key(p)) {
		if (p == young_list (sp, depth)) {
		    (young_list (sp, depth) = node)->next = p;
		    return;
		} else {
		   (q->next = node)->next = p;
		   return;
		} /* if-then-else */
	    } /* if-then */
	} /* for */
	if (! p) (q->next = node)->next = NULL;
    }
    else (young_list (sp, depth) = node)->next = NULL;
}


/** delete a node from a b+ tree **/
node_ *young_delete (sp, depth)
    search_ *sp;
    int depth;
{   node_ *q;

    if (young_list (sp, depth))
	young_list (sp, depth) = (q = young_list (sp, depth))->next;
    else return NULL;

    return q;
}


int is_young_empty (sp, depth)
    search_ *sp;
    int depth;
{   return (young_list (sp, depth) == NULL); }

#undef young_list
#undef adult_count
#undef bandwidth
#undef young_sort_key

