
void	bandwidth_enlargement (), bandwidth_reduction ();
void	switch_to_band_search ();

/* return SEARCH_IS_COMPLETED, search is completed;
 * return SEARCH_IS_ABORTED, search is aborted due to constraints.
 */
search_ending_ band_primitive (sp, offset_sp, msgp)
    search_ *sp, *offset_sp;
    search_message_ *msgp;
{   node_ *node, *children, *child, *p, *temp, *temp0, *next_child, *new_list;
    int i;
    solution_ *sol;
    yesno_ bound_dom, to_update_gllb;
    float GUIDE_H ();
#ifdef HARE_GUIDANCE
    domain p_upb, p_lowb, g_cost, p_h;
#endif
#ifdef DEBUG
if (cmd.debug >= 2) printf ("\nenter band_primitive\n");
if (cmd.debug >= 1) printf ("bw_fx=%s\n", dbg_bw_fx[cmd.bw_fx]);
#endif

    times (&start_time);
    last_time = start_time;
    if (msgp->style == FRESH_START) {
	idpd_init ();
	put_band_node (create_root (sp), sp);
    }

    while ((node = get_band_node (sp))) {

#ifdef HARE_GUIDANCE
      p_upb = node->upb;
      p_lowb = node->lowb;
      g_cost = node->g_cost;
      p_h = node->lowb - g_cost;
#endif

	if (is_bounded (node, sp)) {
	    sp->stat.bounded++;
	    free_node (node);
	    continue;
	}

	to_update_gllb = (sp->gllb == node->lowb) ? YES : NO;
	eval_rt_approx (sp);
	update_stat (sp);
	if (msgp->rt_pf == YES) pf_run_time (sp, offset_sp);
	if (msgp->st_pf == YES) pf_space_vs_time (sp, offset_sp);
	if (is_constr_violated (sp)) return SEARCH_IS_ABORTED;
	if (msgp->rp) alg_rt_pf (sp, msgp->rp);
	if (msgp->alg_entry == YES) alg_entry (sp, node);

#ifdef DEBUG
if (cmd.debug >= 2) debug_node ("\nexpand", node, sp);
#endif

	/** expand will return a list of children **/
	if ((children = expand (node, ALL_CHILDREN, DONT_CARE)))
	    sp->stat.expanded++;
	free_node (node);

	/** bandwidth adaptions **/
	switch (cmd.bw_fx) {
	case DYNAMIC_BW:
	case GDFS_dBS_BW:
		if (children == NULL) bandwidth_enlargement (sp, node->depth);
		break;
	default: break;
	}

#if defined (LOWB_GUIDANCE) || defined (UPB_GUIDANCE) || defined (UGDFS) || defined (HARE_GUIDANCE)
#define BOUND_GUIDANCE
#endif

	new_list = NULL;
	for (child = children; child; child = next_child) {
	    sp->stat.generated++;
	    next_child = child->brother;

	    bound_dom = YES;	/** init skip **/

	    if (is_infeasible (child))  {
#ifdef DEBUG
if (cmd.debug >= 3) printf ("infeasible\n");
#endif
		sp->stat.infeasible++;
		free_node (child);
		continue;
	    }

#ifdef DEBUG
if (cmd.debug >= 3) printf ("eval bounds\n");
#endif

	    evaluate_lower_bound (child);
	    sol = evaluate_upper_bound (child, get_sol_buf ());
	    if (msgp->alg_entry == YES) alg_entry_aux (sp, child);

#ifdef DEBUG
	    if (cmd.debug >= 3) debug_node (NULL, child, sp);
#endif

	    if (child->upb < sp->glub) {
#ifdef DEBUG
if (cmd.debug >= 3) printf ("update incumbent\n");
#endif
		sp->glub = child->upb;
		eval_rt_approx (sp);
		free_sol_buf (sp->incumbent);
		sp->incumbent = sol;
		bounding (sp);
		dominating (child, sp);
		bound_dom = NO;	/** skip over  bounded & dominated  tests **/

		if (_is_first_sol_found_ == NO) {
		    _is_first_sol_found_ = YES;
		    switch_to_band_search (sp);
		}

		/** first-solution search **/
		if (cmd.first_sol == YES) return SEARCH_IS_COMPLETED;

		/** optimal-solution search **/
		if (end_of_opt_sol_search (sp) == YES)
		    return SEARCH_IS_COMPLETED;

		/** bandwidth adaptions **/
		switch (cmd.bw_fx) {
		case DYNAMIC_BW:
		case GDFS_dBS_BW:
			bandwidth_reduction (sp, node->depth);
			break;
		default: break;
		}
	    }
	    else free_sol_buf (sol);

	    if (is_feasible_or_equiv (child)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf ("feasible\n");
#endif
		sp->stat.feasible++;
		free_node (child);
		continue;
	    }

	    if (is_bounded (child, sp)) { 
#ifdef DEBUG
if (cmd.debug >= 3) printf ("bounded\n");
#endif
		sp->stat.bounded++;
		free_node (child);
		continue;
	    }

	    if (bound_dom == YES) {
		if (is_dominated (child, sp)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf ("dominated\n");
#endif
		    sp->stat.dominated++;
		    free_node (child);
		    continue;
		}
	    }

	    if (is_hard_bounded (child, sp)) {
#ifdef DEBUG    
if (cmd.debug >= 3) printf ("bounded\n");     
#endif	      
		if (Next_IDA_Threshold > child->lowb)
		    Next_IDA_Threshold = child->lowb;
		sp->stat.hard_bounded++;
		free_node(child); 
		continue;
	    }

	    /** by non-leaf, lousy-upper-bound node **/
	    if (bound_dom == YES) dominating (child, sp);

	    child->next = new_list;
	    new_list = child;

	} /** children loop **/

	children = new_list;
	new_list = NULL;

#ifdef BOUND_GUIDANCE
	/** sort these live children by fancy guidance **/
	for (child = children; child; child = child->next) {
	    if (new_list) {
		for (temp0 = temp = new_list;
		     temp;
		     temp = (temp0 = temp)->brother) {

#if defined (HARE_GUIDANCE)
          if (GUIDE_H ( child->upb,
                        child->lowb,
                        child->g_cost,
                        child->lowb - child->g_cost,
                        ((domain) child->depth),
                        p_upb - child->upb,
                        child->lowb - p_lowb,
                        child->g_cost - g_cost,
                        p_h - (child->lowb - child->g_cost))
              <
              GUIDE_H ( temp->upb,
                        temp->lowb,
                        temp->g_cost,
                        temp->lowb - temp->g_cost,
                        ((domain) temp->depth),
                        p_upb - temp->upb,
                        temp->lowb - p_lowb,
                        temp->g_cost - g_cost,
                        p_h - (temp->lowb - temp->g_cost))
              ) {
#elif defined (UPB_GUIDANCE) || defined (UGDFS)
	            if (child->upb < temp->upb) {
#elif defined (LOWB_GUIDANCE)
		    if (child->lowb < temp->lowb) {
#else /** bad guidance **/
		    if (1) {
#endif
			if (temp == new_list) (new_list = child)->brother = temp;
			else (temp0->brother = child)->brother = temp;
			break;
		    } 
		}
		if (! temp) (temp0->brother = child)->brother = NULL;
	    }
	    else (new_list = child)->brother = NULL;
	}
	children = new_list;

#else DEFAULT_GUIDANCE
	for (child = children; child; child = child->next) {
	    child->brother = new_list;
	    new_list = child;
	}
	children = new_list;

#endif BOUND_GUIDANCE


	for (child = children; child; child = child->brother) 
	    child->next = NULL;

	put_band_node (children, sp);

#ifndef NO_UPDATE_GLB
	if (to_update_gllb == YES) {
	    /** set init value for global_lowb **/
	    sp->gllb = sp->glub;
	    /** check band priority list **/
	    if (sp->open.btree)
	        /** global lowb == min { lowb of all active nodes } **/
		for (p = sp->open.btree; p; p = p->next)
		    if (p->lowb < sp->gllb) sp->gllb = p->lowb;
	    /** check young priority lists **/
	    for (i = sp->band.ceiling_depth; i <= sp->band.floor_depth; i++)
		if ((sp->band.tree+i)->young)
	        /** global lowb == min { lowb of all active nodes } **/
		    for (p = (sp->band.tree+i)->young; p; p = p->next)
			if (p->lowb < sp->gllb) sp->gllb = p->lowb;
	}
#endif

    } /** selection loop **/

    /** When the search is completed, the global lowb and the global upb
     ** should be the same, regardless of approximation is used or not.
     ** However, it is not the case, sometimes.
     ** One possible reason is that
     ** global lowb is supposed to be updated by a feasible sol; however,
     ** the feasible sol is freed during the children loop.
     ** Therefore, both are coersed to be equal.
     **/
    if (sp->approx.approx == 0.0) sp->gllb = sp->glub;
    sp->approx.run_time = 0.0;

    times (&break_time);
    sp->stat.unix_utime += (break_time.tms_utime - start_time.tms_utime);
    sp->stat.unix_stime += (break_time.tms_stime - start_time.tms_stime);

    if (sp->constr.bound != problem.huge)
	if (sp->glub <= sp->constr.bound) msgp->feasible = YES;

    return SEARCH_IS_COMPLETED;
} /** band primitive **/


int bandwidth_function (depth, init_width)
    int depth, init_width;
{   int width = init_width;
    static float a, b;
    static int first_entry = 1;

    switch (cmd.bw_fx) {
    case GDFS_dBS_BW:
    case GDFS_fBS_BW:
		if (_is_first_sol_found_ == NO) return 1;
		break;
    default:	break;
    }

    if (first_entry) {
	first_entry = 0;
	switch (cmd.bw_fx) {
	case LINEAR_BW:
		a = ((float) (1 - init_width)) / (problem.size - 1);
		b = init_width - a;
		break;

	case EXP_BW:
		a = ((float) (log ((double) init_width))) / (problem.size - 1);
		break;
	default: break;
	}
    }

    switch (cmd.bw_fx) {
    case LINEAR_BW:
		width = a * depth + b;
	 	break;

    case EXP_BW:
		width = init_width * exp ((double) (a * (1 - depth)));
		break;

    case FIXED_BW:
    default:	break;
    }

    if (width < 1) width = 1;
    return width;
}


void bandwidth_enlargement (sp, depth)
    search_ *sp;
    int depth;
{   int i;
    tree_ *treep;
    node_ *node;

    switch (cmd.bw_fx) {
    case GDFS_dBS_BW:
    case GDFS_fBS_BW:
		if (_is_first_sol_found_ == NO) return;
		break;
    default:	break;
    }

#ifdef MONITOR_BW_ADAPTION
printf ("Bandwidth enlargement => depth=%d, ", depth);
#endif

    for (i = sp->band.ceiling_depth; i <= depth; i++) {
	/** increment width by one if it is less than a threshold **/
	treep = sp->band.tree + i;

	if (treep->width < sp->band.bw_bound) {
	    treep->width ++;

	    /** move nodes, if necessary **/
	    for ( ; treep->count < treep->width; treep->count ++)
		if ((node = young_delete (sp, i))) btree_insert (node, sp);
		else break;
#ifdef MONITOR_BW_ADAPTION
printf ("TAKEN\n");
#endif
	}
#ifdef MONITOR_BW_ADAPTION
	else printf ("NOT_TAKEN\n");
#endif
    }
}


void bandwidth_reduction (sp, depth)
    search_ *sp;
    int depth;
{   int i;
    tree_ *treep;

    switch (cmd.bw_fx) {
    case GDFS_dBS_BW:
		if (_is_first_sol_found_ == NO) return;
		break;
    default:	return; break;
    }

#ifdef MONITOR_BW_ADAPTION
printf ("Bandwidth reduction => depth=%d\n", depth);
#endif

    for (i = sp->band.ceiling_depth; i <= depth; i++) {
	/** decrement width by one **/
	treep = sp->band.tree + i;
	treep->width --;
	if (treep->width < 1) treep->width = 1;
    }
}


void switch_to_band_search (sp)
    search_ *sp;
{   int depth, num, init_width;
    tree_ *treep;

    switch (cmd.bw_fx) {
    case GDFS_fBS_BW:
		treep = sp->band.tree;
		init_width = sp->band.init_width;
		num = MaxDepth + 1;

		for (depth = 0; depth < num; depth++)
		(treep+depth)->width = bandwidth_function (depth, init_width);
		break;
    default: break;
    }
}

