
/* return SEARCH_IS_COMPLETED, search is completed;
 * return SEARCH_IS_ABORTED, search is aborted due to constraints.
 */
search_ending_ first_primitive (sp, offset_sp, msgp, num_active)
  search_ *sp, *offset_sp;
  search_message_ *msgp;
  int num_active;
{ node_ *node, *children, *child, *temp;
  solution_ *sol;
  yesno_ bound_dom;
  long count;
#ifdef DEBUG
if (cmd.debug >= 2) printf("\nenter first_primitive\n");
#endif

  times(&start_time);
  last_time = start_time;
  if (msgp->style == FRESH_START) {
    idpd_init();
    btree_insert(create_root(sp), sp);
  }
  
  while (1) {
    for (count = 0, temp = sp->open.btree; temp; temp = temp->next)
      if (temp->type == ACTIVE_SINGLE) count++;
    if (count >= num_active) return SEARCH_IS_COMPLETED;
    if (! (node = btree_delete(sp))) break;
    sp->gllb = node->lowb; 
    update_stat(sp);
    if (is_constr_violated(sp)) return SEARCH_IS_ABORTED;
    /* expand will return a list of children */
    if ((children = expand(node, ALL_CHILDREN, DONT_CARE))) sp->stat.expanded++;
    free_node(node);

    for (child = children; child; child = child->brother) {
      sp->stat.generated++;

      if (is_infeasible(child))  {
        sp->stat.infeasible++;
        free_node(child);
        continue;
      }

      evaluate_lower_bound(child);
      sol = evaluate_upper_bound(child, get_sol_buf());

      bound_dom = YES;	/* init skip */

      if (child->upb < sp->glub) {
        sp->glub = child->upb;
	eval_rt_approx(sp);
        free_sol_buf(sp->incumbent);
        sp->incumbent = sol;
        bounding(sp);
        dominating(child, sp);
        bound_dom = NO;	/* skip over  bounded & dominated  tests */

	_is_first_sol_found_ = YES;

	/** first-solution search **/
	if (cmd.first_sol == YES) return SEARCH_IS_COMPLETED;

	/** optimal-solution search **/
	if (end_of_opt_sol_search (sp) == YES)
	    return SEARCH_IS_COMPLETED;

      }
      else free_sol_buf(sol);

      if (is_feasible_or_equiv(child)) {
        sp->stat.feasible++;
        free_node(child);
        continue;
      }

      if (is_bounded(child, sp)) { 
        sp->stat.bounded++;
        free_node(child);
        continue;
      }

      if (bound_dom == YES) {
        if (is_dominated(child, sp)) {
          sp->stat.dominated++;
          free_node(child);
          continue;
        }
      }

      if (is_hard_bounded(child, sp)) {
        sp->stat.hard_bounded++;
        free_node(child); 
        continue;
      }

      if (bound_dom == YES)
        dominating(child, sp);	/* by non-leaf, lousy-upper-bound node */

      btree_insert(child, sp);
    } /* children loop */
  } /* selection loop */

  times(&break_time);
  sp->stat.unix_utime += (break_time.tms_utime - start_time.tms_utime);
  sp->stat.unix_stime += (break_time.tms_stime - start_time.tms_stime);

  return SEARCH_IS_COMPLETED;
} /* first primitive */
