
/* return SEARCH_IS_COMPLETED, search is completed;
 * return SEARCH_IS_ABORTED, search is aborted due to constraints.
 */
search_ending_ path_primitive (sp, offset_sp, msgp)
  search_ *sp, *offset_sp;
  search_message_ *msgp;
{ node_ *node, *children, *child, *p, *best, *next_child;
  solution_ *sol, *best_sol;
  yesno_ bound_dom;
#ifdef DEBUG
if (cmd.debug >= 2) printf("\nenter path_primitive\n");
#endif

  times(&start_time);  last_time = start_time;
  for (node = list_delete(sp); node; node = list_delete(sp)) {
#ifdef DEBUG
if (cmd.debug >= 2) debug_node("\nexpand", node, sp);
#endif
    /* expand will return a list of children */
    children = expand(node, ALL_CHILDREN, DONT_CARE);
    free_node(node);

    best = NULL; best_sol = NULL;
    for (child = children; child; child = next_child) {
      next_child = child->brother;
      if (is_infeasible(child))  { free_node(child); continue; }
      if (child->type == ACTIVE_SINGLE) {

        evaluate_lower_bound(child);
        sol = evaluate_upper_bound(child, get_sol_buf());

        if (best) {
          if (child->lowb < best->lowb) {
            best = child;
            best_sol = sol;
            free_node (best);
          } else free_node (child);
        } else {
          best = child;
          best_sol = sol;
        }
      } else free_node (child);

    }
    sol = best_sol;
    if (! (child = best)) continue;

    if (child->upb < sp->glub) {
      sp->glub = child->upb;
      eval_rt_approx(sp);
      free_sol_buf(sp->incumbent);
      sp->incumbent = sol;

      _is_first_sol_found_ = YES;

      /** first-solution search **/
      if (cmd.first_sol == YES) return SEARCH_IS_COMPLETED;

      /** optimal-solution search **/
      if (end_of_opt_sol_search (sp) == YES)
	return SEARCH_IS_COMPLETED;

    } else free_sol_buf(sol);

    if (is_feasible_or_equiv(child)) {
      sp->stat.feasible++;
      free_node(child);
      return SEARCH_IS_COMPLETED;
    }

    list_insert(child, sp);
  } /* selection loop */

  times(&break_time);
  sp->stat.unix_utime += (break_time.tms_utime - start_time.tms_utime);
  sp->stat.unix_stime += (break_time.tms_stime - start_time.tms_stime);
  return SEARCH_IS_ABORTED;
} /* path primitive */
