
/* return SEARCH_IS_COMPLETED, search is completed;
 * return SEARCH_IS_ABORTED, search is aborted due to constraints;
 * return SEARCH_IS_IDLE, search process is idle.
 */
search_ending_ pbfs_primitive (sp, offset_sp, msgp, gllb, ngenp)
  search_ *sp, *offset_sp;
  search_message_ *msgp;
  domain gllb;
  long *ngenp;
{ node_ *node, *children, *child;
  solution_ *sol;
  yesno_ bound_dom;
#ifdef DEBUG
if (cmd.debug >= 2) printf("\nenter pbfs_primitive\n");
#endif

  *ngenp = 0;
  times(&start_time);
  last_time = start_time;

  if ((node = btree_delete(sp))) {

    if (is_bounded(node, sp)) {
      sp->stat.bounded++;
      free_node(node);
      goto pbfs_exit;
    }

    sp->gllb = (gllb < node->lowb) ? gllb : node->lowb; 
    eval_rt_approx(sp);
    update_stat(sp);
    if (msgp->rt_pf == YES) pf_run_time(sp, offset_sp);
    if (msgp->st_pf == YES) pf_space_vs_time(sp, offset_sp);
    if (is_constr_violated(sp)) return SEARCH_IS_ABORTED;
    if (msgp->rp) alg_rt_pf(sp, msgp->rp);
    if (msgp->alg_entry == YES) alg_entry(sp, node);
#ifdef DEBUG
if (cmd.debug >= 2) debug_node("\nexpand", node, sp);
#endif
    /* expand will return a list of children */
    if ((children = expand(node, ALL_CHILDREN, DONT_CARE)))
      sp->stat.expanded++;
    free_node(node);

    for (child = children; child; child = child->brother) {
      sp->stat.generated++;
      (*ngenp)++;

      if (is_infeasible(child))  {
#ifdef DEBUG
if (cmd.debug >= 3) printf("infeasible\n");
#endif
        sp->stat.infeasible++;
        free_node(child);
        continue;
      }

#ifdef DEBUG
if (cmd.debug >= 3) printf("eval bounds\n");
#endif

      evaluate_lower_bound(child);
      sol = evaluate_upper_bound(child, get_sol_buf());
      if (msgp->alg_entry == YES) alg_entry_aux (sp, child);

#ifdef DEBUG
if (cmd.debug >= 3) debug_node(NIL, child, sp);
#endif
      bound_dom = YES;	/* init skip */

      if (child->upb < sp->glub) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("update incumbent\n");
#endif
        sp->glub = child->upb;
	eval_rt_approx(sp);
        free_sol_buf(sp->incumbent);
        sp->incumbent = sol;
        bounding(sp);
        dominating(child, sp);
        bound_dom = NO;	/* skip over  bounded & dominated  tests */
      }
      else free_sol_buf(sol);

      if (is_feasible_or_equiv(child)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("feasible\n");
#endif
        sp->stat.feasible++;
        free_node(child);
        continue;
      }

      if (is_bounded(child, sp)) { 
#ifdef DEBUG
if (cmd.debug >= 3) printf("bounded\n");
#endif
          sp->stat.bounded++;
          free_node(child);
          continue;
      }

      if (bound_dom == YES) {
        if (is_dominated(child, sp)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("dominated\n");
#endif
          sp->stat.dominated++;
          free_node(child);
          continue;
        }
      }

      if (is_hard_bounded(child, sp)) {
#ifdef DEBUG  
if (cmd.debug >= 3) printf("bounded\n");   
#endif        
        sp->stat.hard_bounded++;
        free_node(child); 
        continue;
      }

      if (bound_dom == YES)
        dominating(child, sp);	/* by non-leaf, lousy-upper-bound node */

      btree_insert(child, sp);
    } /* children loop */
  } else {
    return SEARCH_IS_IDLE;
  }

pbfs_exit:

  times(&break_time);
  sp->stat.unix_utime += (break_time.tms_utime - start_time.tms_utime);
  sp->stat.unix_stime += (break_time.tms_stime - start_time.tms_stime);

  return SEARCH_IS_COMPLETED;
} /* pbfs primitive */

