
/* return SEARCH_IS_COMPLETED, search is completed;
 * return SEARCH_IS_ABORTED, search is aborted due to constraints,
 * return SEARCH_IS_IDLE, search process is idle.
 */
search_ending_ pgdfs_primitive (sp, offset_sp, msgp, gllb, ngenp)
  search_ *sp, *offset_sp;
  search_message_ *msgp;
  domain gllb;
  long *ngenp;
{ node_ *node, *children, *child, *p, *temp, *temp0, *next_child, *new_list;
  solution_ *sol;
  yesno_ bound_dom;
#ifdef HARE_GUIDANCE
    domain p_upb, p_lowb, g_cost, p_h;
#endif

#ifdef DEBUG
if (cmd.debug >= 2) printf("\nenter pgdfs_primitive\n");
#endif

  *ngenp = 0;
  times(&start_time);
  last_time = start_time;

  if ((node = list_delete(sp))) {

#ifdef HARE_GUIDANCE
      p_upb = node->upb;
      p_lowb = node->lowb;
      g_cost = node->g_cost;
      p_h = node->lowb - g_cost;
#endif

    if (is_bounded(node, sp)) { 
      sp->stat.bounded++;
      free_node(node);
      goto pgdfs_exit;
    }

    sp->gllb = (gllb < node->lowb) ? gllb : node->lowb;
    eval_rt_approx(sp);
    update_stat(sp);
    if (msgp->rt_pf == YES) pf_run_time(sp, offset_sp);
    if (msgp->st_pf == YES) pf_space_vs_time(sp, offset_sp);
    if (is_constr_violated(sp)) return SEARCH_IS_ABORTED;
    if (msgp->rp) alg_rt_pf(sp, msgp->rp);
    if (msgp->alg_entry == YES) alg_entry(sp, node);
#ifdef DEBUG
if (cmd.debug >= 2) debug_node("\nexpand", node, sp);
#endif
    /* expand will return a list of children */
    if ((children = expand(node, ALL_CHILDREN, DONT_CARE)))
      sp->stat.expanded++;
    free_node(node);

#if defined (LOWB_GUIDANCE) || defined (UPB_GUIDANCE) || defined (UGDFS)
#define BOUND_GUIDANCE
#endif

    new_list = NULL;
    for (child = children; child; child = next_child) {
      sp->stat.generated++;
      (*ngenp)++;
      next_child = child->brother;

      bound_dom = YES;  /** init skip **/

      if (is_infeasible(child))  {
#ifdef DEBUG
if (cmd.debug >= 3) printf("infeasible\n");
#endif
        sp->stat.infeasible++;
        free_node(child);
        continue;
      }

#ifdef DEBUG
if (cmd.debug >= 3) printf("eval bounds\n");
#endif

      evaluate_lower_bound(child);
      sol = evaluate_upper_bound(child, get_sol_buf());
      if (msgp->alg_entry == YES) alg_entry_aux (sp, child);

#ifdef DEBUG
if (cmd.debug >= 3) debug_node(NIL, child, sp);
#endif

      if (child->upb < sp->glub) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("update incumbent\n");
#endif
	sp->glub = child->upb;
	eval_rt_approx(sp);
	free_sol_buf(sp->incumbent);
	sp->incumbent = sol;
        bounding(sp);
        dominating(child, sp);
        bound_dom = NO;	/* skip over  bounded & dominated  tests */
      }
      else free_sol_buf(sol);

      if (is_feasible_or_equiv(child)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("feasible\n");
#endif
        sp->stat.feasible++;
        free_node(child);
        continue;
      }

      if (is_bounded(child, sp)) { 
#ifdef DEBUG
if (cmd.debug >= 3) printf("bounded\n");
#endif
        sp->stat.bounded++;
        free_node(child);
        continue;
      }

      if (bound_dom == YES) {
        if (is_dominated(child, sp)) {
#ifdef DEBUG
if (cmd.debug >= 3) printf("dominated\n");
#endif
          sp->stat.dominated++;
          free_node(child);
          continue;
        }
      }

      if (is_hard_bounded(child, sp)) {
#ifdef DEBUG  
if (cmd.debug >= 3) printf("bounded\n");   
#endif        
        sp->stat.hard_bounded++;
        free_node(child); 
        continue;
      }

      /* by non-leaf, lousy-upper-bound node */
      if (bound_dom == YES) dominating(child, sp);

      child->next = new_list;
      new_list = child;
 
    } /* children loop */
 
    children = new_list;
    new_list = NULL;
     
#ifdef BOUND_GUIDANCE
    /** sort these live children by fancy guidance **/
    for (child = children; child; child = child->next) {
      if (new_list) {
        for (temp0 = temp = new_list; temp; temp = (temp0 = temp)->brother) {
#if defined (HARE_GUIDANCE)
          if (GUIDE_H ( child->upb,
                        child->lowb,
                        child->g_cost,
                        child->lowb - child->g_cost,
                        ((domain) child->depth),
                        p_upb - child->upb,
                        child->lowb - p_lowb,
                        child->g_cost - g_cost,
                        p_h - (child->lowb - child->g_cost))
              <
              GUIDE_H ( temp->upb,
                        temp->lowb,
                        temp->g_cost,
                        temp->lowb - temp->g_cost,
                        ((domain) temp->depth),
                        p_upb - temp->upb,
                        temp->lowb - p_lowb,
                        temp->g_cost - g_cost,
                        p_h - (temp->lowb - temp->g_cost))
              ) {
#elif defined (UPB_GUIDANCE) || defined (UGDFS)
          if (child->upb < temp->upb) {
#elif defined (LOWB_GUIDANCE)
          if (child->lowb < temp->lowb) {
#else /** bad guidance **/
          if (1) {
#endif
            if (temp == new_list) (new_list = child)->brother = temp;
            else (temp0->brother = child)->brother = temp;
            break;
          }
        }
        if (! temp) (temp0->brother = child)->brother = NULL;
      }
      else (new_list = child)->brother = NULL;
    }
    children = new_list;
 
#else DEFAULT_GUIDANCE
    for (child = children; child; child = child->next) {
      child->brother = new_list;
      new_list = child;
    }
    children = new_list;
 
#endif BOUND_GUIDANCE

    for (child = children; child; child = child->brother) {
      child->next = NULL;
      list_insert(child, sp);
    }

  } else {

    return SEARCH_IS_IDLE;
  }

pgdfs_exit:

  times(&break_time);
  sp->stat.unix_utime += (break_time.tms_utime - start_time.tms_utime);
  sp->stat.unix_stime += (break_time.tms_stime - start_time.tms_stime);

  if (sp->constr.bound != problem.huge)
    if (sp->glub <= sp->constr.bound) msgp->feasible = YES;

  return SEARCH_IS_COMPLETED;
} /* pgdfs primitive */

