domain ats_eval_lowb (), ats_eval_upb ();


/**  Lower bound is calculated by spanning-tree heuristic **/

void evaluate_lower_bound (node)
  node_ *node;
{
#ifdef DEBUG
if (cmd.debug >= 4) printf("eval lowb\n");
#endif

  node->lowb = (is_pre_goal (node)) ? node->g_cost : ats_eval_lowb (node);
}


domain ats_eval_lowb (node)
  node_ *node;
{ domain cost = node->g_cost;
  int num_live_cities = nCities - node->nvisit;
  int num_roads = node->nvisit - 1;	/* visited roads */
  int entity = node->entity;
  int i, j, road, x, y, s1, s2, the_first_not_done, the_last_not_done;
  int vertex[MaxProblemSize];

  /* find which cities not visited yet */
  trace_live(node);
  /** LiveCity[i] = 1 ==> live; otherwise, 0. **/
  LiveCity[entity] = 1;
  num_live_cities++;
  for (i = 0; i < nCities; i++) vertex[i] = LiveCity[i] ? i : -1;

  /* find which roads have been traversed and put into VisitedRoad */
  trace_road(node);

#ifdef DEBUG
if (cmd.debug >= 5) {
  printf("live cities=%d\n", num_live_cities);
  for (i = 0; i < nCities; i++) printf("%d-th city=%d\n", i, LiveCity[i]);
  printf("num_roads=%d\n", num_roads);
  for (i = 0; i < num_roads; i++) printf("%d-th road=%d\n", i, VisitedRoad[i]);
}
#endif

  /* spanning-tree cost */
  for (i = 0; (i < nRoads) && (num_live_cities > 0); i++) {
    road = sortW(i);
    if (! is_member(road, VisitedRoad, 0, num_roads)) {
      x = decodeX(road);
      y = decodeY(road);
      s1 = vertex[x];
      s2 = vertex[y];
      if ((s1 != -1) && (s2 != -1) && (s1 != s2)) {
	num_live_cities--;
        cost += decodeW(road);
	for (j = 0; j < nCities; j++)
	  if (vertex[j] == s2) vertex[j] = s1;
      }
    }
  }

  /** cost of connecting spanning tree and the tree explored already **/
  for (i = 0; i < nRoads; i++) {
    road = sortW (i);
    if (! is_member (road, VisitedRoad, 0, num_roads)) {
      x = decodeX (road);
      y = decodeY (road);
      if (x == 0)
        if (LiveCity[y]) { cost += decodeW (road); break; }
        else if (y == 0)
          if (LiveCity[x]) { cost += decodeW (road); break; }
    }
  }

  return cost;
}


/**  Upper bound is calculated by the hill-climbing method if applicable,
 ** otherwise, a greedy back-tracking method is applied. **/

solution_ *evaluate_upper_bound (node, sol)
  node_ *node;
  solution_ *sol;
{ int i, num_cities = node->nvisit;

#ifdef DEBUG
if (cmd.debug >= 4) printf("eval upb\n");
#endif

  if (is_feasible(node)) {
    if (num_cities != nCities) error("evaluate_upper_bound");
    for (i = 0; i < num_cities; i++) sol->city[i] = *(node->visited + i);
    node->upb = node->g_cost;
  }
#ifdef NO_UPPER_BOUND
  else node->upb = problem.huge;
#else
  else node->upb = ats_eval_upb(node, sol);
#endif
  return sol;
}


domain ats_eval_upb (node, sol)
  node_ *node;
  solution_ *sol;
{ int x, y, who, i, num_cities = node->nvisit;
  domain cost = node->g_cost, temp, temp0, best, bk_find_sol();

  for (i = 0; i < num_cities; i++) sol->city[i] = *(node->visited + i);
#ifdef DEBUG
if (cmd.debug >= 5)
  for (i = 0; i < num_cities; i++) printf("i=%d, city=%d\n", i, sol->city[i]);
#endif

  trace_live(node);

  x = node->entity;
  while (1) {
    /* find the best next city */
    best = HUGE_FLOAT;
    who = -1;
    for (y = 1; y < nCities; y++) {
      if (LiveCity[y] && (temp = Weight(x,y)) != problem.huge) {
	if (num_cities == nCities - 1)
	  if ((temp0 = Weight(y,0)) != problem.huge) temp += temp0;
	  else continue;
	if (best > temp) { best = temp; who = y; }
      }
    }
#ifdef DEBUG
if (cmd.debug >= 5) printf("num_cities=%d, who=%d\n", num_cities, who);
#endif
    /* record the best next city */
    if (who == -1) {
      /** OLD VERSION, NOT APPROPRIATE.
       ** if (node == ROOT) return bk_find_sol(node, sol);
       ** else return problem.huge; **/
      return problem.huge;
    }
    sol->city[num_cities++] = who;
    if (num_cities == nCities) return (cost + best);
    LiveCity[who] = 0;
    cost += best;
    x = who;
  }
}


domain bk_find_sol (node, sol)
  node_ *node;
  solution_ *sol;
{ int num_cities = node->nvisit;
  int x = node->entity;
  int next = 1, y, i;
  domain cost = node->g_cost;
  domain temp, temp0;

#ifdef DEBUG
if (cmd.debug >= 4) printf("enter bk_find_sol\n");
if (cmd.debug >= 5)
  for (i = 0; i < node->nvisit; i++)
    printf("i=%d, city=%d\n", i, sol->city[i]);
#endif

  trace_live(node);

  while (num_cities < nCities) {
    for (y = next; y < nCities; y++) {
      if (LiveCity[y] && (temp = Weight(x,y)) != problem.huge) {
	if (num_cities == nCities - 1)
	  if ((temp0 = Weight(y,0)) != problem.huge) temp += temp0;
	  else continue;
#ifdef DEBUG
if (cmd.debug >= 5)
  printf("num_cities=%d, city=%d\n", num_cities, y);
#endif
	cost += temp;
	LiveCity[y] = 0;
	sol->city[num_cities++] = (x = y);
	next = 1;
	break;
      }
    } 
    if (y >= nCities) {
      if (node->nvisit == num_cities) return problem.huge;  /* infeasible */
      /* recover to prev stage */
      y = sol->city[--num_cities];
      LiveCity[y] = 1;
      x = sol->city[num_cities - 1];
      cost -= Weight(x,y);
      next = y + 1;
    }
  }
  return cost;
}


trace_live (node)
  node_ *node;
{ int i; 

  /* reset LiveCity array */
  for (i = 0; i < nCities; i++) LiveCity[i] = 1;

  /* mark real live city */
  for (i = 0; i < node->nvisit; i++) LiveCity[*(node->visited + i)] = 0;
}


trace_road (node)
  node_ *node;
{ int i, j, city_1, city_2; 

  if (node->nvisit >= 2) {
    for (i = 0, j = 1; j < node->nvisit; i++, j++) {
      city_1 = *(node->visited + i);
      city_2 = *(node->visited + j);
      VisitedRoad[i] = encodeW(city_1, city_2);
    }
  }
}

