
/* ******************************************************************** *
 * Traveling Salesperson Problem Generator:				*
 * 1. incomplete graph.							*
 * 2. asymmetric graph.							*
 * 3. triangular inequality is satisfied for existing arcs.		*
 *    (existing arcs == arcs of finite distance)			*
 * 4. completeness factor. [0,1]					*
 * ******************************************************************** */


void gen_sample_problem ()
{ float  best, tmp, dx, dy, dist, x[MaxProblemSize], y[MaxProblemSize];
  int i, j, best_one, the_end, the_one;

#ifndef SYM_TSP
  FILE *fp, *fopen ();
  No_Upper_Bound = YES;
#endif

  nCities = problem.size;

  /** build the map **/
  for (i = 0; i < nCities; i++) {
    /* create a new city */
    x[i] = gen_random_int (0, GeoLimit);
    y[i] = gen_random_int (0, GeoLimit);
    /* check if the new city overlaps other cities */
    for (j = 0; j < i; j++)
      if (x[i] == x[j])
        while (y[i] == y[j])
	  /* repeat trying until no overlap */
          y[i] = gen_random_int (0, GeoLimit);
  }

  /** init weight matrix **/
  for (i = 0; i < nCities; i++)
    for (j = 0; j < nCities; j++)
      Weight(i,j) = problem.huge;

  /** calc distance **/
  nRoads = 0;
  for (i = 0; i < nCities; i++) {
    for (j = i+1; j < nCities; j++) {
      if (gen_random_float () <= complFactor ||
          j == i+1 ||
          (i == 0 && j == nCities-1)) {
        dx = x[i] - x[j];
        dy = y[i] - y[j];
        dist = dx * dx + dy * dy;
        Weight(i,j) = Weight(j,i) = (domain) sqrt (todouble (dist));
	sortW(nRoads) = encodeW(i,j);
        nRoads++;
      }
    }
  }

  /* establish the sorted version of weights of edges */
  for (i = 0; i < nRoads-1; i++) {
    best_one = i;
    best = decodeW(sortW(best_one));
    for (j = i+1; j < nRoads; j++)
      if (decodeW(sortW(j)) < best) { best_one = j; best = decodeW(sortW(j)); }
    tmp = sortW(i); sortW(i) = sortW(best_one); sortW(best_one) = tmp;
  }

#ifdef DEBUG
if (cmd.debug >= 10) print_conf(stdout);
if (cmd.debug >= 10) {
  printf("\n");
  for (i = 0; i < nRoads; i++)
    printf("(sortW[%d] = %d, %f), ", i, sortW(i), decodeW(sortW(i)));
}
#endif

  /* PDSD size */
  set_node_size((nCities + nCities) * sizeof(int));

#ifndef SYM_TSP
  fp = fopen (".ats", "a");
  fprintf (fp, "ATSP(%d,%d,%d)\n", problem.size, nRoads, problem.seed);
  fclose (fp);
#endif
}

