
domain grcs_eval_lowb (), grcs_eval_upb ();


void evaluate_lower_bound (node)
  node_ *node;
{
#ifdef DEBUG
if (cmd.debug >= 4) printf("eval lowb\n");
#endif

  node->lowb = is_feasible (node) ? node->g_cost : grcs_eval_lowb (node);
}


domain grcs_eval_lowb (node)
  node_ *node;
{ domain lowb = node->g_cost;
  domain texec = (domain) 0.0;
  domain tavail = (domain) 0.0;
  domain tmax = (domain) 0.0;
  int i, j, who;

  /** calc processors' bottleneck **/

  for (i = 0; i < nTasks; i++)
    if (is_task_alive (node, i)) texec += execTime[i];

  for (i = 0; i < nProcessors; i++)
    if (tmax < *(node->proc_ck+i)) tmax = *(node->proc_ck+i);

  tavail = (domain) 0.0;
  for (i = 0; i < nProcessors; i++) tavail += tmax - *(node->proc_ck+i);

  texec -= tavail;
  if (texec > (domain) 0.0) lowb += (texec / nProcessors);

  /** calc resources' bottlenecks **/
  for (j = 0; j < nResources; j++) {
    texec = (domain) 0.0;
    for (i = 0; i < nTasks; i++)
      if (is_task_alive (node, i) && REQ[i][j]) texec += execTime[i];
    if (lowb < texec) lowb = texec;
  }

  if (node->parent)
    if (lowb < node->parent->lowb) lowb = node->parent->lowb;

  return lowb;
}


solution_ *evaluate_upper_bound (node, sol)
  node_ *node;
  solution_ *sol;
{ 
#ifdef DEBUG
if (cmd.debug >= 4) printf("eval upb\n");
#endif

  node->upb = is_feasible (node) ? node->g_cost : grcs_eval_upb (node, sol);

  return sol;
}


domain r_ck[MaxProblemSize], p_ck[MaxProblemSize]; 
domain grcs_eval_upb (node, sol)
  node_ *node;
  solution_ *sol;
{ domain tmax = (domain) 0.0;
  int i, j, who;
  domain alloc_proc ();

  for (i = 0; i < nResources; i++) r_ck[i] = *(node->resource_ck+i);
  for (i = 0; i < nProcessors; i++) p_ck[i] = *(node->proc_ck+i);

  for (i = 0; i < nTasks; i++)
    if (is_task_sched (node, i)) sol->schedTime[i] = get_task_time (node, i);

  for (i = 0; i < nTasks; i++) {
    if (is_task_alive (node, i)) {

      tmax = alloc_proc (p_ck, &who); 

      for (j = 0; j < nResources; j++)
        if (REQ[i][j])
	  /** resource j is required **/
	  if (tmax < r_ck[j]) tmax = r_ck[j];

      sol->schedTime[i] = tmax;
      tmax += execTime[i];
      p_ck[who] = tmax;

      for (j = 0; j < nResources; j++)
        if (REQ[i][j]) r_ck[j] = tmax;
    }
  }

  tmax = (domain) 0.0;
  for (i = 0; i < nProcessors; i++)
    if (tmax < p_ck[i]) tmax = p_ck[i];
 
  return tmax;
}


domain alloc_proc (p_ck_arr, whop)
  domain p_ck_arr[];
  int *whop;
{ domain tmin;
  int i;

  *whop = 0;
  tmin = p_ck_arr[0];
  for (i = 1; i < nProcessors; i++)
    if (tmin > p_ck_arr[i]) {
      tmin = p_ck_arr[i];
      *whop = i;
    }

  return tmin;
}

