
#define	d_node_	struct d_node__
d_node_ {
	int 	w;
	int	cost;
	node_	*who;		/* pointer to search node */
	d_node_	*next;		/* pointer to next d_node */
} *d_node_manager = NULL, *Dom[MaxProblemSize + 2];

extern void dom_init(), d_list_free(), d_node_free(), d_node_delete();
extern d_node_ *d_node_alloc();


void dom_init ()
{ int  i;

  for (i = 0; i <= problem.size + 1; i++) {
    d_list_free(Dom[i]);
    Dom[i] = NULL;
  }
}


void d_list_free (p)
  d_node_ *p;
{ d_node_ *q;

  while (p) { q = p->next; d_node_free(p); p = q; }
}


void d_node_free (p)
  d_node_ *p;
{
  p->next = d_node_manager;
  d_node_manager = p;
}


d_node_ *d_node_alloc (node)
  node_ *node;
{ d_node_ *p;
  int i;

  /* if the pool of d nodes are empty, then */
  /* allocate 1K d nodes at a time to avoid spreadout of */
  /* d nodes to a large range of virtual pages */
  if (! d_node_manager) {
    p = d_node_manager = (d_node_ *) malloc(1024 * sizeof(d_node_));
    for (i = 0; i < 1023; i++) (p + i)->next = p + i + 1;
    (p + 1023)->next = NULL;
  /*  memory_update( 1024 * sizeof(d_node_));  */
  }

  d_node_manager = (p = d_node_manager)->next;
  p->who = node;
  p->w = node->w;
  p->cost = node->g_cost;
  return p;
}


int is_dominated (node)
  node_ *node;
{ domain g_cost = node->g_cost;
  int depth = node->depth;
  int w = node->w;
  int exist_bit = 0;
  d_node_ *p, *q, *r;

  /* even if DFS, dominance is still necessary
   * if(Search_Strategy == _DFS_) { return(0); }
   */

  /* if the dom list of this depth is empty
   * then just insert it and return not-dominated
   */
  if (! Dom[depth]) { 
    (Dom[depth] = d_node_alloc(node))->next = NULL;
    return 0;
  }

  /* trace down the dom list of this depth
   * let cost be the penalty 
   * j is dominated by i, if
   * w(j) <= w(i)  &&  cost(j) >= cost(i)
   */
  for (p = q = Dom[depth]; p; p = (q = p)->next) {
    if (w <= p->w) {
      if (g_cost >= p->cost) {
        /* node is dominated by p */
        if (node == p->who) { exist_bit = 1; continue;}/* dominated by itself */
        /* else,
         * node is dominated by someone else
         * locate node's d_node and delete it
         */
        else { d_node_delete(node); return 1; }
      }
    }
    else {
      /* w > p->w */
      if (! exist_bit) {
        /* allocate an entry to this dom list */
        r = d_node_alloc(node);
        if (q == Dom[depth] && q == p) (Dom[depth] = r)->next = p;
        else (q->next = r)->next = p;
      }
      return 0;
    }
  }

  /* pass through all the elements of the dom list */
  /* all elements' w's are smaller than node's w */
  /* then the loop will end with q is the last element of the list */
  if (! exist_bit) (q->next = d_node_alloc(node))->next = NULL;

  return 0;
}


void d_node_delete (node)
  node_ *node;
{ d_node_ *p, *q;

  for (q = p = Dom[node->depth]; p; p = (q = p)->next)
    if (node == p->who) {
      if (q == Dom[node->depth] && q == p) Dom[node->depth] = p->next;
      else q->next = p->next;
      d_node_free(p);
      break;
    }
}

