
typedef float decomp_;
decomp_ decomp_h_fun ();
decomp_ dh_val[MaxProblemSize];


node_ *expand (node, child_type, nchild)
    node_ *node;
    child_ child_type;
    int nchild;
{   int entity = node->entity; 
    domain g_cost = node->g_cost;
    node_ *left_child, *right_child;
    int i, who, gen_left, gen_left_only;
    decomp_ best;

    if (node->ndecomp >= 2) return NULL;  /** infeasible **/

    /** binary decomposition (take or not take) **/
    /** decomposition indicates which vertice should be considered first **/
 
    if (child_type == NEXT_N_CHILDREN && nchild == 1)
	child_type = NEXT_CHILD;

    /** check whether the left child needs to be generated or not **/
    gen_left_only = (child_type == NEXT_CHILD && node->ndecomp == 0) ? 1 : 0;
    gen_left = (child_type == ALL_CHILDREN || gen_left_only) ? 1 : 0;

    /** init, less penalty is better **/
    for (i = 0; i < U; i++) dh_val[i] = HUGE_FLOAT;

    /** consider all live children **/
    for (i = 0; i < U; i++)
	if (*(node->item+i))
	    dh_val[i] = decomp_h_fun (P[i+1], W[i+1], node->depth);

    /** find the best decomposition by decomp heuristic values **/
    best = HUGE_FLOAT;
    who = -1;
    for (i = 0; i < U; i++)
	if (best > dh_val[i]) { best = dh_val[i]; who = i+1; }

    if (who == -1) return NULL;   /** infeasible **/

    if (gen_left)
	left_child = allocate_node (node, who, g_cost+P[who], node->w+W[who]);
    if (gen_left_only) { node->ndecomp = 1; return left_child; }

    right_child = get_search_node_from_pool ();
    init_node_struct (right_child, node);
    right_child->entity = node->entity;
    right_child->g_cost = node->g_cost;
    right_child->w = node->w;
    right_child->lowb = node->lowb;
    right_child->upb = node->upb;
    for (i = 0; i < U; i++) *(right_child->item+i) = *(node->item+i);
    *(right_child->item+who-1) = 0;

    node->ndecomp = 2;
    if (gen_left) {
	left_child->brother = right_child;
	return left_child;
    }
    else return right_child;
}


decomp_ decomp_h_fun (p, w, d)
    int p, w, d;
{
#ifdef HARE_DECOMP
    float DECOMP_H ();
    return (DECOMP_H (p, w, d));
#else
    return (tofloat (p) / tofloat (w));
#endif
}


int is_feasible (node)
  node_ *node;
{ return (node->w >= BOUND); }


int is_infeasible (node)
  node_ *node;
{ return 0; }
