
domain pp_eval_lowb (), pp_eval_upb ();


void evaluate_lower_bound (node)
  node_ *node;
{
  node->lowb = is_feasible (node) ? node->g_cost : pp_eval_lowb(node);
}


domain pp_eval_lowb (node)
  node_ *node;
{ int k = node->entity + 1;
  int m, i, j; 
  domain i_value = node->i_value;
  domain cost = node->g_cost;
  domain backlog, diff, x[MaxProblemSize];

  /* reset x array */
  for (i = k; i <= problem.size; i++) x[i] = 0;

  /* backlogging over [k ... i] without considering h effect */
  for (i = k; i <= problem.size; i++)
    if ((backlog = rParam[i] - i_value) < 0) { i_value -= rParam[i]; x[i] = 0; }
    else {
      i_value = 0;
      for (j = 1; j <= problem.size; j++) {
        m = pSort[j];
        if (in(m,k,i))
            if (backlog > (diff = cParam[m] - x[m])) {
              x[m] = cParam[m];
              backlog -= diff;
            }
            else { x[m] += backlog; backlog = 0; break; }
      }

      if (backlog > 0)    /* infeasible solution */ return(problem.huge);
    }

  /* calculate the cost */
  for (i = k; i <= problem.size; i++)
    if (x[i]) cost += (pParam[i] * x[i]);

  return cost;
}


solution_ *evaluate_upper_bound (node, sol)
  node_ *node;
  solution_ *sol;
{
    node->upb = is_feasible (node) ? node->g_cost : pp_eval_upb(node);
  return sol;
}


domain pp_eval_upb (node)
  node_ *node;
{ domain i_value = node->i_value;
  domain cost = node->g_cost;
  domain backlog, diff, x[MaxProblemSize];
  int k = node->entity + 1;
  int i, j;

  /* reset x array */
  for (i = k; i <= problem.size; i++) x[i] = 0;

  /* greedy backlogging over [i ... k]- */
  for (i = k; i <= problem.size; i++) 
    if ((backlog = rParam[i] - i_value) < 0) { i_value -= rParam[i]; x[i] = 0; }
    else {
      i_value = 0;
      for (j = i; j >= k; j--)
        if (backlog > (diff = cParam[j] - x[j])) {
          x[j] = cParam[j];
          backlog -= diff;
        }
        else {
          x[j] += backlog;
          backlog = 0;
          break;
        }
      /* infeasible */
      if (backlog > 0) return(problem.huge);
    }

  /* calculate the cost */
  i_value = node->i_value;
  for (i = k; i <= problem.size; i++) {
    i_value += (x[i] - rParam[i]);
    cost += (hParam[i] * i_value);
    if (x[i]) cost += (bParam[i] + pParam[i] * x[i]);
  }
  return cost;
}

