
void evaluate_lower_bound (node)
  node_ *node;
{
#ifdef DEBUG
if (cmd.debug >= 4) printf("eval lowb\n");
#endif
  if (is_compound(node)) update_bounds(node);
  else node->lowb = (is_feasible(node)) ? node->g_cost : _eval_lowb(node);
}


domain _eval_lowb (node)
  node_ *node;
{ int ntask, nproc, i, j, t;
  float val, max_one, ftmp;
  int b[MaxResource];

  /* calculate floor(T/M) */
  for (nproc = 0, i = 0; i < M; i++)
    if (*(node->proc + i)) nproc++;
  for (ntask = 0, t = 0; t < T; t++)
    if (*(node->live_task + t)) ntask++;
  val = (ntask > nproc) ? (((float) (ntask - nproc)) / M) : 0.0;

  /* calculate max j { [ sum Rjt ] / Bj } */
  for (j = 0; j < R; j++) b[j] = - *(node->rb + j);
  for (t = 0; t < T; t++)
    if (*(node->live_task + t))
      for (j = 0; j < R; j++) b[j] += REQ[t][j];

  /* lowb is max { T/M, [sum Rjt]/Bj } */
  for (max_one = val, j = 0; j < R; j++) {
    ftmp = tofloat(b[j]) / BOUND[j];
    if (max_one < ftmp) max_one = ftmp;
  }

  return (node->g_cost + toint(max_one));
}


solution_ *evaluate_upper_bound (node, sol)
  node_ *node;
  solution_ *sol;
{ 
#ifdef DEBUG
if (cmd.debug >= 4) printf("eval upb\n");
#endif
  if (! is_compound(node))
    node->upb = (is_feasible(node)) ? node->g_cost : _eval_upb(node, sol);
  return sol;
}


domain _eval_upb (node, sol)
  node_ *node;
  solution_ *sol;
{ int u = node->g_cost;
  int i, j, t, best, best_one, sum; 
  int task[MaxProblemSize], b[MaxResource], proc[MaxProcessor];

  /* preset resource bound, task, and proc */
  for (j = 0; j < R; j++) b[j] = *(node->rb + j);
  for (i = 0; i < M; i++) proc[i] = *(node->proc + i);
  for (t = 0; t < T; t++) task[t] = *(node->live_task + t);

  /* simulate the greedy solution */
  while (is_any(task,0,T-1)) {
    for (best_one = -1, best = problem.huge, t = 0; t < T; t++)
      if (task[t] && is_any(proc,0,M-1)) {
	for (sum = 0, j = 0; j < R; j++)
	  if (b[j] > REQ[t][j]) sum += (b[j] - REQ[t][j]);
	  else { sum = -1; break; }
	if (sum >= 0)
	  if (sum < best) { sum = best; best_one = t; }
      }
    if (best_one != -1) {
      task[best_one] = 0;
      for (j = 0; j < R; j++) b[j] -= REQ[best_one][j]; 
      for (i = 0; i < M; i++)
      if (proc[i]) { proc[i] = 0; break; }
    }
    else {
      for (j = 0; j < R; j++) b[j] = BOUND[j];
      for (i = 0; i < M; i++) proc[i] = 1;
      u++;
    }
  }

  return u;
}


void update_bounds (node)
  node_ *node;
{ int entity = node->entity;
  domain g_cost = node->g_cost;
  domain value = node->value;
  int round = node->round;
  domain lowb = problem.huge;
  int i, j, val;
  int rb[MaxProblemSize], proc[MaxProblemSize]; 
#ifdef DEBUG
if (cmd.debug >= 4) printf("update bounds\n");
#endif

  for (i = 0; i < T; i++)
    if (*(node->open_task + i)) {
      for (j = 0; j < R; j++) rb[j] = *(node->rb + j);
      for (j = 0; j < M; j++) proc[j] = *(node->proc + j);

      *(node->live_task + (node->entity = i)) = 0;
      if (! satisfy_constr(node->entity,node)) another_round(node);
      for (j = 0; j < R; j++) *(node->rb + j) -= REQ[node->entity][j];
      val = _eval_lowb (node);
      if (val < lowb) lowb = val;

      *(node->live_task + i) = 1;
      node->entity = entity;
      node->g_cost = g_cost;
      node->value = value;
      node->round = round;
      for (j = 0; j < R; j++) *(node->rb + j) = rb[j];
      for (j = 0; j < M; j++) *(node->proc + j) = proc[j];
    }

  node->lowb = lowb;
}

