
node_ *expand (node, child_type, nchild)
  node_ *node;
  child_ child_type;
  int nchild;
{ int left_child_entity;
  node_ *left, *right;

  if (node->ndecomp >= 2 || (left_child_entity = decomposition(node)) < 0)
    return NULL;

  switch (child_type) {
    case ALL_CHILDREN:
    case NEXT_N_CHILDREN:
      left = get_left_child(node, left_child_entity);
      right = get_right_child(node, left_child_entity);
      (left->brother = right)->brother = NULL;
      return left;
    case NEXT_CHILD:
      switch (node->ndecomp) {
        case 0:  /* return left child */
          (left = get_left_child(node, left_child_entity))->brother = NULL;
          return left;
        case 1:  /* return right child */
          (right = get_right_child(node, left_child_entity))->brother = NULL;
          return right;
        default: break;
      }
    default: break;
  }    
  return NULL;
}


node_ *get_left_child (node, left_child_entity)
  node_ *node;
  int left_child_entity;
{
  node->ndecomp = 1;
  return allocate_node( node,
			left_child_entity,
			alloc_resource(node->proc, 0, M-1),
		 	node->g_cost + node->round);
}
 

node_ *get_right_child (node, left_child_entity)
  node_ *node;
  int left_child_entity;
{ node_ *right;

  node->ndecomp = 2;
  (right = sprout_child(node))->ndecomp = 0;
 
  set_compound(right);
  *(right->open_task + left_child_entity) = 0;
  return right;
}


int alloc_resource (base, begin, end) 
  int *base, begin, end;
{ int k;

  for (k = begin; k <= end; k++)
    if (*(base + k)) { *(base + k) = 0; return k; }
  return 0;
}

 
/* select a decomposition, return entry -1 if no such value available */
int decomposition (node)
  node_ *node;
{ int best_one = -1, t, count = 1;
  float best = HUGE_FLOAT, estimate;

  /* if no open task, then return -1 */
  if (! is_any(node->open_task, 0, T-1)) return -1;

  /* if no processor is available, signal 0 to activate another round */
  if (! alloc_resource(node->proc, 0, M-1)) another_round(node);

  /* try each open task, pick the best one */
  while (count <= 2) {
    for (t = 0; t < T; t++)
      if (*(node->open_task + t)) 
	if (satisfy_constr(t, node)) {
	  estimate = h_fun(t, node->rb, node->proc, node->open_task);
	  if (estimate < best) { best = estimate; best_one = t; } /* most fit */
  	}

	if (best_one != -1) break;

	another_round(node);
	count++;
  }

  return best_one;
}


float h_fun (task, rbp, procp, taskp)
  int task, *rbp, *procp, *taskp;
{ int n_proc = 0, n_task = 0, i, j, t;
  float fmin, ftmp;

  /* calculate # processors available */
  for (i = 0; i < M; i++)
    if (*(procp + i)) n_proc++;

  /* calculate # tasks to solve */
  for (t = 0; t < T; t++)
    if(*(taskp + t)) n_task++;

  /* pick the min heuristic values */
  fmin = HUGE_FLOAT;
  for (j = 0; j < R; j++) {
    ftmp = _h_fun(*(rbp + j), REQ[task][j], n_proc, n_task);
    if (ftmp < fmin) fmin = ftmp;
  }

  return fmin;
}


float _h_fun (b, r, np, nt)
  int b, r, np, nt;
{ int val = b - r;

  return (tofloat(val));
}


int satisfy_constr (task, node)
  int task; 
  node_ *node;
{ int j;

  for (j = 0; j < R; j++)
    if (REQ[task][j] > *(node->rb + j)) return 0;
  return 1;
}


void another_round (node)
  node_ *node;
{ int i, j;

  /* advance g_cost */
  node->round++;

  /* reset resouce bound */
  for (j = 0; j < R; j++) *(node->rb + j) = BOUND[j];

  /* free all processors */
  for (i = 0; i < M; i++) *(node->proc + i) = 1;
}


int is_infeasible (node)
  node_ *node;
{ int t;

  /* T is the absolute max limit */
  if (node->g_cost > T) return 1;

  return 0;
}


int is_feasible (node)
  node_ *node;
{ int t;

  for (t = 0; t < T; t++)
    if (*(node->live_task + t)) return 0;

  return 1;
}

