
domain vc_eval_lowb (), vc_eval_upb ();

void evaluate_lower_bound (node)
    node_ *node;
{
    node->lowb = is_feasible (node) ? node->g_cost : vc_eval_lowb (node);
}

int sort_vertices[MaxProblemSize];

domain vc_eval_lowb (node)
    node_ *node;
{   int num_edges = node->num_edges;
    domain lowb = node->g_cost;
    int neleft = NumEdges - node->num_edges;	/** num of edges uncovered **/
    int sum, nalive, who, best, temp, i, j;

    /** record who are alive **/
    for (nalive = i = 0; i < NumVertices; i++)
	if(is_vertex_alive (node, i)) sort_vertices[nalive++] = i;

    if (nalive <= 0) return problem.huge;   /** infeasible **/

    /** sort these alive **/
    for (i = 0; i < nalive-1; i++) {
	who = i;
	best = get_vertex_degree (node, sort_vertices[who]);
	for (j = i+1; j < nalive; j++)
	    if (best < get_vertex_degree (node, sort_vertices[j])) {
		/** update who and best **/
		who = j;
		best = get_vertex_degree (node, sort_vertices[j]);
	    }
        /** swap i and who **/
	temp = sort_vertices[i];
	sort_vertices[i] = sort_vertices[who];
	sort_vertices[who] = temp;
    }

    /** calc lower bound **/
    for (sum = i = 0; i < nalive; i++) {
	lowb++;
	sum += get_vertex_degree (node, sort_vertices[i]);
	if (sum >= neleft) {
	    if (node->parent)
		if (node->parent->lowb > lowb) lowb = node->parent->lowb;
	    return lowb;
	}
    }

    return problem.huge;
}


solution_ *evaluate_upper_bound (node, sol)
    node_ *node;
    solution_ *sol;
{
    node->upb = vc_eval_upb (node, sol);
    return sol;
}


int check_edges[MaxNumberEdges];

domain vc_eval_upb (node, sol)
    node_ *node;
    solution_ *sol;
{   domain upb = node->g_cost;
    int num_edges = 0;
    int sum, nalive, who, best, temp, i, j;

    /** record who are alive **/
    for (nalive = i = 0; i < NumVertices; i++)
	if(is_vertex_alive (node, i)) sort_vertices[nalive++] = i;

    if (nalive <= 0) return problem.huge;  /** infeasible **/

    /** sort these alive **/
    for (i = 0; i < nalive-1; i++) {
	who = i;
	best = get_vertex_degree (node, sort_vertices[who]);
	for (j = i+1; j < nalive; j++)
	    if (best < get_vertex_degree (node, sort_vertices[j])) {
		/** update who and best **/
		who = j;
		best = get_vertex_degree (node, sort_vertices[j]);
	    }
        /** swap i and who **/
	temp = sort_vertices[i];
	sort_vertices[i] = sort_vertices[who];
	sort_vertices[who] = temp;
    }

    /** clear up edge-checking array **/
    for (i = 0; i < NumEdges; i++) check_edges[i] = VC_ALIVE;

    /** fill in edge-checking array **/
    for (i = 0; i < NumVertices; i++) {
	if (is_vertex_taken (node, i)) {
	    sol->vertices[i] = 1;
	    num_edges += mark_edges (i, check_edges, VC_TAKEN);
	}
    }

    for (i = 0; i < nalive; i++) {
	if (num_edges < NumEdges) {
	    upb++;
	    sol->vertices[sort_vertices[i]] = 1;
	    num_edges += mark_edges (sort_vertices[i], check_edges, VC_TAKEN);
	}
	else break;
    }

    return ((num_edges >= NumEdges) ? upb : problem.huge);
}


mark_edges (i, arr, mark)
    int i;
    int arr[];
    int mark;
{   int j, count = 0;

    for (j = 0; j < NumVertices; j++)
	if (Adjacency[i][j] >= 0)
	    if (arr[Adjacency[i][j]] == VC_ALIVE) {
		arr[Adjacency[i][j]] = mark;
		count++;
	    }

    return count;
}

